(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit uRenameRefactoring;

interface

uses
  ucwClasses,
  ucwRefactoring;

type
  TRenameRefactoring = class(TcwRefactoring)
  private
    procedure RenameRefactor(const ASymbol: TcwSymbol; const ANewName: string;
      const ABaseFolderName: string);
  protected
    procedure Execute; override;
    function GetCanExecute: Boolean; override;
    function GetName: String; override;
    function GetMenuCaption: String; override;
  end;

implementation

uses
  Windows,
  Dialogs,
  SysUtils,
  Classes,
  Contnrs,
  mwPasLexTypes,
  ucwMessage,
  ucwRefactorNodes;

procedure TRenameRefactoring.Execute;
var
  ABaseFolderName: string;
  ANewName: string;
begin
  ANewName := gcwRefactoringHost.SelectedSymbol.Name;
  if InputQuery('Rename', 'Rename', ANewName) then
  begin
    ABaseFolderName := GetBaseFolderName;
    RenameRefactor(gcwRefactoringHost.SelectedSymbol, ANewName, ABaseFolderName);
    ViewRefactoring(ABaseFolderName);
  end;
end;

function TRenameRefactoring.GetCanExecute: Boolean;
begin
  Result := Assigned(gcwRefactoringHost.SelectedSymbol);
end;

function TRenameRefactoring.GetMenuCaption: String;
begin
  Result := 'Rename...';
end;

function TRenameRefactoring.GetName: String;
begin
  Result := 'Rename Refactoring';
end;

procedure TRenameRefactoring.RenameRefactor(const ASymbol: TcwSymbol; const ANewName: string;
  const ABaseFolderName: string);

  procedure RenameRefactorLocation(AModuleNode: TcwModuleNode; const ALocation: TcwCodeLocation;
    const ANewName: string);
  var
    AStrConstNode: TcwTokenNode;
    ATokenNode: TcwTokenNode;
  begin
    ATokenNode := FindNodeForLocation(AModuleNode, ALocation) as TcwTokenNode;
    if Assigned(ATokenNode) then
    begin
      Assert(ATokenNode.TokenKind = ptIdentifier);
      ATokenNode.Code := ANewName;

      if ASymbol is TcwUnit then
      begin
        // check to see if ATokenNode is a unit name, e.g. uMyUnit in 'uMyUnit.pas'
        // if so, rename the unit name in the file specified...
        if Assigned(ATokenNode.GetPriorPeerTokenNode(ptUses)) and
          Assigned(ATokenNode.GetNextPeerTokenNode(ptIn)) then
        begin
          AStrConstNode := ATokenNode.GetNextPeerTokenNode(ptIn).
            GetNextPeerTokenNode(ptStringConst);
          if Assigned(AStrConstNode) then
            AStrConstNode.Code := StringReplace(AStrConstNode.Code, ASymbol.Name, ANewName,
              [rfIgnoreCase]);
        end;
      end;
    end else
      gcwRefactoringHost.AcceptRefactoringMessage(Self,
        TcwMessage.Create('Token not found', ALocation.UnitIntf.FileIndex, ALocation.Y,
          ALocation.X, False));
  end;

var
  Index: Integer;
  ALocation: TcwUsedByLocation;
  AUnitSymbol: TcwUnit;
  AUnitPath: string;
  AUnitFullFileName: string;
  ADFMName: string;
  ANewDFMName: string;
  ADFMPath: string;
begin
  // first rename the symbol's declaration...
  BeginRefactoring(ABaseFolderName, ASymbol.UnitIntf);
  RenameRefactorLocation(ModuleNode[ASymbol.UnitIntf], ASymbol, ANewName);
  
  // rename all places the symbol is referenced...
  for Index := 0 to ASymbol.UsedBy.Count - 1 do
  begin
    ALocation := ASymbol.UsedBy.Locations[Index];
    if not Assigned(ModuleNode[ALocation.UnitIntf]) then
      BeginRefactoring(ABaseFolderName, ALocation.UnitIntf);
    RenameRefactorLocation(ModuleNode[ALocation.UnitIntf], ALocation, ANewName);
  end;

  // if the symbol is a unit, then generate to the new file name...
  if ASymbol is TcwUnit then
  begin
    AUnitSymbol := ASymbol as TcwUnit;
    AUnitPath := ExcludeTrailingPathDelimiter(ExtractFileDir(AUnitSymbol.FullFileName));
    AUnitFullFileName := Format('%s\%s%s', [AUnitPath, ANewName,
      ExtractFileExt(AUnitSymbol.FullFileName)]);
    ADFMName := Format('%s\%s.%s', [AUnitPath, AUnitSymbol.Name, 'dfm']);
    if FileExists(ADFMName) then
    begin
      ANewDFMName := GetFilenameToGenerate(ADFMName, ABaseFolderName, True, ADFMPath);
      ANewDFMName := Format('%s\%s.%s', [ADFMPath, ANewName, 'dfm']);
      // rename the dfm file too...
      CopyFile(PChar(ADFMName), PChar(ANewDFMName), True);
    end;
    ModuleNode[AUnitSymbol].Generate(GetFilenameToGenerate(AUnitFullFileName, ABaseFolderName,
      True, AUnitPath), not RefactorInPlace);
    // delete from list since we've generated manually
    RefactoringList.Delete(RefactoringList.IndexOf(AUnitSymbol));
  end;
  EndRefactorings(ABaseFolderName);
end;

var
  ARenameRefactoring: TcwRefactoring;

initialization

  ARenameRefactoring := TRenameRefactoring.Create;
  gcwRefactoringHost.RegisterRefactoring(ARenameRefactoring);

finalization

  gcwRefactoringHost.UnregisterRefactoring(ARenameRefactoring);

end.
