(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit uUnUsedUnitRefactoring;

interface

uses
  ucwClasses,
  ucwRefactoring,
  ucwRefactorNodes,
  ucwMessage,
  frmUnUsedUnitRefactoring;

type
  TAbstractUnUsedUnitRefactoring = class(TcwRefactoring)
  private
    FUnUsedUnitRefactoringForm: TUnUsedUnitRefactoringForm;
    FDidRefactor: Boolean;
    FUnit: TcwUnit;
    function GetRemoveAllUnusedUnits: Boolean;
    function GetMoveUsedUnitsToImplementation: Boolean;
  protected
    procedure DoRemoveUnusedUnits(const AUnit: TcwUnit; const ABaseFolderName: string);
    procedure DoBeforeExecute(out AAbort: Boolean);
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TSingleUnUsedUnitRefactoring = class(TAbstractUnUsedUnitRefactoring)
  protected
    procedure Execute; override;
    function GetCanExecute: Boolean; override;
    function GetMenuCaption: string; override;
    function GetName: string; override;
  end;

  TMultiUnUsedUnitRefactoring = class(TAbstractUnUsedUnitRefactoring)
  protected
    procedure Execute; override;
    function GetCanExecute: Boolean; override;
    function GetMenuCaption: string; override;
    function GetName: string; override;
  end;

implementation

uses
  Controls,
  Types,
  SysUtils,
  Contnrs,
  mwPasLexTypes;

type
  // a unit to be removed
  TUnitToBeRemoved = class
  private
    FUsedSet: TVisibilityKindSet;
    FUnit: TcwUnit;
    FIdentifierNode: TcwCodeNode;
  public
    constructor Create(const AUnit: TcwUnit; AUsedSet: TVisibilityKindSet;
      AIdentifierNode: TcwCodeNode);
    procedure SetUsedSet(AUsedSet: TVisibilityKindSet);
    property UsedSet: TVisibilityKindSet read FUsedSet;
    property UnitIntf: TcwUnit read FUnit;
    property IdentifierNode: TcwCodeNode read FIdentifierNode;
  end;

  TUnitsToBeRemovedList = class
  private
    FObjectList: TObjectList;
    function GetUnitToBeRemoved(Index: Integer): TUnitToBeRemoved;
  public
    constructor Create;
    destructor Destroy; override;
    function Count: Integer;
    function Add(AUnitToBeRemoved: TUnitToBeRemoved): Integer;
    function IndexOf(AUnit: TcwUnit): Integer;
    procedure Delete(Index: Integer);
    property UnitsToBeRemoved[Index: Integer]: TUnitToBeRemoved read GetUnitToBeRemoved; default;
  end;

{ TUnitToBeRemoved }

constructor TUnitToBeRemoved.Create(const AUnit: TcwUnit;
  AUsedSet: TVisibilityKindSet; AIdentifierNode: TcwCodeNode);
begin
  inherited Create;
  FUnit := AUnit;
  FUsedSet := AUsedSet;
  FIdentifierNode := AIdentifierNode;
end;

procedure TUnitToBeRemoved.SetUsedSet(AUsedSet: TVisibilityKindSet);
begin
  FUsedSet := AUsedSet;
end;

{ TUnitsToBeRemovedList }

constructor TUnitsToBeRemovedList.Create;
begin
  inherited;
  FObjectList := TObjectList.Create(True);
end;

destructor TUnitsToBeRemovedList.Destroy;
begin
  FObjectList.Free;
  inherited;
end;

function TUnitsToBeRemovedList.Add(AUnitToBeRemoved: TUnitToBeRemoved): Integer;
begin
  Result := FObjectList.Add(AUnitToBeRemoved);
end;

function TUnitsToBeRemovedList.Count: Integer;
begin
  Result := FObjectList.Count;
end;

function TUnitsToBeRemovedList.GetUnitToBeRemoved(Index: Integer): TUnitToBeRemoved;
begin
  Result := FObjectList[Index] as TUnitToBeRemoved;
end;

function TUnitsToBeRemovedList.IndexOf(AUnit: TcwUnit): Integer;
var
  Index: Integer;
begin
  Result := -1;
  for Index := 0 to Count - 1 do
    if UnitsToBeRemoved[Index].UnitIntf.FileIndex = AUnit.FileIndex then
    begin
      Result := Index;
      Break;
    end;
end;

procedure TUnitsToBeRemovedList.Delete(Index: Integer);
begin
  FObjectList.Delete(Index);
end;

{ TAbstractUnUsedUnitRefactoring }

constructor TAbstractUnUsedUnitRefactoring.Create;
begin
  inherited;
  FUnUsedUnitRefactoringForm := TUnUsedUnitRefactoringForm.Create(nil);
end;

destructor TAbstractUnUsedUnitRefactoring.Destroy;
begin
  FUnUsedUnitRefactoringForm.Release;
  inherited;
end;

procedure TAbstractUnUsedUnitRefactoring.DoRemoveUnusedUnits(
  const AUnit: TcwUnit; const ABaseFolderName: string);

  procedure RemoveUsesListNode(var AUsesListNode: TcwUsesListNode;
    AUsesListVisibility: TVisibilityKind);

    procedure UnitToBeRemoved(var AUsesListNode: TcwUsesListNode; AIdentifierNode: TcwCodeNode);
    var
      AUsesNode: TcwTokenNode;
      ACommaNode: TcwTokenNode;
    begin
      FDidRefactor := True;
      ACommaNode := AIdentifierNode.GetNextPeerTokenNode(ptComma);
      if Assigned(ACommaNode) then
      begin
        TcwNode.DeleteIfExists(ACommaNode.GetNextPeerNodeClass(TcwJunkNode, False));
        ACommaNode.Delete;
      end else
        TcwNode.DeleteIfExists(AIdentifierNode.GetNextPeerNodeClass(TcwJunkNode, False));

      // if this is the last and only unit in the list then remove parent, and next semicolon...
      if AUsesListNode.IdentifierList.Count = 1 then
      begin
        TcwNode.DeleteIfExists(AUsesListNode.GetPriorPeerNodeClass(TcwJunkNode, False));
        AUsesNode := AUsesListNode.GetPriorPeerTokenNode(ptUses, False);
        if Assigned(AUsesNode) then
        begin
          TcwNode.DeleteIfExists(AUsesNode.GetPriorPeerNodeClass(TcwJunkNode, False));
          AUsesNode.Delete;
        end;
        TcwNode.DeleteIfExists(AUsesListNode.GetNextPeerTokenNode(ptSemiColon, False));
        AUsesListNode.Delete;
        AUsesListNode := nil;
      end else
      begin
        // if this is the last unit in the list, then remove the previous comma....
        if AIdentifierNode = AUsesListNode.
          IdentifierList[AUsesListNode.IdentifierList.Count - 1] then
        begin
          ACommaNode := AIdentifierNode.GetPriorPeerTokenNode(ptComma, True);
          if Assigned(ACommaNode) then
          begin
            TcwNode.DeleteIfExists(ACommaNode.GetNextPeerNodeClass(TcwJunkNode, False));
            ACommaNode.Delete;
          end;
        end;
        AIdentifierNode.Delete;
      end;
    end;

    procedure AddUnitToImplementationUses(AUnitName: string);

      procedure AddUnitToUsesListNode(AUnitName: string; AUsesListNode: TcwUsesListNode;
        AFirstUnit: Boolean);
      var
        AUsesNode: TcwTokenNode;
        ACode: string;
        AJunkNode: TcwJunkNode;
      begin
        AUsesNode := AUsesListNode.GetPriorPeerTokenNode(ptUses);
        if Assigned(AUsesNode) then
        begin
          if not AFirstUnit then
          begin
            AJunkNode := AUsesNode.GetNextPeerNodeClass(TcwJunkNode, False) as TcwJunkNode;
            if Assigned(AJunkNode) then
              ACode := AJunkNode.Code
            else
              ACode := '';
            TcwJunkNode.CreateCode(AUsesListNode, ACode, 0);
            TcwTokenNode.CreateCode(AUsesListNode, ',', ptComma, 0);
          end;
          TcwIdentifierNode.CreateCode(AUsesListNode, AUnitName, ptIdentifier, 0);
        end else
          gcwRefactoringHost.AcceptRefactoringMessage(Self, TcwMessage.Create(
            'Uses token not found.', FUnit.FileIndex,
            AUsesListNode.Point.X, AUsesListNode.Point.Y, False));
      end;

    var
      AJunkNode: TcwJunkNode;
      AImplementationNode: TcwTokenNode;
      AStartIndex: Integer;
      ADirectiveNode: TcwDirectiveNode;
      AUsesListNode: TcwUsesListNode;
      ASectionNode: TcwNode;
    begin
      ASectionNode := ModuleNode[FUnit].GetFirstNodeClass(TcwImplementationSectionNode);
      if Assigned(ASectionNode) then
      begin
        AUsesListNode := ASectionNode.GetFirstNodeClass(TcwUsesListNode) as TcwUsesListNode;
        if Assigned(AUsesListNode) then
          AddUnitToUsesListNode(AUnitName, AUsesListNode, False)
        else
        begin
          // find where to put uses clause...
          AStartIndex := -1;
          ADirectiveNode := ASectionNode.GetFirstNodeClass(TcwDirectiveNode) as TcwDirectiveNode;
          if Assigned(ADirectiveNode) and SameText(ADirectiveNode.Code, '{$R *.DFM}') then
            AStartIndex := ADirectiveNode.GetIndex
          else
          begin
            AImplementationNode := ASectionNode.GetFirstTokenNode(ptImplementation);
            if Assigned(AImplementationNode) then
              AStartIndex := AImplementationNode.GetIndex;
          end;
          AJunkNode := ASectionNode.GetNextNodeClass(AStartIndex, TcwJunkNode, False) as
            TcwJunkNode;
          if Assigned(AJunkNode) then
            AStartIndex := AJunkNode.GetIndex;
          Inc(AStartIndex);
          TcwJunkNode.CreateCode(ASectionNode, sLineBreak + sLineBreak, AStartIndex);
          TcwTokenNode.CreateCode(ASectionNode, ';', ptSemiColon, AStartIndex);
          // don't assign code to AUsesListNode, since we're adding child nodes later...
          AUsesListNode := TcwUsesListNode.Create(ASectionNode, Point(-1, -1), -1, AStartIndex);
          TcwJunkNode.CreateCode(ASectionNode, ' ', AStartIndex);
          TcwTokenNode.CreateCode(ASectionNode, 'uses', ptUses, AStartIndex);
          AddUnitToUsesListNode(AUnitName, AUsesListNode, True);
        end;
      end else
        gcwRefactoringHost.AcceptRefactoringMessage(Self, TcwMessage.Create(
          'Implementation section not found.', FUnit.FileIndex,
          FUnit.X, FUnit.Y, False));
    end;

    function UnitIsActuallyUsed(const AUnit: TcwUnit; out AUsedSet: TVisibilityKindSet ): Boolean;
    var
      I, J: Integer;
      ASymbol: TcwSymbol;
      ANode: TcwNode;
    begin
      Result := False;
      AUsedSet := [];
      for I := 0 to AUnit.SymbolCount - 1 do
      begin
        ASymbol := AUnit.Symbols[I];
        for J := 0 to ASymbol.UsedBy.Count - 1 do
          if ASymbol.UsedBy[J].FileIndex = FUnit.FileIndex then
          begin
            Result := True;
            if AUsesListVisibility = vkInterface then
            begin
              ANode := FindNodeForLocation(ModuleNode[FUnit], ASymbol.UsedBy[J]);
              if Assigned(ANode) and ANode.ParentedBy(TcwInterfaceSectionNode) then
                Include(AUsedSet, vkInterface)
              else
                // if assigned then it is implementation, may not be parented by the section node if
                // e.g. its in a initialization/finalization section...
                if Assigned(ANode) {and ANode.ParentedBy(TcwImplementationSectionNode) }then
                  Include(AUsedSet, vkImplementation);
              if AUsedSet = [vkInterface, vkImplementation] then
                Break;
            end else
              Break;
          end;

        if ((AUsesListVisibility = vkImplementation) and Result) or
          ((AUsesListVisibility = vkInterface) and
           (AUsedSet = [vkInterface, vkImplementation])) then
          Break;
      end;
    end;

    procedure RemoveUnusedUnits(AUnitsToBeRemovedList: TUnitsToBeRemovedList);

      procedure UndoUnitToBeRemovedForComponent(AUnitsToBeRemovedList: TUnitsToBeRemovedList;
        AFormFrameComponent: TcwComponent; AFormFrameClass: TcwClass; AComponent: TcwComponent;
        AComponentSymbol: TcwSymbol = nil);
      var
        AUnitIndex: Integer;
        ACurrentClass: TcwClass;
        AComponentIndex: Integer;
      begin
        // for each component on the form, look at its ancestory and for each class in the
        // ancestory, see if the class's unit is among our list to remove
        if not Assigned(AComponentSymbol) then
          AComponentSymbol := AFormFrameClass.FindSymbol(AUnit, AFormFrameClass, AComponent.Name);
        if Assigned(AComponentSymbol) and (AComponentSymbol.TypeIntf is TcwClass) then
        begin
          ACurrentClass := AComponentSymbol.TypeIntf as TcwClass;
          repeat
            AUnitIndex := AUnitsToBeRemovedList.IndexOf(ACurrentClass.UnitIntf);
            if AUnitIndex >= 0 then
              AUnitsToBeRemovedList.Delete(AUnitIndex);
            ACurrentClass := ACurrentClass.InheritsFromClass;
          until not Assigned(ACurrentClass);
        end;
        for AComponentIndex := 0 to AComponent.ChildComponents.Count - 1 do
          UndoUnitToBeRemovedForComponent(AUnitsToBeRemovedList, AFormFrameComponent,
            AFormFrameClass, AComponent.ChildComponents[AComponentIndex]);
      end;

    var
      ARemoveIndex: Integer;
      AFormFrameComponentSymbol: TcwSymbol;
      AFormFrameClass: TcwClass;
      AFormFrameComponent: TcwComponent;
      AUnitToBeRemoved: TUnitToBeRemoved;
      ADFMUnit: TcwUnit;
    begin
      if not GetRemoveAllUnusedUnits and (AUsesListVisibility = vkInterface) then
      begin
        ADFMUnit := FUnit;
        repeat
          if Assigned(ADFMUnit.DFMResource) and (ADFMUnit.DFMResource.Components.Count > 0) then
          begin
            AFormFrameComponent := ADFMUnit.DFMResource.Components[0];
            AFormFrameComponentSymbol := ADFMUnit.Classes.Find(AFormFrameComponent.CClassName);
            if AFormFrameComponentSymbol is TcwClass then
            begin
              AFormFrameClass := AFormFrameComponentSymbol as TcwClass;
              UndoUnitToBeRemovedForComponent(AUnitsToBeRemovedList, AFormFrameComponent,
                AFormFrameClass, AFormFrameComponent, AFormFrameComponentSymbol);

              if AFormFrameComponent.IsInherited and
                Assigned(AFormFrameClass.InheritsFromClass) and
                (AFormFrameClass.InheritsFromClass.UnitIntf.FileIndex <> ADFMUnit.FileIndex) then
                ADFMUnit := AFormFrameClass.InheritsFromClass.UnitIntf
              else
                Break;
            end else
              Break;
          end else
            Break;
        until False;
      end;

      // now remove units, or move them to implementation uses clause...
      for ARemoveIndex := 0 to AUnitsToBeRemovedList.Count - 1 do
      begin
        AUnitToBeRemoved := AUnitsToBeRemovedList[ARemoveIndex];
        if (AUsesListVisibility = vkImplementation) or (AUnitToBeRemoved.UsedSet = []) then
          UnitToBeRemoved(AUsesListNode, AUnitToBeRemoved.IdentifierNode)
        else
          if GetMoveUsedUnitsToImplementation and (AUsesListVisibility = vkInterface) and
            (AUnitToBeRemoved.UsedSet = [vkImplementation]) then
          begin
            AddUnitToImplementationUses(AUnitToBeRemoved.IdentifierNode.Code);
            UnitToBeRemoved(AUsesListNode, AUnitToBeRemoved.IdentifierNode);
          end;
      end;
    end;

  var
    AUsedSet: TVisibilityKindSet;
    AIdentifierNode: TcwCodeNode;
    AUnit: TcwUnit;
    Index: Integer;
    AUnitsToBeRemovedList: TUnitsToBeRemovedList;
  begin
    if Assigned(AUsesListNode) then
    begin
      AUnitsToBeRemovedList := TUnitsToBeRemovedList.Create;
      try
        Index := AUsesListNode.IdentifierList.Count - 1;
        while Index >= 0 do
        begin
          AIdentifierNode := AUsesListNode.IdentifierList[Index];
          AUnit := gcwProject.AllUnits.FindUnit(AIdentifierNode.Code);
          if Assigned(AUnit) then
          begin
            if not (uwInterface in AUnit.UnitWarnings) then
            begin
              if not UnitIsActuallyUsed(AUnit, AUsedSet) or
                ((AUsesListVisibility = vkInterface) and
                 ((AUsedSet = []) or (AUsedSet = [vkImplementation]))) then
                AUnitsToBeRemovedList.Add(TUnitToBeRemoved.Create(AUnit, AUsedSet, AIdentifierNode));
            end else
            begin
              gcwRefactoringHost.AcceptRefactoringMessage(Self, TcwMessage.Create(Format(
                'Refactoring cancelled since used unit, %s, has errors in its interface.',
                [AUnit.Name]), FUnit.FileIndex, AUnit.X, AUnit.Y, False));
              FDidRefactor := False;
              Break;
            end;
          end else
          begin
            gcwRefactoringHost.AcceptRefactoringMessage(Self, TcwMessage.Create(Format(
              'Used unit, %s, not found in project.',
              [AUsesListNode.IdentifierList[Index].Code]), FUnit.FileIndex,
              AIdentifierNode.Point.X, AIdentifierNode.Point.Y, False));
            FDidRefactor := False;
            Break;
          end;
          Dec(Index);
          if not Assigned(AUsesListNode) then
            Break;
        end;

        RemoveUnusedUnits(AUnitsToBeRemovedList);
      finally
        AUnitsToBeRemovedList.Free;
      end;
    end;
  end;

var
  ASectionNode: TcwNode;
  AUsesListNode: TcwUsesListNode;
begin
  FUnit := AUnit;
  if FUnit.UnitType = dtProject then
  begin
    Assert(Assigned(FUnit));
    FDidRefactor := False;
    BeginRefactoring(ABaseFolderName, FUnit);
    try
      if FUnit.UnitWarnings = [] then
      begin
        if Assigned(ModuleNode[FUnit]) then
        begin
          ASectionNode := ModuleNode[FUnit].GetFirstNodeClass(TcwInterfaceSectionNode);
          if Assigned(ASectionNode) then
          begin
            AUsesListNode := ASectionNode.GetFirstNodeClass(TcwUsesListNode) as TcwUsesListNode;
            RemoveUsesListNode(AUsesListNode, vkInterface);
            ASectionNode := ModuleNode[FUnit].GetNextNodeClass(ASectionNode.GetIndex,
              TcwImplementationSectionNode);
            AUsesListNode := ASectionNode.GetFirstNodeClass(TcwUsesListNode) as TcwUsesListNode;
            RemoveUsesListNode(AUsesListNode, vkImplementation);
          end else
          begin
            AUsesListNode := ModuleNode[FUnit].GetFirstNodeClass(TcwUsesListNode) as
              TcwUsesListNode;
            RemoveUsesListNode(AUsesListNode, vkImplementation);
          end;
        end else
          gcwRefactoringHost.AcceptRefactoringMessage(Self, TcwMessage.Create(Format(
            'Unit, %s, not found in project.', [FUnit.Name]), FUnit.FileIndex, FUnit.X, FUnit.Y,
            False));
      end else
        gcwRefactoringHost.AcceptRefactoringMessage(Self, TcwMessage.Create(Format(
          'Unit, %s, has errors.', [FUnit.Name]), FUnit.FileIndex, FUnit.X, FUnit.Y,
          False));
    finally
      if FDidRefactor then
        EndRefactoring(ABaseFolderName, FUnit)
      else
        CancelRefactoring(FUnit);
    end;
  end;
end;

procedure TAbstractUnUsedUnitRefactoring.DoBeforeExecute(out AAbort: Boolean);
begin
  AAbort := FUnUsedUnitRefactoringForm.ShowModal <> mrOk;
end;

function TAbstractUnUsedUnitRefactoring.GetMoveUsedUnitsToImplementation: Boolean;
begin
  Result := FUnUsedUnitRefactoringForm.MoveUsedUnitsCheckBox.Checked;
end;

function TAbstractUnUsedUnitRefactoring.GetRemoveAllUnusedUnits: Boolean;
begin
  Result := FUnUsedUnitRefactoringForm.RemoveAllUnUsedUnitsCheckBox.Checked;
end;

{ TSingleUnUsedUnitRefactoring }

procedure TSingleUnUsedUnitRefactoring.Execute;
var
  AAbort: Boolean;
  ABaseFolderName: string;
begin
  DoBeforeExecute(AAbort);
  if not AAbort then
  begin
    ABaseFolderName := GetBaseFolderName;
    DoRemoveUnusedUnits(gcwRefactoringHost.SelectedUnit, ABaseFolderName);
    ViewRefactoring(ABaseFolderName);
  end;
end;

function TSingleUnUsedUnitRefactoring.GetCanExecute: Boolean;
begin
  Result := Assigned(gcwRefactoringHost.SelectedUnit);
end;

function TSingleUnUsedUnitRefactoring.GetMenuCaption: string;
begin
  Result := 'Selected Unit';
end;

function TSingleUnUsedUnitRefactoring.GetName: string;
begin
  Result := 'SingleUnUsedUnitRefactoring';
end;

{ TMultiUnUsedUnitRefactoring }

procedure TMultiUnUsedUnitRefactoring.Execute;
var
  AAbort: Boolean;
  ABaseFolderName: string;
  Index: Integer;
begin
  DoBeforeExecute(AAbort);
  if not AAbort then
  begin
    ABaseFolderName := GetBaseFolderName;
    // skipping main project unit!
    for Index := 1 to gcwProject.AllUnits.Count - 1 do
      DoRemoveUnusedUnits(gcwProject.AllUnits[Index], ABaseFolderName);
    ViewRefactoring(ABaseFolderName);
  end;
end;

function TMultiUnUsedUnitRefactoring.GetCanExecute: Boolean;
begin
  Result := True;
end;

function TMultiUnUsedUnitRefactoring.GetMenuCaption: string;
begin
  Result := 'All Units';
end;

function TMultiUnUsedUnitRefactoring.GetName: string;
begin
  Result := 'TMultiUnUsedUnitRefactoring';
end;

var
  ASingleUnUsedUnitRefactoring: TSingleUnUsedUnitRefactoring = nil;
  AMultiUnUsedUnitRefactoring: TMultiUnUsedUnitRefactoring = nil;

initialization

  ASingleUnUsedUnitRefactoring := TSingleUnUsedUnitRefactoring.Create;
  gcwRefactoringHost.AddGroupName('RemoveUnUsedUnit', 'Remove Unused Unit');
  gcwRefactoringHost.RegisterRefactoring(ASingleUnUsedUnitRefactoring, 'RemoveUnUsedUnit');
  AMultiUnUsedUnitRefactoring := TMultiUnUsedUnitRefactoring.Create;
  gcwRefactoringHost.RegisterRefactoring(AMultiUnUsedUnitRefactoring, 'RemoveUnUsedUnit');

finalization

  gcwRefactoringHost.UnregisterRefactoring(ASingleUnUsedUnitRefactoring);
  ASingleUnUsedUnitRefactoring.Free;
  gcwRefactoringHost.UnregisterRefactoring(AMultiUnUsedUnitRefactoring);
  AMultiUnUsedUnitRefactoring.Free;

end.
