(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwRefactorNodes;

interface

uses
  SysUtils,
  Classes,
  Contnrs,
  Types,
  mwPasLex,
  mwSimplePasPar,
  mwSimplePasParTypes,
  mwPasLexTypes;

type
  TcwNode = class;
  TcwTokenNode = class;

  TNodeCompareEvent = function(ANode: TcwNode; AData: Pointer): Boolean of object;

  TcwNode = class
  private
    FObjectList: TObjectList;
    FParentNode: TcwNode;
    function GetNode(Index: Integer): TcwNode;
    function GetNextNode(AFromIndex: Integer; ASearch: Boolean;
      ANodeCompareEvent: TNodeCompareEvent; AData: Pointer): TcwNode;
    function GetPriorNode(AFromIndex: Integer; ASearch: Boolean;
      ANodeCompareEvent: TNodeCompareEvent; AData: Pointer): TcwNode;
    function GetNextTokenNodeCompareEvent(ANode: TcwNode; AData: Pointer): Boolean;
    function GetNextNodeClassCompareEvent(ANode: TcwNode; AData: Pointer): Boolean;
  public
    constructor Create(AParentNode: TcwNode = nil; Index: Integer = -1);
    destructor Destroy; override;
    procedure SetParentNode(AParentNode: TcwNode; Index: Integer = -1);
    property ParentNode: TcwNode read FParentNode;
    function Count: Integer;
    function Add(ANode: TcwNode): Integer;
    procedure Insert(Index: Integer; ANode: TcwNode); virtual;
    procedure Delete(ANode: TcwNode); overload; virtual;
    procedure Delete; overload;
    class procedure DeleteIfExists(ANode: TcwNode);
    function ParentedBy(AParentNode: TcwNode): Boolean; overload;
    function ParentedBy(AParentNodeClass: TClass): Boolean; overload;
    function IsParentOf(ANode: TcwNode): Boolean;
    procedure Clear;
    function IndexOf(ANode: TcwNode): Integer;
    function GetIndex: Integer;
    // find token nodes of a certain kind...
    function GetFirstTokenNode(ATokenKind: TptTokenKind = ptUnknown): TcwTokenNode;
    function GetNextTokenNode(AFromIndex: Integer; ATokenKind: TptTokenKind = ptUnknown;
      ASearch: Boolean = True): TcwTokenNode;
    function GetLastTokenNode(ATokenKind: TptTokenKind = ptUnknown): TcwTokenNode;
    function GetPriorTokenNode(AFromIndex: Integer; ATokenKind: TptTokenKind = ptUnknown;
      ASearch: Boolean = True): TcwTokenNode;
    function GetPriorPeerTokenNode(ATokenKind: TptTokenKind = ptUnknown;
      ASearch: Boolean = True): TcwTokenNode;
    function GetNextPeerTokenNode(ATokenKind: TptTokenKind = ptUnknown;
      ASearch: Boolean = True): TcwTokenNode;
    // find nodes of a certain class...
    function GetFirstNodeClass(ANodeClass: TClass): TcwNode;
    function GetNextNodeClass(AFromIndex: Integer; ANodeClass: TClass; ASearch: Boolean = True):
      TcwNode;
    function GetLastNodeClass(ANodeClass: TClass): TcwNode;
    function GetPriorNodeClass(AFromIndex: Integer; ANodeClass: TClass; ASearch: Boolean = True):
      TcwNode;
    function GetPriorPeerNodeClass(ANodeClass: TClass; ASearch: Boolean = True): TcwNode;
    function GetNextPeerNodeClass(ANodeClass: TClass; ASearch: Boolean = True): TcwNode;

    property Nodes[Index: Integer]: TcwNode read GetNode; default;
  end;

  TcwNodeCallbackEvent = procedure(const ANode: TcwNode) of object;

  // this is always the "root" node
  TcwModuleNode = class(TcwNode)
  private
  public
    constructor Create;
    procedure IterateNodes(ADepthFirst: Boolean; ACallback: TcwNodeCallbackEvent);
    procedure Generate(AFileName: string; ACheckOverwrite: Boolean);
  end;

  TcwCodeNode = class(TcwNode)
  private
    FCodeModified: Boolean;
    FStartPos: Integer;
    FEndPos: Integer;
    FCode: string;
    FStartPoint: TPoint;
    FEndPoint: TPoint;
    function GetEndPoint: TPoint;
    function GetPoint: TPoint;
    function GetEndPos: Integer;
    function GetStartPos: Integer;
    procedure SetCode(const Value: string);
    function DoGetCode: string;
    procedure TranslatePoint(var ADestPoint: TPoint; ASourcePoint: TPoint);
  protected
    procedure SetEndPoint(APoint: TPoint; AEndPos: Integer; const ACode: PChar);
  public
    constructor Create(AParentNode: TcwNode; APoint: TPoint; APos: Integer; Index: Integer = -1);
      virtual;
    constructor CreateCode(AParentNode: TcwNode; ACode: string; Index: Integer = -1);
    function GetCode(AIncludeLastJunkNode: Boolean = True): string;
    property Point: TPoint read GetPoint;
    property EndPoint: TPoint read GetEndPoint;
    property StartPos: Integer read GetStartPos;
    property EndPos: Integer read GetEndPos;
    property Code: string read DoGetCode write SetCode;
    property CodeModified: Boolean read FCodeModified;
  end;

  TcwCodeNodeClass = class of TcwCodeNode;

  TcwCodeNodeList = class
  private
    FObjectList: TObjectList;
    function GetNode(Index: Integer): TcwCodeNode;
  public
    constructor Create;
    destructor Destroy; override;
    function Add(ANode: TcwCodeNode): Integer;
    procedure Delete(ANode: TcwCodeNode);
    procedure Clear;
    function Count: Integer;
    property Nodes[Index: Integer]: TcwCodeNode read GetNode; default;
  end;

  TcwInterfaceSectionNode = class(TcwCodeNode)
  end;

  TcwImplementationSectionNode = class(TcwCodeNode)
  end;

  TcwIdentifierListNode = class(TcwCodeNode)
  private
    FNodeList: TcwCodeNodeList;
    procedure RebuildList;
  public
    constructor Create(AParentNode: TcwNode; APoint: TPoint; APos: Integer; Index: Integer = -1);
      override;
    destructor Destroy; override;
    procedure Insert(Index: Integer; ANode: TcwNode); override;
    procedure Delete(ANode: TcwNode); override;
    property IdentifierList: TcwCodeNodeList read FNodeList;
  end;

  TcwUsesListNode = class(TcwIdentifierListNode)
  end;

  TcwCommentNode = class(TcwCodeNode)
  end;

  TcwDirectiveNode = class(TcwCommentNode)
  end;

  TcwJunkNode = class(TcwCodeNode)
  public
    function IsWhiteSpaceOnly: Boolean;
  end;

  TcwTokenNode = class(TcwCodeNode)
  private
    FTokenKind: TptTokenKind;
  protected
    procedure SetTokenKind(ATokenKind: TptTokenKind);
  public
    constructor CreateCode(AParentNode: TcwNode; ACode: string; ATokenKind: TptTokenKind;
      Index: Integer = -1);
    property TokenKind: TptTokenKind read FTokenKind;
  end;

  TcwIdentifierNode = class(TcwTokenNode)
  public
    function IsQualified: Boolean;
  end;

  TcwClassNode = class(TcwCodeNode)
  end;

  TcwClassMemberListNode = class(TcwCodeNode)
  end;

  TcwClassMethodDeclarationNode = class(TcwCodeNode)
  end;

  TcwClassPropertyNode = class(TcwCodeNode)
  end;

  TcwMethodDeclarationNode = class(TcwCodeNode)
  end;

  TcwBlockNode = class;

  TcwMethodImplementationNode = class(TcwCodeNode)
  private
    FBlockNode: TcwBlockNode;
  public
    procedure Insert(Index: Integer; ANode: TcwNode); override;
    procedure Delete(ANode: TcwNode); override;
    property BlockNode: TcwBlockNode read FBlockNode;
  end;

  TcwVariableDeclarationListNode = class;

  TcwStatementListNode = class(TcwCodeNode)
  private
    FNodeList: TcwCodeNodeList;
  public
    constructor Create(AParentNode: TcwNode; APoint: TPoint; APos: Integer;
      Index: Integer = -1); override;
    destructor Destroy; override;
    procedure Insert(Index: Integer; ANode: TcwNode); override;
    procedure Delete(ANode: TcwNode); override;
    property StatementList: TcwCodeNodeList read FNodeList;
  end;

  TcwBlockNode = class(TcwStatementListNode)
  private
    FMethodList: TcwCodeNodeList;
    FVariableSectionList: TcwCodeNodeList;
  public
    constructor Create(AParentNode: TcwNode; APoint: TPoint; APos: Integer; Index: Integer = -1);
      override;
    destructor Destroy; override;
    procedure Insert(Index: Integer; ANode: TcwNode); override;
    procedure Delete(ANode: TcwNode); override;
    function AddVariableSectionCode: TcwVariableDeclarationListNode;
    function GetInnerMostVariableSection: TcwVariableDeclarationListNode;
    property MethodList: TcwCodeNodeList read FMethodList;
    property VariableSectionList: TcwCodeNodeList read FVariableSectionList;
  end;

  TcwStatementNode = class(TcwStatementListNode)
  public
    function GetBlockNode: TcwBlockNode;
    function GetStatementKind: TptTokenKind;
  end;

  TcwVariableDeclarationListNode = class(TcwCodeNode)
  public
    procedure AddVariableCode(AName: string; AType: string);
    function IsInnerMostSection: Boolean;
    function GetBlockNode: TcwBlockNode;
  end;

  TcwConstDeclarationListNode = class(TcwCodeNode)
  end;

  TcwTypeDeclarationListNode = class(TcwCodeNode)
  end;

  TcwParameterListNode = class(TcwCodeNode)
  end;

  TcwVariableReferenceListNode = class(TcwCodeNode)
  private
    FNodeList: TcwCodeNodeList;
  public
    constructor Create(AParentNode: TcwNode; APoint: TPoint; APos: Integer; Index: Integer = -1);
      override;
    destructor Destroy; override;
    procedure Insert(Index: Integer; ANode: TcwNode); override;
    procedure Delete(ANode: TcwNode); override;
    property VariableReferenceList: TcwCodeNodeList read FNodeList;
  end;

  TcwVariableReferenceNode = class(TcwVariableReferenceListNode)
  public
    function GetContainerNode(out AAsOperator: Boolean): TcwIdentifierNode;
  end;

  TcwNodeBuilder = class(TmwSimplePasPar)
  private
    FSkipJunk: Boolean;
    FFileName: string;
    FStack: TObjectStack;
    FCurrentNode: TcwNode;
    FModuleNode: TcwModuleNode;
    procedure BeginCodeNode(ACodeNodeClass: TcwCodeNodeClass);
    procedure EndCodeNode;
    //procedure CancelCodeNode;
  protected
    procedure ParseFile; override;
    procedure SkipJunk; override;
    procedure SkipDirective; override;
    procedure NextToken; override;
    procedure InterfaceSection; override;
    procedure ImplementationSection; override;
    procedure UsedUnitsList; override;
    procedure Block; override;
    procedure Statement; override;
    procedure ClassType; override;
    procedure ProcedureMethodDeclaration; override;
    procedure FunctionMethodDeclaration; override;
    procedure VarSection; override;
    procedure FormalParameterList; override;
    procedure Expected(Sym: TptTokenKind); override;
    procedure UnitFile; override;
    procedure ClassProperty; override;
    procedure ConstSection; override;
    procedure TypeSection; override;
    procedure ClassMemberList; override;
    procedure ClassMethodHeading; override;
    procedure ExportedHeading; override;
    procedure VariableReference; override;
  public
    constructor Create(AModuleNode: TcwModuleNode);
    destructor Destroy; override;
    procedure Run(UnitName: String; SourceStream: TMemoryStream = nil); override;
  end;

implementation

{ TcwNode }

uses
  Dialogs,
  Math,
  ucwClasses;

constructor TcwNode.Create(AParentNode: TcwNode = nil; Index: Integer = -1);
var
  AIndex: Integer;
begin
  inherited Create;
  FObjectList := TObjectList.Create(False);
  FParentNode := AParentNode;
  if Assigned(FParentNode) then
  begin
    if Index >= 0 then
      AIndex := Index
    else
      AIndex := FParentNode.Count;
    FParentNode.Insert(AIndex, Self)
  end;
end;

destructor TcwNode.Destroy;
begin
  Clear;
  FObjectList.Free;
  if Assigned(FParentNode) then
    FParentNode.Delete(Self);
  inherited;
end;

procedure TcwNode.Clear;
begin
  while Count > 0 do
    Nodes[0].Free;
end;

function TcwNode.GetNode(Index: Integer): TcwNode;
begin
  Result := FObjectList[Index] as TcwNode;
end;

function TcwNode.Add(ANode: TcwNode): Integer;
begin
  FObjectList.Insert(FObjectList.Count, ANode);
  Result := FObjectList.Count - 1;
end;

function TcwNode.Count: Integer;
begin
  Result := FObjectList.Count;
end;

procedure TcwNode.Delete(ANode: TcwNode);
begin
  FObjectList.Delete(FObjectList.IndexOf(ANode));
end;

procedure TcwNode.Insert(Index: Integer; ANode: TcwNode);
begin
  FObjectList.Insert(Index, ANode);
end;

function TcwNode.ParentedBy(AParentNode: TcwNode): Boolean;
var
  ATempNode: TcwNode;
begin
  ATempNode := Self;
  repeat
    Result := ATempNode.ParentNode = AParentNode;
    if Result then
      Break
    else
      ATempNode := ATempNode.ParentNode;
  until not Assigned(ATempNode);
end;

function TcwNode.ParentedBy(AParentNodeClass: TClass): Boolean;
var
  ATempNode: TcwNode;
begin
  ATempNode := Self;
  repeat
    Result := Assigned(ATempNode.ParentNode) and
      (ATempNode.ParentNode.ClassType = AParentNodeClass);
    if Result then
      Break
    else
      ATempNode := ATempNode.ParentNode;
  until not Assigned(ATempNode);
end;

function TcwNode.IsParentOf(ANode: TcwNode): Boolean;

  function RecurseNodes(ANodeToFind, ANode: TcwNode): Boolean;
  var
    Index: Integer;
  begin
    Result := ANodeToFind = ANode;
    if not Result then
      for Index := 0 to ANode.Count - 1 do
      begin
        Result := RecurseNodes(ANodeToFind, ANode[Index]);
        if Result then
          Break;
      end;
  end;

begin
  if ANode <> Self then
    Result := RecurseNodes(ANode, Self)
  else
    Result := False;
end;

procedure TcwNode.SetParentNode(AParentNode: TcwNode; Index: Integer = -1);
var
  AIndex: Integer;
begin
  Assert((Self <> AParentNode) and not IsParentOf(AParentNode));
  if Assigned(ParentNode) then
    ParentNode.Delete(Self);
  if Index = -1 then
    AIndex := AParentNode.Count - 1
  else
    AIndex := Index;
  AParentNode.Insert(AIndex, Self);
  FParentNode := AParentNode;
end;

function TcwNode.GetIndex: Integer;
begin
  if Assigned(ParentNode) then
    Result := ParentNode.IndexOf(Self)
  else
    Result := -1;
end;

function TcwNode.IndexOf(ANode: TcwNode): Integer;
begin
  Result := FObjectList.IndexOf(ANode);
end;

function TcwNode.GetNextNode(AFromIndex: Integer; ASearch: Boolean;
  ANodeCompareEvent: TNodeCompareEvent; AData: Pointer): TcwNode;
var
  Index: Integer;
begin
  Result := nil;
  for Index := AFromIndex + 1 to Count - 1 do
  begin
    if ANodeCompareEvent(Nodes[Index], AData) then
    begin
      Result := Nodes[Index];
      Break;
    end;
    if not ASearch then
      Break;
  end;
end;

function TcwNode.GetPriorNode(AFromIndex: Integer; ASearch: Boolean;
  ANodeCompareEvent: TNodeCompareEvent; AData: Pointer): TcwNode;
var
  Index: Integer;
begin
  Result := nil;
  for Index := AFromIndex - 1 downto 0 do
  begin
    if ANodeCompareEvent(Nodes[Index], AData) then
    begin
      Result := Nodes[Index];
      Break;
    end;
    if not ASearch then
      Break;
  end;
end;

function TcwNode.GetNextTokenNodeCompareEvent(ANode: TcwNode; AData: Pointer): Boolean;
var
  ATokenKind: TptTokenKind;
  ATokenNode: TcwTokenNode;
begin
  Result := False;
  ATokenKind := TptTokenKind(Integer(AData));
  if (ANode is TcwTokenNode) then
  begin
    ATokenNode := ANode as TcwTokenNode;
    if ((ATokenKind = ptUnknown) or (ATokenNode.TokenKind = ATokenKind)) then
      Result := True;
  end;
end;

function TcwNode.GetFirstTokenNode(ATokenKind: TptTokenKind = ptUnknown): TcwTokenNode;
begin
  Result := GetNextTokenNode(-1, ATokenKind);
end;

function TcwNode.GetNextTokenNode(AFromIndex: Integer; ATokenKind: TptTokenKind = ptUnknown;
  ASearch: Boolean = True): TcwTokenNode;
begin
  Result := GetNextNode(AFromIndex, ASearch, GetNextTokenNodeCompareEvent,
    Pointer(Integer(ATokenKind))) as TcwTokenNode;
end;

function TcwNode.GetPriorTokenNode(AFromIndex: Integer; ATokenKind: TptTokenKind = ptUnknown;
  ASearch: Boolean = True): TcwTokenNode;
begin
  Result := GetPriorNode(AFromIndex, ASearch, GetNextTokenNodeCompareEvent,
    Pointer(Integer(ATokenKind))) as TcwTokenNode;
end;

function TcwNode.GetLastTokenNode(ATokenKind: TptTokenKind = ptUnknown): TcwTokenNode;
begin
  Result := GetPriorTokenNode(Count, ATokenKind);
end;

function TcwNode.GetNextPeerTokenNode(ATokenKind: TptTokenKind = ptUnknown;
  ASearch: Boolean = True): TcwTokenNode;
begin
  if Assigned(ParentNode) then
    Result := ParentNode.GetNextTokenNode(GetIndex, ATokenKind, ASearch)
  else
    Result := nil;
end;

function TcwNode.GetPriorPeerTokenNode(ATokenKind: TptTokenKind = ptUnknown;
  ASearch: Boolean = True): TcwTokenNode;
begin
  if Assigned(ParentNode) then
    Result := ParentNode.GetPriorTokenNode(GetIndex, ATokenKind, ASearch)
  else
    Result := nil;
end;

function TcwNode.GetNextNodeClassCompareEvent(ANode: TcwNode; AData: Pointer): Boolean;
begin
  Result := ANode.ClassType = TClass(AData);
end;

function TcwNode.GetFirstNodeClass(ANodeClass: TClass): TcwNode;
begin
  Result := GetNextNodeClass(-1, ANodeClass);
end;

function TcwNode.GetNextNodeClass(AFromIndex: Integer; ANodeClass: TClass; ASearch: Boolean = True):
  TcwNode;
begin
  Result := GetNextNode(AFromIndex, ASearch, GetNextNodeClassCompareEvent,
    Pointer(ANodeClass)) as TcwNode;
end;

function TcwNode.GetPriorNodeClass(AFromIndex: Integer; ANodeClass: TClass;
  ASearch: Boolean = True): TcwNode;
begin
  Result := GetPriorNode(AFromIndex, ASearch, GetNextNodeClassCompareEvent,
    Pointer(ANodeClass)) as TcwNode;
end;

function TcwNode.GetLastNodeClass(ANodeClass: TClass): TcwNode;
begin
  Result := GetPriorNodeClass(Count, ANodeClass);
end;

function TcwNode.GetNextPeerNodeClass(ANodeClass: TClass; ASearch: Boolean = True): TcwNode;
begin
  if Assigned(ParentNode) then
    Result := ParentNode.GetNextNodeClass(GetIndex, ANodeClass, ASearch)
  else
    Result := nil;
end;

function TcwNode.GetPriorPeerNodeClass(ANodeClass: TClass; ASearch: Boolean = True): TcwNode;
begin
  if Assigned(ParentNode) then
    Result := ParentNode.GetPriorNodeClass(GetIndex, ANodeClass, ASearch)
  else
    Result := nil;
end;

procedure TcwNode.Delete;
begin
  ParentNode.Delete(Self);
  // Self is now invalid!
end;

class procedure TcwNode.DeleteIfExists(ANode: TcwNode);
begin
  if Assigned(ANode) then
    ANode.Delete;
end;

{ TcwModuleNode }

constructor TcwModuleNode.Create;
begin
  inherited Create(nil);
end;

procedure TcwModuleNode.IterateNodes(ADepthFirst: Boolean; ACallback: TcwNodeCallbackEvent);

  procedure ProcessNode(ANode: TcwNode);

    procedure ProcessChildNodes(ANode: TcwNode);
    var
      ASaveCount: Integer;
      Index: Integer;
    begin
      Index := 0;
      while Index <= ANode.Count - 1 do
      begin
        ASaveCount := ANode.Count;
        // RECURSE!
        ProcessNode(ANode[Index]);
        // if nodes were added in the event, then 'jump' over those nodes...
        Inc(Index, ANode.Count - ASaveCount + 1)
      end;
    end;

  begin
    if ADepthFirst then
      ProcessChildNodes(ANode);
    ACallback(ANode);
    if not ADepthFirst then
      ProcessChildNodes(ANode);
  end;

begin
  ProcessNode(Self);
end;

procedure TcwModuleNode.Generate(AFileName: string; ACheckOverwrite: Boolean);

  procedure ProcessNode(ANode: TcwNode; var ACode: string);
  var
    Index: Integer;
    ACodeNode: TcwCodeNode;
    ASetCode: Boolean;
  begin
    ASetCode := False;
    if ANode is TcwCodeNode then
    begin
      ACodeNode := ANode as TcwCodeNode;
      if (ANode.Count = 0) or ACodeNode.CodeModified then
      begin
        ACode := ACode + ACodeNode.Code;
        ASetCode := True;
      end;
    end;
    if not ASetCode then
      for Index := 0 to ANode.Count - 1 do
        ProcessNode(ANode.Nodes[Index], ACode);
  end;

var
  AStringList: TStringList;
  ACode: string;
begin
  if ACheckOverwrite and FileExists(AFileName) then
    raise Exception.CreateFmt('Attempt to overwrite file, %s.', [AFileName]);
  AStringList := TStringList.Create;
  try
    ProcessNode(Self, ACode);
    AStringList.Text := ACode;
    AStringList.SaveToFile(AFileName);
  finally
    AStringList.Free;
  end;
end;

{ TcwCodeNode }

constructor TcwCodeNode.Create(AParentNode: TcwNode; APoint: TPoint; APos: Integer;
  Index: Integer = -1);
begin
  inherited Create(AParentNode, Index);
  TranslatePoint(FStartPoint, APoint);
  FStartPos := APos;
end;

constructor TcwCodeNode.CreateCode(AParentNode: TcwNode; ACode: string;
  Index: Integer = -1);
begin
  Create(AParentNode, Types.Point(-1, -1), -1, Index);
  SetCode(ACode);
end;

function TcwCodeNode.DoGetCode: string;
begin
  Result := GetCode(True);
end;

function TcwCodeNode.GetCode(AIncludeLastJunkNode: Boolean = True): string;
var
  Index: Integer;
begin
  // get code from child nodes, unless Code has been set explicitly...
  if (Count = 0) or FCodeModified then
    Result := FCode
  else
  begin
    if Count > 0 then
    begin
      Result := '';
      for Index := 0 to Count - 2 do
        if Nodes[Index] is TcwCodeNode then
          Result := Result + (Nodes[Index] as TcwCodeNode).Code;
      if not (Nodes[Count - 1] is TcwJunkNode) or AIncludeLastJunkNode then
        Result := Result + (Nodes[Count - 1] as TcwCodeNode).Code;
    end else
      Result := FCode;
  end;
end;

function TcwCodeNode.GetEndPoint: TPoint;
begin
  Result := FEndPoint;
end;

function TcwCodeNode.GetEndPos: Integer;
begin
  Result := FEndPos;
end;

function TcwCodeNode.GetPoint: TPoint;
begin
  Result := FStartPoint;
end;

function TcwCodeNode.GetStartPos: Integer;
begin
  Result := FStartPos;
end;

procedure TcwCodeNode.SetCode(const Value: string);
begin
  FCode := Value;
  FCodeModified := True;
end;

procedure TcwCodeNode.TranslatePoint(var ADestPoint: TPoint; ASourcePoint: TPoint);
begin
  ADestPoint.X := ASourcePoint.X + 1;
  ADestPoint.Y := ASourcePoint.Y + 1;
end;

procedure TcwCodeNode.SetEndPoint(APoint: TPoint; AEndPos: Integer; const ACode: PChar);
var
  ALen: Integer;
begin
  TranslatePoint(FEndPoint, APoint);
  FEndPos := AEndPos;
  if (FStartPos >= 0) and (Count = 0) then
  begin
    //Assert(FStartPos <= Length(ACode) - 1, IntToStr(FStartPos));
    //Assert((AEndPos >= 0) and (AEndPos <= Length(ACode) - 1), IntToStr(AEndPos));
    //Assert(AEndPos >= FStartPos);
    ALen := AEndPos - FStartPos + 1;
    if ALen > 0 then
      SetString(FCode, PChar(@ACode[FStartPos]), ALen);
  end;
end;

{ TcwCodeNodeList }

constructor TcwCodeNodeList.Create;
begin
  inherited;
  FObjectList := TObjectList.Create(False);
end;

destructor TcwCodeNodeList.Destroy;
begin
  FObjectList.Free;
  inherited;
end;

function TcwCodeNodeList.Add(ANode: TcwCodeNode): Integer;
begin
  Result := FObjectList.Add(ANode);
end;

function TcwCodeNodeList.Count: Integer;
begin
  Result := FObjectList.Count;
end;

procedure TcwCodeNodeList.Delete(ANode: TcwCodeNode);
begin
  FObjectList.Delete(FObjectList.IndexOf(ANode));
end;

function TcwCodeNodeList.GetNode(Index: Integer): TcwCodeNode;
begin
  Result := FObjectList[Index] as TcwCodeNode;
end;

procedure TcwCodeNodeList.Clear;
begin
  FObjectList.Clear;
end;

{ TcwIdentifierListNode }

constructor TcwIdentifierListNode.Create(AParentNode: TcwNode; APoint: TPoint; APos: Integer;
  Index: Integer = -1);
begin
  inherited;
  FNodeList := TcwCodeNodeList.Create;
end;

destructor TcwIdentifierListNode.Destroy;
begin
  inherited;
  FreeAndNil(FNodeList);
end;

procedure TcwIdentifierListNode.Insert(Index: Integer; ANode: TcwNode);
begin
  inherited;
  if ANode is TcwIdentifierNode then
    // can't simply add to FNodeList since ANode is being inserted, possibly at beginning of list
    RebuildList;
end;

procedure TcwIdentifierListNode.Delete(ANode: TcwNode);
begin
  inherited;
  Assert(Assigned(FNodeList));
  if ANode is TcwIdentifierNode then
    RebuildList;
end;

procedure TcwIdentifierListNode.RebuildList;
var
  Index: Integer;
begin
  FNodeList.Clear;
  for Index := 0 to Count - 1 do
    if Nodes[Index] is TcwIdentifierNode then
      FNodeList.Add(Nodes[Index] as TcwIdentifierNode);
end;

{ TcwJunkNode }

function TcwJunkNode.IsWhiteSpaceOnly: Boolean;
var
  TempStr: string;
begin
  TempStr := StringReplace(Code, #13, '', [rfReplaceAll]);
  TempStr := StringReplace(TempStr, #10, '', [rfReplaceAll]);
  TempStr := StringReplace(TempStr, #9, '', [rfReplaceAll]);
  TempStr := StringReplace(TempStr, ' ', '', [rfReplaceAll]);
  Result := TempStr = '';
end;

{ TcwTokenNode }

constructor TcwTokenNode.CreateCode(AParentNode: TcwNode; ACode: string; ATokenKind: TptTokenKind;
  Index: Integer = -1);
begin
  inherited CreateCode(AParentNode, ACode, Index);
  SetTokenKind(ATokenKind);
end;

procedure TcwTokenNode.SetTokenKind(ATokenKind: TptTokenKind);
begin
  FTokenKind := ATokenKind;
end;

{ TcwIdentifierNode }

function TcwIdentifierNode.IsQualified: Boolean;
var
  APriorTokenNode: TcwTokenNode;
  AVariableReferenceNode: TcwVariableReferenceNode;
begin
  Assert(Assigned(ParentNode));
  if (GetIndex = 0) and (ParentNode is TcwVariableReferenceNode) then
  begin
    AVariableReferenceNode := ParentNode as TcwVariableReferenceNode;
    Assert(Assigned(AVariableReferenceNode.ParentNode));
    APriorTokenNode := AVariableReferenceNode.GetPriorPeerTokenNode;
  end else
    APriorTokenNode := GetPriorPeerTokenNode;
  Result := Assigned(APriorTokenNode) and (APriorTokenNode.TokenKind = ptPoint);
end;

{ TcwMethodImplementationNode }

procedure TcwMethodImplementationNode.Insert(Index: Integer; ANode: TcwNode);
begin
  inherited;
  if ANode is TcwBlockNode then
    FBlockNode := ANode as TcwBlockNode;
end;

procedure TcwMethodImplementationNode.Delete(ANode: TcwNode);
begin
  inherited;
  if ANode is TcwBlockNode then
    FBlockNode := nil;
end;

{ TcwStatementListNode }

constructor TcwStatementListNode.Create(AParentNode: TcwNode; APoint: TPoint; APos: Integer;
  Index: Integer = -1);
begin
  inherited;
  FNodeList := TcwCodeNodeList.Create;
end;

destructor TcwStatementListNode.Destroy;
begin
  inherited;
  FreeAndNil(FNodeList);
end;

procedure TcwStatementListNode.Insert(Index: Integer; ANode: TcwNode);
begin
  inherited;
  if ANode is TcwStatementNode then
    FNodeList.Add(ANode as TcwStatementNode);
end;

procedure TcwStatementListNode.Delete(ANode: TcwNode);
begin
  inherited;
  if ANode is TcwStatementNode then
    FNodeList.Delete(ANode as TcwStatementNode);
end;

{ TcwBlockNode }

constructor TcwBlockNode.Create(AParentNode: TcwNode; APoint: TPoint;
  APos: Integer; Index: Integer = -1);
begin
  inherited;
  FMethodList := TcwCodeNodeList.Create;
  FVariableSectionList := TcwCodeNodeList.Create;
end;

destructor TcwBlockNode.Destroy;
begin
  inherited;
  FreeAndNil(FMethodList);
  FreeAndNil(FVariableSectionList);
end;

procedure TcwBlockNode.Insert(Index: Integer; ANode: TcwNode);
begin
  inherited;
  if ANode is TcwMethodImplementationNode then
    FMethodList.Add(ANode as TcwMethodImplementationNode);
  if ANode is TcwVariableDeclarationListNode then
    FVariableSectionList.Add(ANode as TcwVariableDeclarationListNode);
end;

procedure TcwBlockNode.Delete(ANode: TcwNode);
begin
  inherited;
  if ANode is TcwMethodImplementationNode then
    FMethodList.Delete(ANode as TcwMethodImplementationNode);
  if ANode is TcwVariableDeclarationListNode then
    FVariableSectionList.Delete(ANode as TcwVariableDeclarationListNode);
end;

function TcwBlockNode.AddVariableSectionCode: TcwVariableDeclarationListNode;
var
  ABeginNode: TcwTokenNode;
begin
  ABeginNode := GetFirstTokenNode;
  Assert(ABeginNode.TokenKind = ptBegin);
  // insert at the beginnode, which is after any sub-procedures
  Result := TcwVariableDeclarationListNode.Create(Self,
    Types.Point(-1, -1), -1, ABeginNode.GetIndex);
  TcwTokenNode.CreateCode(Result, 'var', ptVar, 0);
  TcwJunkNode.CreateCode(Result, #13#10'  ', 1);
end;

function TcwBlockNode.GetInnerMostVariableSection: TcwVariableDeclarationListNode;
var
  Index: Integer;
begin
  Result := nil;
  for Index := 0 to FVariableSectionList.Count - 1 do
    if (FVariableSectionList[Index] as TcwVariableDeclarationListNode).IsInnerMostSection then
    begin
      Result := FVariableSectionList[Index] as TcwVariableDeclarationListNode;
      Break;
    end;
end;

{ TcwStatementNode }

function TcwStatementNode.GetBlockNode: TcwBlockNode;
var
  ANode: TcwNode;
begin
  Result := nil;
  ANode := ParentNode;
  while Assigned(ANode) do
  begin
    if ANode is TcwBlockNode then
    begin
      Result := ANode as TcwBlockNode;
      Break;
    end;
    ANode := ANode.ParentNode;
  end;
end;

function TcwStatementNode.GetStatementKind: TptTokenKind;
var
  ATokenNode: TcwTokenNode;
begin
  ATokenNode := GetFirstTokenNode;
  if Assigned(ATokenNode) then
    Result := ATokenNode.TokenKind
  else
    Result := ptUnknown; //TODO: what if assignment statement? should be ptAssign
end;

{ TcwVariableDeclarationListNode }

procedure TcwVariableDeclarationListNode.AddVariableCode(AName: string; AType: string);
var
  AOriginalHadVars: Boolean;
  AJunkCode: string;
begin
  Assert(Count >= 2);
  Assert((Nodes[0] as TcwTokenNode).TokenKind = ptVar);
  Assert(Nodes[1] is TcwJunkNode);
  // not asserting this because of TcwBlockNode.AddVariableSectionCode...
  //Assert(Nodes[2] is TcwIdentifierNode);
  AOriginalHadVars := (Count > 2) and (Nodes[2] is TcwIdentifierNode);
  TcwIdentifierNode.CreateCode(Self, AName, ptIdentifier, 2);
  TcwTokenNode.CreateCode(Self, ':', ptColon, 3);
  TcwJunkNode.CreateCode(Self, ' ', 4);
  TcwIdentifierNode.CreateCode(Self, AType, ptIdentifier, 5);
  TcwTokenNode.CreateCode(Self, ';', ptSemiColon, 6);
  if AOriginalHadVars then
    // copy junk after the 'var' node...
    AJunkCode := (Nodes[1] as TcwJunkNode).Code
  else
    AJunkCode := sLineBreak;
  TcwJunkNode.CreateCode(Self, AJunkCode, 7);
end;

function TcwVariableDeclarationListNode.GetBlockNode: TcwBlockNode;
var
  ANode: TcwNode;
begin
  Result := nil;
  ANode := ParentNode;
  while Assigned(ANode) do
  begin
    if ANode is TcwBlockNode then
    begin
      Result := ANode as TcwBlockNode;
      Break;
    end;
    ANode := ANode.ParentNode;
  end;
end;

function TcwVariableDeclarationListNode.IsInnerMostSection: Boolean;
var
  ATokenNode: TcwTokenNode;
  ABlockNode: TcwBlockNode;
begin
  Result := False;
  ABlockNode := GetBlockNode;
  if Assigned(ABlockNode) then
  begin
    ATokenNode := ABlockNode.GetFirstTokenNode;
    if Assigned(ATokenNode) and (ATokenNode.TokenKind = ptBegin) then
      if ABlockNode[ATokenNode.GetIndex - 1] = Self then
        Result := True;
  end;
end;

{ TcwVariableReferenceListNode }

constructor TcwVariableReferenceListNode.Create(AParentNode: TcwNode; APoint: TPoint; APos: Integer;
  Index: Integer = -1);
begin
  inherited;
  FNodeList := TcwCodeNodeList.Create;
end;

destructor TcwVariableReferenceListNode.Destroy;
begin
  inherited;
  FreeAndNil(FNodeList);
end;

procedure TcwVariableReferenceListNode.Insert(Index: Integer; ANode: TcwNode);
begin
  inherited;
  if ANode is TcwVariableReferenceNode then
    FNodeList.Add(ANode as TcwVariableReferenceNode);
end;

procedure TcwVariableReferenceListNode.Delete(ANode: TcwNode);
begin
  inherited;
  Assert(Assigned(FNodeList));
  if ANode is TcwVariableReferenceNode then
    FNodeList.Delete(ANode as TcwVariableReferenceNode);
end;

{ TcwVariableReferenceNode }

function TcwVariableReferenceNode.GetContainerNode(out AAsOperator: Boolean): TcwIdentifierNode;
var
  ATokenNode: TcwTokenNode;
  ANextAsOperator: boolean;
  AIndex: Integer;
  AStack: TStack;
  ALastIdentifier: TcwIdentifierNode;
begin
  // locate the container in this node's expression
  // for example, find TMyObject in "Sender as TMyObject"
  // or TMyObject in "TMyObject(Sender)"
  // or MyObject in "SomeObject[0].MyObject"
  AStack := TStack.Create;
  try
    AIndex := 0;
    Result := nil;
    repeat
      Assert(AIndex <= Count - 1);
      if Nodes[AIndex] is TcwJunkNode then
        Inc(AIndex);
      if AIndex <= Count - 1 then
      begin
        if (Nodes[AIndex] is TcwVariableReferenceNode) then
        begin
          if not Assigned(Result) then
            // RECURSE!
            Result := (Nodes[AIndex] as TcwVariableReferenceNode).GetContainerNode(ANextAsOperator);
        end else
        begin
          if Nodes[AIndex] is TcwIdentifierNode then
            Result := Nodes[AIndex] as TcwIdentifierNode
          else
          begin
            Assert(Nodes[AIndex] is TcwTokenNode);
            ATokenNode := Nodes[AIndex] as TcwTokenNode;
            case ATokenNode.TokenKind of
              ptSquareOpen, ptRoundOpen: AStack.Push(Result);
              ptSquareClose, ptRoundClose: begin
                ALastIdentifier := AStack.Pop;
                if Assigned(ALastIdentifier) then
                  Result := ALastIdentifier;
              end;
              ptAs,
              ptPoint: begin
                Result := nil;
                if ATokenNode.TokenKind = ptAs then
                  AAsOperator := True;
              end;
              else
            end;
          end;
        end;
        Inc(AIndex);
      end;
    until AIndex > Count - 1;
  finally
    AStack.Free;
  end;
end;

{ TcwNodeBuilder }

constructor TcwNodeBuilder.Create(AModuleNode: TcwModuleNode);
var
  ADefines: TmwAssignableHash;
begin
  inherited Create(GlobalKeyHashTable, GlobalDirectiveHash);
  ADefines := TmwAssignAbleHash.Create(gcwProject.CWOptions.Defines.Text);
  try
    Lexer.DefinedList.Assign(ADefines);
  finally
    ADefines.Free;
  end;
  FModuleNode := AModuleNode;
  FStack := TObjectStack.Create;
end;

destructor TcwNodeBuilder.Destroy;
begin
  FStack.Free;
  inherited;
end;

procedure TcwNodeBuilder.ParseFile;
var
  AJunkNode: TcwJunkNode;
begin
  inherited;
  if Lexer.TokenPos - 1 < Length(Lexer.Origin) - 1 then
  begin
    AJunkNode := TcwJunkNode.Create(FModuleNode, Lexer.PosXY, Lexer.TokenPos);
    AJunkNode.SetEndPoint(Types.Point(0, 0), Length(Lexer.Origin) - 1, Lexer.Origin);
  end;
end;

procedure TcwNodeBuilder.BeginCodeNode(ACodeNodeClass: TcwCodeNodeClass);
begin
  Assert(Assigned(FCurrentNode));
  FStack.Push(FCurrentNode);
  FCurrentNode := ACodeNodeClass.Create(FCurrentNode, Lexer.PosXY, Lexer.TokenPos);
  FStack.Push(FCurrentNode);
end;

procedure TcwNodeBuilder.EndCodeNode;
var
  ANode: TcwCodeNode;
begin
  ANode := FStack.Pop as TcwCodeNode;
  ANode.SetEndPoint(Lexer.PosXY, Lexer.TokenPos - 1, Lexer.Origin);
  FCurrentNode := FStack.Pop as TcwNode;
end;

//procedure TcwNodeBuilder.CancelCodeNode;
//var
//  ANode: TcwCodeNode;
//begin
//  ANode := FStack.Pop as TcwCodeNode;
//  ANode.Free;
//  FCurrentNode := FStack.Pop as TcwNode;
//end;

procedure TcwNodeBuilder.SkipDirective;
begin
  BeginCodeNode(TcwDirectiveNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.SkipJunk;

  procedure SkipExpected(ATokenKind: TptTokenKind);
  begin
    if Lexer.TokenID = ATokenKind then
      Lexer.Next
    else
      inherited Expected(ATokenKind);
  end;

begin
  if Lexer.IsJunk then
  begin
    // TokenID is ptUnknown, ptBorDirective, ptAnsiDirective, ptSpace, ptCRLF, ptAnsiComment, ptBorComment, ptCRLFCo, ptSlashesComment
    repeat
      if Lexer.TokenID in [ptAnsiDirective, ptBorDirective] then
        SkipDirective
      else
        if Lexer.TokenID in [ptAnsiComment, ptBorComment, ptSlashesComment] then
        begin
          BeginCodeNode(TcwCommentNode);
          try
            case Lexer.TokenID of
              ptAnsiComment: SkipExpected(ptAnsiComment);
              ptBorComment: SkipExpected(ptBorComment);
              ptSlashesComment: SkipExpected(ptSlashesComment);
              else
                Assert(False);
            end;
          finally
            EndCodeNode;
          end;
        end else
        begin
          FSkipJunk := True;
          BeginCodeNode(TcwJunkNode);
          try
            repeat
              case Lexer.TokenID of
                ptSpace: SkipExpected(ptSpace);
                ptCRLFCo: SkipExpected(ptCRLFCo);
                ptCRLF: SkipExpected(ptCRLF);
                else
                  Assert(False);
              end;
            until not (Lexer.TokenID in [ptSpace, ptCRLFCo, ptCRLF]);
          finally
            FSkipJunk := False;
            EndCodeNode;
          end;
        end;
    until not Lexer.IsJunk;
  end;
end;

procedure TcwNodeBuilder.NextToken;
begin
  Assert(not FSkipJunk);
  BeginCodeNode(TcwTokenNode);
  (FCurrentNode as TcwTokenNode).SetTokenKind(Lexer.TokenID);
  try
    Lexer.Next;
  finally
    EndCodeNode;
  end;
  SkipJunk;
end;

procedure TcwNodeBuilder.ImplementationSection;
begin
  BeginCodeNode(TcwImplementationSectionNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.InterfaceSection;
begin
  BeginCodeNode(TcwInterfaceSectionNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.UsedUnitsList;
begin
  BeginCodeNode(TcwUsesListNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.Block;
begin
  BeginCodeNode(TcwBlockNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.Statement;
begin
  BeginCodeNode(TcwStatementNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.ClassType;
begin
  BeginCodeNode(TcwClassNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.ProcedureMethodDeclaration;
begin
  BeginCodeNode(TcwMethodImplementationNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.FunctionMethodDeclaration;
begin
  BeginCodeNode(TcwMethodImplementationNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.VarSection;
begin
  BeginCodeNode(TcwVariableDeclarationListNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.FormalParameterList;
begin
  BeginCodeNode(TcwParameterListNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.Expected(Sym: TptTokenKind);
var
  ATokenNodeClass: TcwCodeNodeClass;
begin
  Assert(not FSkipJunk);
  if Sym = Lexer.TokenID then
  begin
    if Sym = ptIdentifier then
      ATokenNodeClass := TcwIdentifierNode
    else
      ATokenNodeClass := TcwTokenNode;
    BeginCodeNode(ATokenNodeClass);
    (FCurrentNode as TcwTokenNode).SetTokenKind(Sym);
    try
      Lexer.Next;
    finally
      EndCodeNode;
    end;
    SkipJunk;
  end else
    inherited;
end;

procedure TcwNodeBuilder.Run(UnitName: String; SourceStream: TMemoryStream);
begin
  FFileName := UnitName;
  FCurrentNode := FModuleNode;
  inherited;
end;

procedure TcwNodeBuilder.UnitFile;
begin
  inherited;
  SkipJunk;
end;

procedure TcwNodeBuilder.ClassProperty;
begin
  BeginCodeNode(TcwClassPropertyNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.ConstSection;
begin
  BeginCodeNode(TcwConstDeclarationListNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.TypeSection;
begin
  BeginCodeNode(TcwTypeDeclarationListNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.ClassMemberList;
begin
  BeginCodeNode(TcwClassMemberListNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.ClassMethodHeading;
begin
  BeginCodeNode(TcwClassMethodDeclarationNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.ExportedHeading;
begin
  BeginCodeNode(TcwMethodDeclarationNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

procedure TcwNodeBuilder.VariableReference;
begin
  BeginCodeNode(TcwVariableReferenceNode);
  try
    inherited;
  finally
    EndCodeNode;
  end;
end;

end.
