(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwRefactoring;

interface

uses
  Classes,
  Contnrs,
  ucwClasses,
  ucwRefactorNodes;

type
  TcwRefactoring = class;

  TcwRefactoringListNode = class
  private
    FUnit: TcwUnit;
    FModuleNode: TcwModuleNode;
  public
    constructor Create(ARefactoring: TcwRefactoring; const AUnit: TcwUnit);
    destructor Destroy; override;
    property ModuleNode: TcwModuleNode read FModuleNode;
    property _Unit: TcwUnit read FUnit;
  end;

  TcwRefactoringList = class
  private
    FRefactoring: TcwRefactoring;
    FObjectList: TObjectList;
    function GetNode(Index: Integer): TcwRefactoringListNode;
  public
    constructor Create(ARefactoring: TcwRefactoring);
    function Add(ARefactoringListNode: TcwRefactoringListNode): Integer;
    procedure Delete(Index: Integer);
    destructor Destroy; override;
    function Count: Integer;
    function IndexOf(AUnit: TcwUnit): Integer;
    property Nodes[Index: Integer]: TcwRefactoringListNode read GetNode;
  end;

  TcwRefactoring = class(TInterfacedObject, IcwRefactoring)
  private
    FRefactoringList: TcwRefactoringList;
    FRefactorInPlace: Boolean;
    function GetModuleNode(AUnit: TcwUnit): TcwModuleNode;
    function GetLocked(AUnit: TcwUnit): Boolean;
    function CheckRefactoringUnit(AUnit: TcwUnit): Integer;
  protected
    function TryCreateModuleNode(const AUnit: TcwUnit): TcwModuleNode;
    function FindNodeForLocation(AStartNode: TcwNode; const ALocation: TcwCodeLocation):
      TcwCodeNode;
    function FindCodeLocationForNode(const AUnit: TcwUnit; ANode: TcwCodeNode): TcwCodeLocation;
    function GetFilenameToGenerate(const AFileName: string; const ABaseFolderName: string;
      ACreateDirs: Boolean; out APath: string): string; virtual;
    function GetBaseFolderName: string; virtual;
    function GetQualifiedBaseFolderName(ABaseFolderName: string): string; virtual;

    // IcwRefactoring
    function GetGroupName: string; virtual;
    function GetMenuCaption: string; virtual; abstract;
    function GetName: string; virtual; abstract;
    function GetCanExecute: Boolean; virtual; abstract;
    procedure Execute; virtual; abstract;

    procedure ViewRefactoring(ABaseFolderName: string);
    procedure BeginRefactoring(ABaseFolderName: string; const AUnit: TcwUnit);
    procedure EndRefactoring(ABaseFolderName: string; const AUnit: TcwUnit);
    procedure EndRefactorings(ABaseFolderName: string);
    procedure CancelRefactoring(const AUnit: TcwUnit);
    property Locked[AUnit: TcwUnit]: Boolean read GetLocked;
    property ModuleNode[AUnit: TcwUnit]: TcwModuleNode read GetModuleNode;
    property RefactorInPlace: Boolean read FRefactorInPlace write FRefactorInPlace;
    property RefactoringList: TcwRefactoringList read FRefactoringList;
  public
    constructor Create;
    destructor Destroy; override;

    // IcwRefactoring
    property Name: string read GetName;
    property MenuCaption: string read GetMenuCaption;
    property CanExecute: Boolean read GetCanExecute;
    property GroupName: string read GetGroupName;
  end;

implementation

uses
  Windows,
  ShellAPI,
  SysUtils;

{ TcwRefactoringListNode }

constructor TcwRefactoringListNode.Create(ARefactoring: TcwRefactoring; const AUnit: TcwUnit);
begin
  inherited Create;
  FUnit := AUnit;
  FModuleNode := ARefactoring.TryCreateModuleNode(AUnit);
end;

destructor TcwRefactoringListNode.Destroy;
begin
  FModuleNode.Free;
  inherited;
end;

{ TcwRefactoringList }

constructor TcwRefactoringList.Create(ARefactoring: TcwRefactoring);
begin
  inherited Create;
  FRefactoring := ARefactoring;
  FObjectList := TObjectList.Create(True);
end;

destructor TcwRefactoringList.Destroy;
begin
  FObjectList.Free;
  inherited;
end;

function TcwRefactoringList.Add(ARefactoringListNode: TcwRefactoringListNode): Integer;
begin
  Result := FObjectList.Add(ARefactoringListNode);
end;

function TcwRefactoringList.Count: Integer;
begin
  Result := FObjectList.Count;
end;

function TcwRefactoringList.GetNode(Index: Integer): TcwRefactoringListNode;
begin
  Result := FObjectList[Index] as TcwRefactoringListNode;
end;

function TcwRefactoringList.IndexOf(AUnit: TcwUnit): Integer;
var
  Index: Integer;
begin
  Result := -1;
  for Index := 0 to Count - 1 do
    if Nodes[Index]._Unit = AUnit then
    begin
      Result := Index;
      Break;
    end;
end;

procedure TcwRefactoringList.Delete(Index: Integer);
begin
  FObjectList.Delete(Index);
end;

{ TcwRefactoring }

constructor TcwRefactoring.Create;
begin
  inherited;
  FRefactoringList := TcwRefactoringList.Create(Self);
end;

destructor TcwRefactoring.Destroy;
begin
  FRefactoringList.Free;
  inherited;
end;

procedure TcwRefactoring.ViewRefactoring(ABaseFolderName: string);
var
  APath: string;
  AParams: string;
begin
  if not FRefactorInPlace then
  begin
    GetFilenameToGenerate(gcwProject.cwOptions.FileToAnalyze, ABaseFolderName, False, APath);
    APath := Format('"%s"', [APath]);
    if gcwProject.SystemOptions.RefactorOptions.FolderComparisonExe = '' then
      ShellExecute(0, 'open', PChar(APath), nil, nil, SW_SHOW)
    else
    begin
      AParams := gcwProject.SystemOptions.RefactorOptions.FolderComparisonExeParams;
      AParams := StringReplace(AParams, '%1',
        Format('"%s"', [ExtractFileDir(gcwProject.cwOptions.FileToAnalyze)]), []);
      AParams := StringReplace(AParams, '%2', APath, []);
      ShellExecute(0, 'open', PChar(gcwProject.SystemOptions.RefactorOptions.FolderComparisonExe),
        PChar(AParams), nil, SW_SHOW)
    end;
  end;
end;

function TcwRefactoring.TryCreateModuleNode(const AUnit: TcwUnit): TcwModuleNode;
var
  AParser: TcwNodeBuilder;
begin
  if gcwProject.FileNames[AUnit.FileIndex].FileExists then
  begin
    Result := TcwModuleNode.Create;
    AParser := TcwNodeBuilder.Create(Result);
    AParser.Run(AUnit.FullFileName);
    AParser.Free;
  end else
    Result := nil;
end;

function TcwRefactoring.FindNodeForLocation(AStartNode: TcwNode;
  const ALocation: TcwCodeLocation): TcwCodeNode;

  function Recurse(ANode: TcwNode): TcwCodeNode;
  var
    Index: Integer;
    AModuleCodeNode: TcwCodeNode;
  begin
    Result := nil;
    if (ANode.Count = 0) and (ANode is TcwCodeNode) then
    begin
      AModuleCodeNode := ANode as TcwCodeNode;
      if (AModuleCodeNode.Point.X = ALocation.X) and
        (AModuleCodeNode.Point.Y = ALocation.Y) then
        Result := AModuleCodeNode;
    end;
    if not Assigned(Result) then
      for Index := 0 to ANode.Count - 1 do
      begin
        Result := Recurse(ANode[Index]);
        if Assigned(Result) then
          Break;
      end;
  end;

begin
  Result := Recurse(AStartNode);
end;

function TcwRefactoring.GetFilenameToGenerate(const AFileName: string; const
  ABaseFolderName: string;
  ACreateDirs: Boolean; out APath: string): string;

  function IsUNC(AFileName: string): Boolean;
  begin
    Result := Copy(AFileName, 1, 2) = '\\';
  end;

  function GetAndSetUNCServerName(var AFileName: string): string;
  var
    APos: Integer;
  begin
    AFileName := Copy(AFileName, 3, Length(AFileName));
    APos := Pos('\', AFileName);
    if APos > 0 then
    begin
      Result := Copy(AFileName, 1, APos - 1);
      AFileName := Copy(AFileName, APos + 1, Length(AFileName));
    end
    else
      Result := '';
  end;

  function GetAndSetDriveName(var AFileName: string): string;
  begin
    Result := ExtractFileDrive(AFileName);
    if Length(Result) > 1 then
      Result := Result[1];
    AFileName := Copy(AFileName, 4, Length(AFileName));
  end;

var
  AConvertedFileName: string;
begin
  if FRefactorInPlace then
    Result := AFileName
  else
  begin
    AConvertedFileName := AFileName;
    if IsUNC(AConvertedFileName) then
      APath := 'UNC\' + GetAndSetUNCServerName(AConvertedFileName)
    else
      APath := 'Drives\' + GetAndSetDriveName(AConvertedFileName);
    APath := Format('%s\%s', [GetQualifiedBaseFolderName(ABaseFolderName), APath]);
    Result := Format('%s\%s', [APath, AConvertedFileName]);
    APath := ExtractFileDir(ExpandFileName(Result));
    if not ForceDirectories(APath) then
      raise Exception.CreateFmt('Could not create directory, %s', [APath]);
  end;
end;

function TcwRefactoring.GetQualifiedBaseFolderName(ABaseFolderName: string): string;
begin
  Result := '.\Generated\' + ABaseFolderName;
end;

function TcwRefactoring.GetBaseFolderName: string;
var
  AGUID: TGUID;
begin
  CreateGUID(AGUID);
  Result := GUIDToString(AGUID);
end;

procedure TcwRefactoring.BeginRefactoring(ABaseFolderName: string; const AUnit: TcwUnit);
var
  ARefactoringListNode: TcwRefactoringListNode;
begin
  if Locked[AUnit] then
    raise Exception.Create('Unit is already being refactored.')
  else
  begin
    ARefactoringListNode := TcwRefactoringListNode.Create(Self, AUnit);
    FRefactoringList.Add(ARefactoringListNode);
  end;
end;

function TcwRefactoring.CheckRefactoringUnit(AUnit: TcwUnit): Integer;
begin
  Result := FRefactoringList.IndexOf(AUnit);
  if Result = -1 then
    raise Exception.CreateFmt('Unit, %s, is not being refactored.', [AUnit.Name]);
end;

procedure TcwRefactoring.EndRefactoring(ABaseFolderName: string; const AUnit: TcwUnit);
var
  AModuleNode: TcwModuleNode;
  AIndex: Integer;
  APath: string;
begin
  AIndex := CheckRefactoringUnit(AUnit);
  AModuleNode := FRefactoringList.Nodes[AIndex].ModuleNode;
  AModuleNode.Generate(GetFilenameToGenerate(AUnit.FullFileName, ABaseFolderName, True, APath),
    not RefactorInPlace);
  gcwRefactoringHost.RefactoringStatus(AUnit);
  FRefactoringList.Delete(AIndex);
end;

procedure TcwRefactoring.CancelRefactoring(const AUnit: TcwUnit);
var
  AIndex: Integer;
begin
  AIndex := CheckRefactoringUnit(AUnit);
  FRefactoringList.Delete(AIndex);
  gcwRefactoringHost.RefactoringStatus(AUnit);
end;

function TcwRefactoring.GetModuleNode(AUnit: TcwUnit): TcwModuleNode;
var
  AIndex: Integer;
begin
  AIndex := FRefactoringList.IndexOf(AUnit);
  if AIndex >= 0 then
    Result := FRefactoringList.Nodes[AIndex].ModuleNode
  else
    Result := nil;
end;

function TcwRefactoring.FindCodeLocationForNode(const AUnit: TcwUnit; ANode: TcwCodeNode):
  TcwCodeLocation;
var
  Index: Integer;
  ACodeLocation: TcwCodeLocation;
begin
  Result := nil;
  for Index := 0 to AUnit.UsedByLocations.Count - 1 do
  begin
    ACodeLocation := AUnit.UsedByLocations[Index] as TcwCodeLocation;
    if (ACodeLocation.X = ANode.Point.X) and (ACodeLocation.Y = ANode.Point.Y) then
    begin
      Result := ACodeLocation;
      Break;
    end;
  end;
end;

function TcwRefactoring.GetGroupName: string;
begin
  Result := '';
end;

function TcwRefactoring.GetLocked(AUnit: TcwUnit): Boolean;
begin
  Result := FRefactoringList.IndexOf(AUnit) >= 0;
end;

procedure TcwRefactoring.EndRefactorings(ABaseFolderName: string);
begin
  while FRefactoringList.Count > 0 do
    EndRefactoring(ABaseFolderName, FRefactoringList.Nodes[0]._Unit);
end;

end.
