(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwAnalyzeParser;

interface

uses
  SysUtils,
  Classes,
  Types,
  mwPasLexTypes,
  mwSimplePasPar,
  ucwClasses,
  ucwTypeParser;

type
  TAnalyzeParser = class(TTypeParser)
  private
    FAsSymbolFound: Boolean;
    FLastAsSymbol: TcwSymbol;
    function ParseAsOperator: TcwSymbol;
  protected
    function ParseQualifiedIdentifier(const aSymbol: TcwSymbol): TcwSymbol; override;
    procedure Factor; override;
    procedure WithStatement; override;
    procedure ForStatement; override;
    function ParseVariable(const aSymbol: TcwSymbol): TcwSymbol;
    function ParseVariableTwo(const aVarOne: TcwSymbol): TcwSymbol;
    function ParseVariableReference(const aSymbol: TcwSymbol): TcwSymbol; override;
    procedure ParseVariableList(const aList: TcwHashedObjectList);
  end;

implementation

uses
  mwPasLex,
  mwSimplePasParTypes;

function TAnalyzeParser.ParseQualifiedIdentifier(const aSymbol: TcwSymbol): TcwSymbol;
var
  vContainer: TcwSymbolContainer;
  vIdentName: string;
  vLastTokenID: TptTokenKind;
  AQualifyingClass: TcwClass;
begin
  Result := nil;
  if Assigned(aSymbol) then
  begin
    vContainer := aSymbol.FindContainer;
    if Assigned(vContainer) then
    begin
      if vContainer is TcwClass then
        AQualifyingClass := vContainer as TcwClass
      else
        AQualifyingClass := FCurrentClass;
      Result := vContainer.FindSymbol(UnitIntf, AQualifyingClass, Lexer.Token, True,
        GetUsedByInfo);
    end;
  end else
    vContainer := nil;

  if not Assigned(Result) then
    Result := FindSymbol(Lexer.Token, True, GetUsedByInfo);

  if Assigned(Result) then
  begin
    FLastAsSymbol := Result;
    FAsSymbolFound := True;

    //TODO: suppress warning messages option
    //if Assigned(OnMessage) and not Assigned(Result.GetType) and Result.TypeRequired then
    //  OnMessage(Self, Format('Type not found for %s.', [Lexer.Token]), Lexer.PosXY.X,
    //    Lexer.PosXY.Y);

    vContainer := Result.FindContainer;
  end else
    //TODO: suppress warning messages option
    if Assigned(OnMessage) then
      OnMessage(Self, Format('Symbol not found, %s.', [Lexer.Token]), Lexer.PosXY.X, Lexer.PosXY.Y);

  Expected(ptIdentifier);
  while TokenID = ptPointerSymbol do
    //pointers levels of indirection
    NextToken;

  if TokenID in [ptSquareOpen, ptPoint] then
    while TokenID in [ptSquareOpen, ptPoint] do
    begin //jdj 1/7/2001
      // look for default array property...
      vLastTokenID := TokenID;
      if (TokenID = ptSquareOpen) then
      begin
        if (Result is TcwProperty) and (Result as TcwProperty).IsArrayProperty then
        begin
          ConstantExpression;
          Break;
        end else
          if Assigned(Result) and (Result.TypeIntf is TcwClass) and
            Assigned((Result.TypeIntf as TcwClass).DefaultProperty) then
            vIdentName := (Result.TypeIntf as TcwClass).DefaultProperty.Name
          else
          begin
            ConstantExpression;
            Break;
          end;
      end else
      begin
        NextToken;
        vIdentName := Lexer.Token;
      end;

      if Assigned(vContainer) then
      begin
        Result := vContainer.FindSymbol(UnitIntf, FCurrentClass, vIdentName, True,
          GetUsedByInfo);
        if Assigned(Result) then
          // if Result is a constructor, then set result to the current class in the hierarchy
          if (vContainer is TcwClass) and (Result is TcwRoutine) and
            ((Result as TcwRoutine).RoutineKind = rkConstructor) then
            Result := vContainer
          else
            vContainer := Result.FindContainer;
      end else
        Result := nil;

      case vLastTokenID of
        ptPoint: Expected(ptIdentifier);
        ptSquareOpen: ConstantExpression;
        else
          Assert(False);
      end;
    end;
end;

procedure TAnalyzeParser.Factor;
begin
  case TokenID of
    ptAsciiChar, ptStringConst: ParseCharString;
    ptAddressOp, ptDoubleAddressOp, ptIdentifier, ptInherited, ptPointerSymbol, ptRoundOpen:
      Designator;
    ptIntegerConst, ptFloat: Number;
    ptNil: NextToken;
    ptMinus: begin
      NextToken;
      Factor;
    end;
    ptNot: begin
      NextToken;
      Factor;
    end;
    ptPlus: begin
      NextToken;
      Factor;
    end;
    ptSquareOpen: SetConstructor;
    ptString: begin
      NextToken;
      Factor;
    end;
  end;
end;

function TAnalyzeParser.ParseAsOperator: TcwSymbol;
begin
  NextToken;
  Result := ParseQualifiedIdentifier(nil);
  FLastAsSymbol := Result;
  FAsSymbolFound := True;
end;

function TAnalyzeParser.ParseVariable(const aSymbol: TcwSymbol): TcwSymbol;
begin
  Result := aSymbol;
  case TokenID of
    ptInherited: begin
      NextToken;
      Result := ParseQualifiedIdentifier(Result);
    end;
    ptPoint: ParseVariableTwo(Result);
    ptPointerSymbol: ParseVariableTwo(Result);
    ptRoundOpen: ParseVariableTwo(Result);
    ptSquareOpen: ParseVariableTwo(Result);
    else
      Result := ParseQualifiedIdentifier(Result);
  end;
  Result := ParseVariableTwo(Result);
  if TokenID = ptAs then
    Result := ParseAsOperator;
end;

function TAnalyzeParser.ParseVariableTwo(const aVarOne: TcwSymbol): TcwSymbol;
begin
  Result := aVarOne;
  case TokenID of
    ptPoint: begin
      NextToken;
      case TokenID of
        ptAddressOp, ptDoubleAddressOp, ptIdentifier: Result := ParseVariableReference(Result);
        ptPointerSymbol, ptRoundOpen, ptSquareOpen: Result := ParseVariableTwo(Result);
      end;
    end;
    ptPointerSymbol: begin
      NextToken;
      case TokenID of
        ptAddressOp, ptDoubleAddressOp, ptIdentifier: Result := ParseVariableReference(Result);
        ptPoint, ptPointerSymbol, ptRoundOpen, ptSquareOpen: Result := ParseVariableTwo(Result);
      end;
    end;
    ptRoundOpen: begin
      NextToken;
      fInRound := True;
      case TokenID of
        ptRoundClose: NextToken;
        else
        begin
          FAsSymbolFound := False;
          FLastAsSymbol := nil;
          case TokenID of
            ptAddressOp, ptDoubleAddressOp: ParseVariableReference(Result);
            ptPoint, ptPointerSymbol, ptSquareOpen: ParseVariableTwo(Result);
            ptRoundOpen: ParseVariableTwo(nil);
          end;
          fInRound := True;
          ExpressionList;
          fInRound := True;
          Expected(ptRoundClose);
          fInRound := False;
          if not Assigned(Result) and FAsSymbolFound then
            Result := FLastAsSymbol;
        end;
      end;

      case TokenID of
        ptAs: ParseAsOperator;
        ptAddressOp, ptDoubleAddressOp: ParseVariableReference(Result);
        ptPoint, ptPointerSymbol, ptRoundOpen, ptSquareOpen: Result := ParseVariableTwo(Result);
      end;
    end;
    ptSquareOpen: begin
      Lexer.InitAhead;
      while Lexer.AheadTokenID <> ptSemiColon do
      begin
        case Lexer.AheadTokenID of
          ptBegin, ptClass, ptConst, ptEnd, ptDotDot, ptIn, ptNull, ptThreadVar, ptType, ptVar:
            Break;
          else
            Lexer.AheadNext;
        end;
      end;
      case Lexer.AheadTokenID of
        ptDotDot: SubrangeType;
        else
        begin
          NextToken;
          case TokenID of
            ptSquareClose: NextToken;
            else
            begin
              case TokenID of
                ptAddressOp, ptDoubleAddressOp: ParseVariableReference(Result);
                ptPoint, ptPointerSymbol, ptRoundOpen, ptSquareOpen: ParseVariableTwo(Result);
              end;
              ExpressionList;
              Expected(ptSquareClose);
            end;
          end;
          case TokenID of
            ptAs: ParseAsOperator;
            ptAddressOp, ptDoubleAddressOp: ParseVariableReference(Result);
            ptPoint, ptPointerSymbol, ptRoundOpen, ptSquareOpen: ParseVariableTwo(Result);
          end;
        end;
      end;
    end
  end;
end;

function TAnalyzeParser.ParseVariableReference(const aSymbol: TcwSymbol): TcwSymbol;
begin
  Result := aSymbol;
  case TokenID of
    ptAddressOp: begin
      NextToken;
      Result := ParseVariable(Result);
    end;
    ptDoubleAddressOp: begin
      NextToken;
      Result := ParseVariable(Result);
    end;
    ptPointerSymbol: begin
      NextToken;
      case TokenID of
        ptRoundClose, ptSquareClose: ;
        ptIdentifier: NextToken; // (cbm)
        else
          // (cbm) commenting because string literals can be coded like, ^C
          // in this case, we don't want to parse "C" as an identifier, so we're skipping
          // to nexttoken (see, ptIdentifer above in this case statement)
          //Result := ParseVariable(Result);
      end;
    end;
    else
      Result := ParseVariable(Result);
  end;
end;

procedure TAnalyzeParser.ParseVariableList(const aList: TcwHashedObjectList);

  procedure AddSymbolToScopeVariables(const aList: TcwHashedObjectList; aSymbol: TcwSymbol);
  var
    aContainer: TcwSymbolContainer;
  begin
    if (aList <> nil) and (aSymbol <> nil) then
    begin
      aContainer := aSymbol.FindContainer;
      if Assigned(aContainer) then
        aList.Add(aContainer.Name, aContainer)
      else
        aList.Add(aSymbol.Name, aSymbol);
    end;
  end;

var
  vSymbol: TcwSymbol;
begin
  vSymbol := ParseVariableReference(nil);
  AddSymbolToScopeVariables(aList, vSymbol);
  while Lexer.TokenID = ptComma do
  begin
    NextToken;
    vSymbol := ParseVariableReference(nil);
    AddSymbolToScopeVariables(aList, vSymbol);
  end;
end;

procedure TAnalyzeParser.WithStatement;
var
  Cnt: Integer;
begin
  Expected(ptWith);
  Cnt := ScopeVariables.Count;
  ParseVariableList(ScopeVariables);
  Expected(ptDo);
  Statement;
  while ScopeVariables.Count > Cnt do
    ScopeVariables.Delete(ScopeVariables.Count - 1);
end;

procedure TAnalyzeParser.ForStatement;
begin
  Expected(ptFor);
  ParseQualifiedIdentifier(nil);
  Expected(ptAssign);
  Expression;
  case TokenID of
    ptTo:
      begin
        NextToken;
      end;
    ptDownTo:
      begin
        NextToken;
      end;
  else
    begin
      SynError(InvalidForStatement);
    end;
  end;
  Expression;
  Expected(ptDo);
  Statement;
end;

end.


