(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwClassParser;

interface

uses
  SysUtils,
  Classes,
  Types,
  mwPasLexTypes,
  mwSimplePasParTypes,
  mwSimplePasPar,
  ucwClasses,
  ucwRoutineParser;

type
  TClassParser = class(TRoutineParser)
  private
    FForwardClasses: TcwHashedObjectList;
  protected
    procedure ParseClassHeritage(const aClass: TcwClass);
    procedure ParseClassVisibility(const aClass: TcwClass; var aVisibility: TVisibilityKind);
    procedure ParseClassField(const aClass: TcwClass; Visibility: TVisibilityKind);
    procedure ParseConstructorHeading(const aClass: TcwClass; Visibility: TVisibilityKind);
    procedure ParseDestructorHeading(const aClass: TcwClass; Visibility: TVisibilityKind);
    procedure ParseClassFunctionHeading(const aClass: TcwClass; Visibility: TVisibilityKind);
    procedure ParseClassProcedureHeading(const aClass: TcwClass; Visibility: TVisibilityKind);
    procedure ParseClassMethodDirective(const aRoutine: TcwRoutine);

    procedure ParseClassMethodHeading(const aClass: TcwClass; Visibility: TVisibilityKind);

    procedure ParseReadAccessIdentifier(const aClass: TcwClass; const aProperty: TcwProperty);
    procedure ParseWriteAccessIdentifier(const aClass: TcwClass; const aProperty: TcwProperty);
    procedure ParseAccessSpecifier(const aClass: TcwClass; const aProperty: TcwProperty);
    procedure ParsePropertySpecifiers(const aClass: TcwClass; const aProperty: TcwProperty);

    procedure ParseClassProperty(const aClass: TcwClass; Visibility: TVisibilityKind);
    procedure ParseClassMethodOrProperty(const aClass: TcwClass; Visibility: TVisibilityKind);
    procedure ParseClassMemberList(const aClass: TcwClass);

    procedure ParseClassForward(const aClassName: string; const aCodeLocation: TPoint);
    procedure ParseClassType(const aClassName: string; const aCodeLocation: TPoint);
  public
    destructor Destroy; override;
  end;

implementation

uses
  mwPasLex,
  ucwUnitParser;

destructor TClassParser.Destroy;
begin
  FForwardClasses.Free;
  inherited;
end;

procedure TClassParser.ParseClassHeritage(const aClass: TcwClass);
var
  vInheritanceList: TStringList;
  vClass: TcwClass;
  i: Integer;
  Loc: TSmallPoint;
begin
  Expected(ptRoundOpen);
  vInheritanceList := TStringList.Create;
  try
    GetQualifiedIdentifierList(vInheritanceList);
    for i := 0 to vInheritanceList.Count - 1 do
    begin
      Loc := TSmallPoint(vInheritanceList.Objects[i]);
      if i = 0 then
      begin
        vClass := FindVisibleClass(vInheritanceList[i]);
        aClass.InheritsFromClass := vClass;
        if Assigned(vClass) then
          vClass.Descendants.Add(aClass);
      end else
      begin
        {vInterface := }FindVisibleInterface(vInheritanceList[i]);
        //TODO: aclass.interfaces will own vInterface, but vInterface is owned somwhere else
        //if Assigned(vInterface) then
        //  aClass.Interfaces.AddInterface(vInterface);
      end;
    end;
  finally
    vInheritanceList.Free;
  end;
  Expected(ptRoundClose);
end;

procedure TClassParser.ParseClassVisibility(const aClass: TcwClass; var aVisibility: TVisibilityKind);
begin
  while ExID in [ptStrict, ptAutomated, ptPrivate, ptProtected, ptPublic, ptPublished] do
  begin
    Lexer.InitAhead;
    case Lexer.AheadExID of
      ptColon, ptComma: ;
    else
      begin
        case ExID of
          ptAutomated:
            begin
              aVisibility := vkAutomated;
              ExpectedEx(ptAutomated);
            end;
          ptPrivate:
            begin
              aVisibility := vkPrivate;
              ExpectedEx(ptPrivate);
            end;
          ptProtected:
            begin
              aVisibility := vkProtected;
              ExpectedEx(ptProtected);
            end;
          ptPublic:
            begin
              aVisibility := vkPublic;
              ExpectedEx(ptPublic);
            end;
          ptPublished:
            begin
              aVisibility := vkPublished;
              ExpectedEx(ptPublished);
            end;
          ptStrict:
            begin
              case Lexer.AheadExID of
                ptPrivate:
                  begin
                    ExpectedEx(ptStrict);
                    ExpectedEx(ptPrivate);
                    aVisibility := vkStrictPrivate;
                  end;
                ptProtected:
                  begin
                    ExpectedEx(ptStrict);
                    ExpectedEx(ptProtected);
                    aVisibility := vkStrictProtected;
                  end;
              end;
            end;
        end;
        if Assigned(aClass) then
          aClass.AddVisibilitySpecifier(Lexer.PosXY.Y, aVisibility); //todo: .sm dont add +1 to YPos; we're already on the next line?!
      end;
    end;
  end;
end;

procedure TClassParser.ParseClassField(const aClass: TcwClass; Visibility: TVisibilityKind);
var
  ATypeName: string;
  Cnt: Integer;
  i: Integer;
  vType: TcwSymbol;
  vVar: TcwVariable;
begin
  //IdentifierList;
  Cnt := aClass.Fields.Count;
  vVar:=aClass.Fields.AddVariable(Lexer.Token, Visibility, False,
    FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
  Expected(ptIdentifier);
  StoreDeclarationEnd(vVar); 
  while (TokenID = ptComma) do
  begin
    NextToken;
    vVar:=aClass.Fields.AddVariable(Lexer.Token, Visibility, False,
      FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
    Expected(ptIdentifier);
    StoreDeclarationEnd(vVar);
  end;

  Expected(ptColon);

  if TokenID = ptIdentifier then
    ATypeName := Lexer.Token
  else
    ATypeName := '';

  vType := ParseTypeKind(aClass, ATypeName, Lexer.PosXY, False);
  if Assigned(vType) then
    for i := Cnt to aClass.Fields.Count - 1 do
      aClass.Fields[i].TypeIntf := vType;
end;

procedure TClassParser.ParseClassMethodDirective(const aRoutine: TcwRoutine);
var
  Directive: TRoutineDirective;
begin
  while (ExId in [ptAbstract, ptCdecl, ptDynamic, ptMessage, ptOverride,
    ptOverload, ptPascal, ptAssembler, ptRegister, ptReintroduce, ptSafeCall, ptStdCall,
    ptVirtual]) or (ExID in [ptDeprecated, ptPlatform]) or (TokenID = ptLibrary) do
  begin
    if (ExID in [ptDeprecated, ptPlatform]) or (TokenID = ptLibrary) then
      HintDirectives
    else
    begin
      Directive := ParseProceduralDirective;
      if Directive <> rdNone then
        aRoutine.AddDirective(Directive);
    end;
    SEMICOLON;
  end;
end;

procedure TClassParser.ParseConstructorHeading(const aClass: TcwClass; Visibility: TVisibilityKind);
var
  vResultVar: TcwVariable;
  vRoutine: TcwRoutine;
begin
  Expected(ptConstructor);
  vRoutine := aClass.Routines.AddRoutine(Lexer.Token, Visibility, rkConstructor,
    FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);

  // (cbm) adding a Result virtual variable, since constructors can act like functions
  vResultVar := TcwVariable.Create(cResultVariableName, vkNone, False, 
    FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
  vRoutine.ReturnType := vResultVar;
  vResultVar.TypeIntf := AClass;
  vRoutine.AddSymbol(vResultVar);

  ConstructorName;
  if TokenID = ptRoundOpen then
    ParseFormalParameterList(vRoutine, True, False);
  StoreDeclarationEnd(vRoutine);
  SEMICOLON;
  ParseClassMethodDirective(vRoutine);
end;

procedure TClassParser.ParseDestructorHeading(const aClass: TcwClass; Visibility: TVisibilityKind);
var
  vRoutine: TcwRoutine;
begin
  Expected(ptDestructor);
  vRoutine := aClass.Routines.AddRoutine(Lexer.Token, Visibility, rkDestructor,
    FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);

  DestructorName;
  if TokenID = ptRoundOpen then
    ParseFormalParameterList(vRoutine, True, False);
  StoreDeclarationEnd(vRoutine);
  SEMICOLON;
  ParseClassMethodDirective(vRoutine);
end;

procedure TClassParser.ParseClassFunctionHeading(const aClass: TcwClass; Visibility: TVisibilityKind);
var
  vRoutine: TcwRoutine;
begin
  Expected(ptFunction);
  vRoutine := aClass.Routines.AddRoutine(Lexer.Token, Visibility, rkFunction,
    FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);

  FunctionMethodName;
  if TokenID = ptRoundOpen then
    ParseFormalParameterList(vRoutine, True, False);
  Expected(ptColon);
  AddReturnType(vRoutine);
  StoreDeclarationEnd(vRoutine);
  SEMICOLON;
  if ExID = ptDispId then
  begin
    NextToken;
    ConstantExpression;
    StoreDeclarationEnd(vRoutine);
    SEMICOLON;
  end;
  ParseClassMethodDirective(vRoutine);
end;

procedure TClassParser.ParseClassProcedureHeading(const aClass: TcwClass; Visibility: TVisibilityKind);
var
  vRoutine: TcwRoutine;
begin
  Expected(ptProcedure);
  vRoutine := aClass.Routines.AddRoutine(Lexer.Token, Visibility, rkProcedure,
    FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);

  ProcedureMethodName;
  if TokenID = ptRoundOpen then
    ParseFormalParameterList(vRoutine, True, False);
  StoreDeclarationEnd(vRoutine); 
  SEMICOLON;
  if ExID = ptDispId then
  begin
    NextToken;
    ConstantExpression;
    StoreDeclarationEnd(vRoutine);
    SEMICOLON;
  end;
  ParseClassMethodDirective(vRoutine);
end;

procedure TClassParser.ParseClassMethodHeading(const aClass: TcwClass; Visibility: TVisibilityKind);
begin
  case TokenID of
    ptConstructor:
      begin
        ParseConstructorHeading(aClass, Visibility);
      end;
    ptDestructor:
      begin
        ParseDestructorHeading(aClass, Visibility);
      end;
    ptFunction:
      begin
        Lexer.InitAhead;
        Lexer.AheadNext;
        case Lexer.AheadTokenID of
          PtPoint:
            begin
              ClassMethodResolution;
            end;
          else
            begin
              ParseClassFunctionHeading(aClass, Visibility);
            end;
        end;
      end;
    ptProcedure:
      begin
        Lexer.InitAhead;
        Lexer.AheadNext;
        case Lexer.AheadTokenID of
          PtPoint:
            begin
              ClassMethodResolution;
            end;
          else
            begin
              ParseClassProcedureHeading(aClass, Visibility);
            end;
        end;
      end;
    else
      SynError(InvalidClassMethodHeading);
  end;
end;

procedure TClassParser.ParseReadAccessIdentifier(const aClass: TcwClass; const aProperty: TcwProperty);
begin
  aProperty.ReadAccess := ParseQualifiedIdentifier(aClass);
end;

procedure TClassParser.ParseWriteAccessIdentifier(const aClass: TcwClass; const aProperty: TcwProperty);
begin
  aProperty.WriteAccess := ParseQualifiedIdentifier(aClass);
end;

procedure TClassParser.ParseAccessSpecifier(const aClass: TcwClass; const aProperty: TcwProperty);
begin
  case ExID of
    ptRead:
      begin
        NextToken;
        ParseReadAccessIdentifier(aClass, aProperty);
      end;
    ptWrite:
      begin
        NextToken;
        ParseWriteAccessIdentifier(aClass, aProperty);
      end;
    ptReadOnly:
      begin
        NextToken;
      end;
    ptWriteOnly:
      begin
        NextToken;
      end;
    else
      begin
        SynError(InvalidAccessSpecifier);
      end;
  end;
end;

procedure TClassParser.ParsePropertySpecifiers(const aClass: TcwClass; const aProperty: TcwProperty);
begin
  if ExID = ptIndex then
  begin
    NextToken;
    ConstantExpression;
  end;
  while ExID in [ptRead, ptReadOnly, ptWrite, ptWriteOnly] do
  begin
    ParseAccessSpecifier(aClass, aProperty);
  end;
  if ExID = ptDispId then
  begin
    NextToken;
    ConstantExpression;
  end;
  while ExID in [ptDefault, ptNoDefault, ptStored] do
  begin
    StorageSpecifier;
  end;
  if ExID = ptImplements then
  begin
    NextToken;
    QualifiedIdentifierList;
  end;
end;

procedure TClassParser.parseClassProperty(const aClass: TcwClass; Visibility: TVisibilityKind);

  // (cbm) get the type from the ancestor class...
  procedure SetPropertyTypeFromAncestorProperty(const aAncestorClass: TcwClass;
    const aProperty: TcwProperty);
  var
    Index: Integer;
    aAncestorProperty: TcwProperty;
  begin
    if Assigned(aAncestorClass) then
    begin

      for Index := 0 to aAncestorClass.Properties.Count - 1 do
      begin
        aAncestorProperty := aAncestorClass.Properties[Index];
        if (aAncestorProperty.Name = aProperty.Name) and
          Assigned(aProperty.UnitIntf) and
          Assigned(aAncestorProperty.UnitIntf) and
          ((aAncestorProperty.UnitIntf.Name = aProperty.UnitIntf.Name) or
           (aAncestorProperty.Visibilty in [vkProtected, vkPublic, vkPublished])) then
        begin
          aProperty.TypeIntf := aAncestorProperty.TypeIntf;
          aProperty.IsArrayProperty := aAncestorProperty.IsArrayProperty;
          Break;
        end;
      end;
      if not Assigned(aProperty.TypeIntf) then
        SetPropertyTypeFromAncestorProperty(aAncestorClass.InheritsFromClass, aProperty);
    end;
  end;

var
  vProperty: TcwProperty;
begin
  Expected(ptProperty);
  vProperty := aClass.Properties.AddProperty(Lexer.Token, Visibility, aClass,
    FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
  PropertyName;
  case TokenID of
    ptColon, ptSquareOpen:
      begin
        if TokenID = ptSquareOpen then
        begin
          vProperty.IsArrayProperty := True;
          PropertyParameterList;
        end;
        Expected(ptColon);
        vProperty.TypeIntf := ParseTypeId;
      end;

    // ptSemiColon means the property has no type, so its a visibility promotion
    // get the type from the ancestor class...
    ptSemiColon: SetPropertyTypeFromAncestorProperty(aClass.InheritsFromClass, vProperty);

    else //IsArrayProperty:= False; (cbm) moved initialization to top
  end;
  ParsePropertySpecifiers(aClass, vProperty);
  StoreDeclarationEnd(vProperty); 
  SEMICOLON;
  if vProperty.IsArrayProperty then
    // check for default array property
    case ExID of
      ptDefault:
        begin
          vProperty.DefaultProperty := True;
          PropertyDirective;
        end;
    end;
end;

procedure TClassParser.ParseClassMethodOrProperty(const aClass: TcwClass; Visibility: TVisibilityKind);
begin
  if TokenID = ptClass then
  begin
    NextToken;
  end;
  case TokenID of
    ptProperty:
      begin
        ParseClassProperty(aClass, Visibility);
      end;
    else
      begin
        ParseClassMethodHeading(aClass, Visibility);
      end;
  end;
end;

procedure TClassParser.ParseClassMemberList(const aClass: TcwClass);
var
  vVisibility: TVisibilityKind;
begin
  vVisibility := vkPublished;
  ParseClassVisibility(aClass, vVisibility);
  while TokenID in [ptClass, ptConstructor, ptDestructor, ptFunction,
    ptIdentifier, ptProcedure, ptProperty] do
  begin
    while TokenID = ptIdentifier do
    begin
      ParseClassField(aClass, vVisibility);
      SEMICOLON;
      ParseClassVisibility(aClass, vVisibility);
    end;
    while TokenID in [ptClass, ptConstructor, ptDestructor, ptFunction,
      ptProcedure, ptProperty] do
    begin
      ParseClassMethodOrProperty(aClass, vVisibility);
    end;
    ParseClassVisibility(aClass, vVisibility);
  end;
end;

procedure TClassParser.ParseClassForward(const aClassName: string; const aCodeLocation: TPoint);
var
  vClass: TcwClass;
begin
  vClass := UnitIntf.Classes.AddClass(aClassName, UnitVisibility, FileIndex,
    aCodeLocation.X + 1, aCodeLocation.Y + 1);
  if not Assigned(FForwardClasses) then
    FForwardClasses := TcwHashedObjectList.Create;
  FForwardClasses.Add(vClass.Name, vClass);
end;

procedure TClassParser.ParseClassType(const aClassName: string; const aCodeLocation: TPoint);

  procedure AssignDefaultHeritage(const AClass: TcwClass);
  var
    ASystemUnit: TcwUnit;
    vClass: TcwClass;
  begin
    ASystemUnit := gcwProject.AllUnits.FindUnit(SSystemPas);
    if Assigned(ASystemUnit) then
    begin
      if (AClass is TcwInterface) then
        vClass := ASystemUnit.Classes.Find('IInterface')
      else
        vClass := ASystemUnit.Classes.Find('TObject');
      if Assigned(vClass) and (vClass <> AClass) then
      begin
        AClass.InheritsFromClass := vClass;
        vClass.Descendants.Add(aClass);
      end;
    end;
  end;

var
  AIndex: Integer;
  vClass: TcwClass;
begin
  if Assigned(FForwardClasses) and FForwardClasses.Find(aClassName, AIndex) then
  begin
    vClass := FForwardClasses[AIndex] as TcwClass;
    vClass.X := aCodeLocation.X + 1;
    vClass.Y := aCodeLocation.Y + 1;
    FForwardClasses.Delete(AIndex);
  end else
    vClass := UnitIntf.Classes.AddClass(aClassName, UnitVisibility, FileIndex,
      aCodeLocation.X + 1, aCodeLocation.Y + 1);
  IncIndentLevel;

  FParsingClassInterface := vClass;
  try
    Expected(ptClass);
    case TokenID of
      ptEnd: begin
        AssignDefaultHeritage(vClass);
        NextToken; { Direct descendant of TObject without new members }
        HintDirectives;
      end;
      ptRoundOpen: begin
        ParseClassHeritage(vClass);
        case TokenID of
          ptEnd: begin
            NextToken; { No new members }
            HintDirectives;
          end;
          ptSemiColon: ; { No new members }
          else
          begin
            ParseClassMemberList(vClass);
            Expected(ptEnd);
            HintDirectives;
          end;
        end;
      end;
      else
      begin
        AssignDefaultHeritage(vClass);
        ParseClassMemberList(vClass); { Direct descendant of TObject }
        Expected(ptEnd);
        HintDirectives;
      end;
    end;
  finally
    FParsingClassInterface := nil;
  end;
  StoreDeclarationEnd(vClass);
end;

end.
