(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwIniOptions;

interface

uses
  Windows, SysUtils, IniFiles, Classes, Variants;

var
  BoolOptions: array[Boolean] of Char = ('F', 'T');

type
  TOptions = class
  private
    FIniFileName: string;
    FIni: TIniFile;
    function DoGetOption(const aIni: TIniFile; const aSection: string; const aOption: string;
      const aDefaultValue: Variant): Variant;
    procedure DoSetOption(const aIni: TIniFile; const aSection: string; const aOption: string;
      const aNewValue: Variant);
    procedure DoEraseSection(const aIni: TIniFile; const aSection: string);
  public
    constructor Create(const aFileName: string);
    destructor Destroy; override;
    property IniFileName: string read FIniFileName;

    function GetOption(const aSection: string; const aOption: string; const aDefaultValue: Variant):
      Variant;
    procedure SetOption(const aSection: string; const aOption: string; const aNewValue: Variant);

    procedure ReadSection(const aSection: string; const aOptionStrings: TStrings);
    procedure ReadSectionValues(const aSection: string; const aOptionStrings: TStrings);
    procedure SaveTStringOptions(const aSection: string; const aOptionStrings: TStrings);
  end;

implementation

function VarToString(const aValue: Variant): string;
begin
  case varType(aValue) of
    varSmallInt,
    varInteger,
    varSingle,
    varWord,
    varLongWord,
    varInt64: Result := IntToStr(aValue);
    varDouble,
    varCurrency: Result := FloatToStr(aValue);
    varDate: Result := DateToStr(aValue);
    varBoolean: Result := BoolOptions[Boolean(aValue)];
    varString: Result := aValue;
  end;
end;

constructor TOptions.Create(const aFileName: string);
begin
  inherited Create;
  FIniFileName := aFileName;
  FIni := TIniFile.Create(aFileName);
end;

destructor TOptions.Destroy;
begin
  FIni.Free;
  inherited Destroy;
end;

function TOptions.DoGetOption(const aIni: TIniFile; const aSection: string; const
  aOption: string; const aDefaultValue: Variant): Variant;
var
  s, DefaultS: string;
begin
  DefaultS := VarToString(aDefaultValue);
  if aIni = nil then
  begin
    Result := aDefaultValue;
    EXIT;
  end;
  s := aIni.ReadString(aSection, aOption, DefaultS);
  case varType(aDefaultValue) of
    varSmallInt,
    varInteger,
    varSingle,
    varWord,
    varLongWord,
    varInt64: Result := StrToInt(s);
    varDouble,
    varCurrency: Result := StrToFloat(s);
    varDate: Result := StrToDate(s);
    varBoolean: if s = 'T' then
        Result := True
      else
        Result := False;
    varString: Result := s;

    else
      raise Exception.CreateFmt('Unknown value type in section, %s, option, %s.',
        [ASection, AOption]);
  end;
end;

function TOptions.GetOption(const aSection: string; const aOption: string;
  const aDefaultValue: Variant): Variant;
begin
  Result := DoGetOption(FIni, aSection, aOption, aDefaultValue);
end;

procedure TOptions.DoSetOption(const aIni: TIniFile; const aSection: string;
  const aOption: string; const aNewValue: Variant);
begin
  if (aSection <> '') and (aOption <> '') then
    aIni.WriteString(aSection, aOption, VarToString(aNewValue));
end;

procedure TOptions.SetOption(const aSection: string; const aOption: string;
  const aNewValue: Variant);
begin
  DoSetOption(FIni, aSection, aOption, aNewValue);
end;

procedure TOptions.DoEraseSection(const aIni: TIniFile; const aSection: string);
begin
  aIni.EraseSection(aSection);
end;

procedure TOptions.ReadSectionValues(const aSection: string; const aOptionStrings: TStrings);
begin
  FIni.ReadSectionValues(aSection, aOptionStrings);
end;

procedure TOptions.ReadSection(const aSection: string; const aOptionStrings: TStrings);
begin
  FIni.ReadSection(aSection, aOptionStrings);
end;

procedure TOptions.SaveTStringOptions(const aSection: string; const aOptionStrings: TStrings);
var
  i: Integer;
  s: string;
  p: Integer;
  vSection: string;
  vOption: string;
begin
  DoEraseSection(FIni, aSection);
  for i := 0 to aOptionStrings.Count - 1 do
  begin
    s := aOptionStrings[i];
    if Trim(s) <> '' then
    begin
      p := Pos('=', s);
      if p > 0 then
      begin
        vSection := Copy(s, 1, p - 1);
        vOption := Copy(s, p + 1, Length(s) - p);
      end else
      begin
        vSection := s;
        vOption := '';
      end;
      FIni.WriteString(aSection, vSection, vOption);
    end;
  end;
end;

end.

