(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwInterfaceParser;

interface

uses
  Types,
  Classes,
  ucwClasses,
  ucwObjectParser,
  mwPasLexTypes;

type
  TInterfaceParser = class(TObjectParser)
  private
    FForwardInterfaces: TcwHashedObjectList;
  protected
    procedure ParseInterfaceHeritage(aInterface: TcwInterface);
    procedure ParseInterfaceGUID(aInterface: TcwInterface);
    procedure ParseInterfaceMemberList(aInterface: TcwInterface);
    procedure ParseInterfaceForward(const aInterfaceName: string; const aCodeLocation: TPoint);
    procedure ParseInterfaceType(const aInterfaceName: string; const aCodeLocation: TPoint);
  public
    destructor Destroy; override;
  end;

implementation

uses
  mwSimplePasParTypes;

destructor TInterfaceParser.Destroy;
begin
  FForwardInterfaces.Free;
  inherited;
end;

procedure TInterfaceParser.ParseInterfaceHeritage(aInterface: TcwInterface);
var
  vInheritanceList: TStringList;
var
  i: Integer;
var
  vInterface: TcwInterface;
var
  Loc: TSmallPoint;
begin
  Expected(ptRoundOpen);
  vInheritanceList := TStringList.Create;
  try
    GetIdentifierList(vInheritanceList);
    for i := 0 to vInheritanceList.Count - 1 do
    begin
      Loc := TSmallPoint(vInheritanceList.Objects[i]);
      vInterface := FindVisibleInterface(vInheritanceList[i]);
      if vInterface <> nil then
      begin
        //TODO: not sure why this is here?
        //aInterface.Interfaces.AddInterface(vInterface);
        aInterface.InheritsFromClass := vInterface; // (cbm)
        vInterface.Descendants.Add(aInterface);
        vInterface.UsedBy.AddUsedBy(UsedByInfo(FileIndex,
          vInterface.Name <> vInheritanceList[i], Loc.X, Loc.Y));
      end;
    end;
  finally
    vInheritanceList.Free;
  end;
  Expected(ptRoundClose);
end;

procedure TInterfaceParser.ParseInterfaceGUID(aInterface: TcwInterface);
begin
  Expected(ptSquareOpen);
  aInterface.GUID := Lexer.Token;
  ParseCharString;
  Expected(ptSquareClose);
end;

procedure TInterfaceParser.ParseInterfaceMemberList(aInterface: TcwInterface);
begin
  while TokenID in [ptFunction, ptProcedure, ptProperty] do
  begin
    ParseClassMethodOrProperty(aInterface, vkPublic);
  end;
end;

procedure TInterfaceParser.ParseInterfaceForward(const aInterfaceName: string;
  const aCodeLocation: TPoint);
var
  vInterface: TcwInterface;
begin
  vInterface := UnitIntf.Interfaces.AddInterface(aInterfaceName, UnitVisibility, 
    FileIndex, aCodeLocation.X + 1, aCodeLocation.Y + 1);
  if not Assigned(FForwardInterfaces) then
    FForwardInterfaces := TcwHashedObjectList.Create;
  FForwardInterfaces.Add(vInterface.Name, vInterface);
end;

procedure TInterfaceParser.ParseInterfaceType(const aInterfaceName: string;
  const aCodeLocation: TPoint);
var
  AIndex: Integer;
  vInterface: TcwInterface;
begin
  if Assigned(FForwardInterfaces) and FForwardInterfaces.Find(aInterfaceName, AIndex) then
  begin
    vInterface := FForwardInterfaces[AIndex] as TcwInterface;
    vInterface.X := aCodeLocation.X + 1;
    vInterface.Y := aCodeLocation.Y + 1;
    FForwardInterfaces.Delete(AIndex);
  end else
    vInterface := UnitIntf.Interfaces.AddInterface(aInterfaceName, UnitVisibility,
      FileIndex, aCodeLocation.X + 1, aCodeLocation.Y + 1);
  case TokenID of
    ptInterface:
      begin
        NextToken;
      end;
    ptDispInterface:
      begin
        vInterface.IsDispatchInterface := True;
        NextToken;
      end
    else
      begin
        SynError(InvalidInterfaceType);
      end;
  end;
  case TokenID of
    ptEnd:
      begin
        NextToken; { Direct descendant without new members }
      end;
    ptRoundOpen:
      begin
        ParseInterfaceHeritage(vInterface);
        case TokenID of
          ptEnd:
            begin
              NextToken; { No new members }
            end;
          ptSemiColon: ; { No new members }
          else
            begin
              if TokenID = ptSquareOpen then
              begin
                ParseInterfaceGUID(vInterface);
              end;
              ParseInterfaceMemberList(vInterface);
              Expected(ptEnd);
            end;
        end;
      end;
    else
      begin
        if TokenID = ptSquareOpen then
        begin
          ParseInterfaceGUID(vInterface);
        end;
        ParseInterfaceMemberList(vInterface); { Direct descendant }
        Expected(ptEnd);
      end;
  end;
end;

end.
