(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwObjectParser;

interface

uses
  Types,
  Classes,
  ucwClasses,
  ucwClassParser,
  mwPasLexTypes;

type
  TObjectParser = class(TClassParser)
  private
    FForwardObjects: TcwHashedObjectList;
  protected
    procedure ParseObjectHeritage(const aObject: TcwClass);
    function ParseObjectVisibility(CurrentVisibility: TVisibilityKind): TVisibilityKind;
    procedure ParseObjectMemberList(const aObject: TcwClass);
    procedure ParseObjectForward(const aClassName: string; const aCodeLocation: TPoint);
    procedure ParseObjectType(const aObjectName: string; const aCodeLocation: TPoint);
  public
    destructor Destroy; override;
  end;

implementation

uses
  mwSimplePasParTypes,
  mwPasLex;

destructor TObjectParser.Destroy;
begin
  FForwardObjects.Free;
  inherited;
end;

procedure TObjectParser.ParseObjectHeritage(const aObject: TcwClass);
var
  vObject: TcwClass;
  vSymbol: TcwSymbol;
begin
  Expected(ptRoundOpen);
  vSymbol := ParseQualifiedIdentifier(nil);
  vObject := vSymbol as TcwClass;
  aObject.InheritsFromClass := vObject;
  if vObject <> nil then
    vObject.Descendants.Add(aObject);
  Expected(ptRoundClose);
end;

function TObjectParser.ParseObjectVisibility(CurrentVisibility: TVisibilityKind): TVisibilityKind;
begin
  Result := CurrentVisibility;
  while ExID in [ptPrivate, ptProtected, ptPublic] do
  begin
    Lexer.InitAhead;
    case Lexer.AheadExID of
      ptColon, ptComma: ;
      else
        case ExID of
          ptPrivate:
            begin
              Result := vkPrivate;
              VisibilityPrivate;
            end;
          ptProtected:
            begin
              Result := vkProtected;
              VisibilityProtected;
            end;
          ptPublic:
            begin
              Result := vkPublic;
              VisibilityPublic;
            end;
        end;
    end;
  end;
end;

procedure TObjectParser.ParseObjectMemberList(const aObject: TcwClass);
var
  vVisibility: TVisibilityKind;
begin {jdj added ptProperty-call to ObjectProperty 02/07/2001}
  vVisibility := vkPublic;
  vVisibility := ParseObjectVisibility(vVisibility);
  while TokenID in [ptConstructor, ptDestructor, ptFunction, ptIdentifier,
    ptProcedure, ptProperty] do
  begin
    while TokenID = ptIdentifier do
    begin
      ParseClassField(aObject, vVisibility);
      SEMICOLON;
      vVisibility := ParseObjectVisibility(vVisibility);
    end;
    while TokenID in [ptConstructor, ptDestructor, ptFunction, ptProcedure, ptProperty] do
    begin
      case TokenID of
        ptConstructor, ptDestructor, ptFunction, ptProcedure:
          ParseClassMethodHeading(aObject, vVisibility);
        ptProperty:
          ParseClassProperty(aObject, vVisibility);
      end;
    end;
    vVisibility := ParseObjectVisibility(vVisibility);
  end;
end;

procedure TObjectParser.ParseObjectForward(const aClassName: string; const aCodeLocation: TPoint);
var
  vObject: TcwClass;
begin
  Expected(ptObject);
  vObject := UnitIntf.Classes.AddClass(aClassName, UnitVisibility, FileIndex,
    aCodeLocation.X + 1, aCodeLocation.Y + 1);
  if not Assigned(FForwardObjects) then
    FForwardObjects := TcwHashedObjectList.Create;
  FForwardObjects.Add(vObject.Name, vObject);
end;

procedure TObjectParser.ParseObjectType(const aObjectName: string; const aCodeLocation: TPoint);
var
  AIndex: Integer;
  vObject: TcwClass;
begin
  if Assigned(FForwardObjects) and FForwardObjects.Find(aObjectName, AIndex) then
  begin
    vObject := FForwardObjects[AIndex] as TcwClass;
    vObject.X := aCodeLocation.X + 1;
    vObject.Y := aCodeLocation.Y + 1;
    FForwardObjects.Delete(AIndex);
  end else
    vObject := UnitIntf.Objects.AddClass(aObjectName, UnitVisibility, 
      FileIndex, aCodeLocation.X + 1, aCodeLocation.Y + 1);

  Expected(ptObject);
  case TokenID of
    ptEnd:
      begin
        NextToken; { Direct descendant without new members }
      end;
    ptRoundOpen:
      begin
        ParseObjectHeritage(vObject);
        case TokenID of
          ptEnd:
            begin
              NextToken; { No new members }
            end;
          ptSemiColon: ; { No new members }
          else
            begin
              ParseObjectMemberList(vObject);
              Expected(ptEnd);
            end;
        end;
      end;
    else
      begin
        ParseObjectMemberList(vObject); { Direct descendant }
        Expected(ptEnd);
      end;
  end;
  StoreDeclarationEnd(vObject); 
end;

end.
