(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwPathUtils;

interface

uses
  Windows,
  SysUtils,
  Classes,
  IniFiles,
  Registry;

//returns in a string list the available dev tools installed on this machine.
//the Objects member of DirsList will point to a newly created stringlist,
//containing the list of search directories available from the environment settings
procedure GetDefaultDirs(const aProjectFileName: string; DirsList: TStrings);

procedure GetSubFolders(const aFolder: string; Folders: TStrings);

procedure ClearAndFreeItems(AStrings: TStrings);

function ResolvePathName(APath: string): string;

implementation

var
  gProductsDirs: TStringList = nil;

function ResolvePathName(APath: string): string;
var
  AVariable: string;
  AVarPath: string;
  j: Integer;
begin
  Result := APath;
  if (Length(Result) > 0) and (Result[1] = '$') then
  begin
    AVariable := Copy(Result, Pos('(', Result) + 1, Length(Result));
    AVariable := Copy(AVariable, 1, Pos(')', AVariable) - 1);

    // delete $(Delphi) from $(Delphi)\Whatever
    Delete(Result, 1, Length(AVariable) + 3);

    AVarPath := ExcludeTrailingPathDelimiter(SysUtils.GetEnvironmentVariable(AVariable));
    Result := AVarPath + Result;
  end;
end;

//returns in a string list the available dev tools installed on this machine.
//the Objects member of DirsList will point to a newly created stringlist,
//containing the list of search directories available from the environment settings
procedure GetDefaultDirs(const aProjectFileName: string; DirsList: TStrings);

  function GetProjectSearchPath: string;
  var
    DOFFile: TIniFile;
    ProjectOptionsName: string;
  begin
    ProjectOptionsName := ChangeFileExt(aProjectFileName, '.dof');
    if FileExists(ProjectOptionsName) then
    begin
      DOFFile := TIniFile.Create(ProjectOptionsName);
      try
        Result := DOFFile.ReadString('Directories', 'SearchPath', '');
      finally
        DOFFile.Free;
      end;
    end
    else
      Result := '';
  end;

  function GetStringKey(RootKey: HKEY; const aKey: string; const aValue: string): string;
  var
    Reg: TRegistry;
  begin
    Reg := TRegistry.Create;
    try
      Reg.RootKey := RootKey;
      if Reg.OpenKey(aKey, False) then
        Result := Reg.ReadString(aValue)
      else
        Result := '';
    finally
      Reg.Free;
    end;
  end;

  procedure AddEntry(const aStrings: TStrings; const aText: string);
  var
    vPath: string;
  begin
    vPath := IncludeTrailingPathDelimiter(aText);
    if aStrings.IndexOf(vPath) < 0 then
      aStrings.Add(vPath);
  end;

  procedure SetSemicolonText(aStrings: TStrings; const aText: string);
  var
    P: Integer;
    vText: string;
    s: string;
  begin
    vText := aText;
    P := Pos(';', vText);
    while P > 0 do
    begin
      s := Trim(Copy(vText, 1, P - 1));
      if Length(s) > 0 then
        AddEntry(aStrings, s);

      Delete(vText, 1, P);
      P := Pos(';', vText);
    end;
    s := Trim(vText);
    if s <> '' then
      AddEntry(aStrings, s);
  end;

  function IsDelphiInstalled(VerNo: Integer): Boolean;
  var
    aKey: string;
    KeyResult: HKEY;
  begin
    aKey := Format('Software\Borland\Delphi\%d.0\', [VerNo]);
    Result := (RegOpenKeyEx(HKEY_CURRENT_USER, PChar(aKey), 0,
      KEY_ALL_ACCESS, KeyResult) = ERROR_SUCCESS) and
      (KeyResult <> 0);
    if Result then
      RegCloseKey(KeyResult);
  end;

var
  i: Integer;
  s: string;
  aStrings: TStringList;
begin
  if gProductsDirs = nil then
  begin
    gProductsDirs := TStringList.Create;
    //Retrieve borland products.
    for i := 2 to 7 do
      if IsDelphiInstalled(i) then
      begin
        aStrings := TStringList.Create;
        s := GetStringKey(HKEY_CURRENT_USER,
          Format('Software\Borland\Delphi\%d.0\Library', [i]),
          'Browsing Path');
        if s <> '' then
          SetSemicolonText(aStrings, s);

        s := GetStringKey(HKEY_CURRENT_USER,
          Format('Software\Borland\Delphi\%d.0\Library', [i]),
          'SearchPath');
        if s <> '' then
          SetSemicolonText(aStrings, s);

        s := GetStringKey(HKEY_CURRENT_USER,
          Format('Software\Borland\Delphi\%d.0\Library', [i]),
          'Search Path');
        if s <> '' then
          SetSemicolonText(aStrings, s);

        if s <> '' then
          SetSemicolonText(aStrings, s);

        aStrings.Add(ExcludeTrailingPathDelimiter(GetCurrentDir) + '\CodeWrenchSystem');

        gProductsDirs.AddObject(Format('Delphi %d.0', [i]),
          aStrings);
      end;
  end;
  DirsList.Assign(gProductsDirs);
end;

procedure GetSubFolders(const aFolder: string; Folders: TStrings);
var
  sr: TSearchRec;
  sFolder: string;
  iResult: integer;
begin
  iResult := FindFirst(IncludeTrailingPathDelimiter(aFolder) + '*.*', faAnyFile, sr);
  if (iResult = 0) then
  try
    repeat
      if (((sr.Attr and faDirectory) = faDirectory) and (sr.Name <> '.') and (sr.Name <> '..')) then
      begin
        sFolder := IncludeTrailingPathDelimiter(aFolder) + sr.Name;
        if (sFolder <> aFolder) then
        begin
          Folders.Add(sFolder);
          GetSubFolders(sFolder, Folders);
        end;
      end;

      iResult := FindNext(sr);
    until (iResult <> 0);

  finally
    FindClose(sr);
  end;
end;

procedure ClearAndFreeItems(AStrings: TStrings);
var
  Index: Integer;
begin
  for Index := 0 to AStrings.Count - 1 do
    AStrings.Objects[Index].Free;
  AStrings.Clear;
end;

initialization

finalization

  if Assigned(gProductsDirs) then
    ClearAndFreeItems(gProductsDirs);
  gProductsDirs.Free;

end.

