(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwPreprocessImpl;

interface

uses
  Classes,
  ucwPreprocessIntf;

type
  TClUsedUnit = class(TInterfacedObject, IClUsedUnit)
  private
    FUnitName: string;
    FFullUnitName: string;

  protected
    function GetUnitName: string;
    function GetFullUnitName: string;
  end;

  TClUsedUnitList = class(TInterfacedObject, IClUsedUnitList)
  private
    FItems: TInterfaceList;

  protected
    function GetItem(aIndex: integer): IClUsedUnit;
    function Add(const aUnitName, aFullUnitName: string): IClUsedUnit;
    procedure Clear;
    function GetCount: integer;

  public
    constructor Create;
    destructor Destroy; override;
  end;

  TClPreProcessorInfo = class(TInterfacedObject, IClPreProcessorInfo)
  private
    FFilename: string;
    FUsedUnits: TClUsedUnitList;

  protected
    function GetUsedUnits: IClUsedUnitList;
    function GetUsedUnitCount: integer;
    function GetFullUnitName(const aUnitName: string): string;

    procedure SetFilename(const aValue: string);

  public
    constructor Create;
    destructor Destroy; override;

    property Filename: string
      read FFilename
      write SetFilename;
  end;

function clCreateUsedUnit(const aUnitName, aFullUnitName: string): IClUsedUnit;
function clCreatePreProcessorInfo(const aFilename: string): IClPreProcessorInfo;

implementation

uses
  SysUtils,
  ucwPreprocessor;

// ------------------------------------------------------------------------------------------

function TClUsedUnit.GetUnitName: string;
begin
  Result := FUnitName;
end;

function TClUsedUnit.GetFullUnitName: string;
begin
  Result := FFullUnitName;
end;

// ------------------------------------------------------------------------------------------

constructor TClUsedUnitList.Create;
begin
  inherited;

  FItems := TInterfaceList.Create;
end;

destructor TClUsedUnitList.Destroy;
begin
  FItems.Free;

  inherited;
end;

function TClUsedUnitList.GetItem(aIndex: integer): IClUsedUnit;
begin
  Result := IClUsedUnit(FItems[aIndex]);
end;

function TClUsedUnitList.Add(const aUnitName, aFullUnitName: string): IClUsedUnit;
begin
  FItems.Add(clCreateUsedUnit(aUnitName, aFullUnitName));
end;

procedure TClUsedUnitList.Clear;
begin
  FItems.Clear;
end;

function TClUsedUnitList.GetCount: integer;
begin
  Result := FItems.Count;
end;

// ------------------------------------------------------------------------------------------

constructor TClPreProcessorInfo.Create;
begin
  inherited;

  FUsedUnits := TClUsedUnitList.Create;
end;

destructor TClPreProcessorInfo.Destroy;
begin
  FUsedUnits.Free;

  inherited;
end;

function TClPreProcessorInfo.GetUsedUnits: IClUsedUnitList;
begin
  Result := (FUsedUnits as IClUsedUnitList);
end;

function TClPreProcessorInfo.GetUsedUnitCount: integer;
begin
  Result := FUsedUnits.GetCount;
end;

function TClPreProcessorInfo.GetFullUnitName(const aUnitName: string): string;
var
  i: integer;
  uu: IClUsedUnit;
begin
  for i:=0 to FUsedUnits.GetCount-1 do
  begin
    uu := FUsedUnits.GetItem(i);
    if SameText(uu.UnitName, aUnitName) then
    begin
      Result := uu.FullUnitName;
      EXIT;
    end;
  end;
end;

procedure TClPreProcessorInfo.SetFilename(const aValue: string);
var
  i: integer;
  slUsedUnits: TStringList;
  sUnitName, sFullUnitName: string;
begin
  // check input
  if (aValue = FFilename) then
    EXIT;

  // init result
  FFilename := aValue;
  FUsedUnits.Clear;
  if not FileExists(FFilename) then
    EXIT;

  // preprocess
  slUsedUnits := TStringList.Create;
  try
    PreProcessUsedUnits(FFilename, slUsedUnits);
    for i:=0 to slUsedUnits.Count-1 do
    begin
      sUnitName := slUsedUnits.Names[i];
      sFullUnitName := slUsedUnits.Values[sUnitName];
      FUsedUnits.Add(sUnitName, sFullUnitName);
    end;

  finally
    slUsedUnits.Free;
  end;
end;

// ------------------------------------------------------------------------------------------

function clCreateUsedUnit(const aUnitName, aFullUnitName: string): IClUsedUnit;
var
  uuResult: TClUsedUnit;
begin
  uuResult := TClUsedUnit.Create;

  uuResult.FUnitName := aUnitName;
  uuResult.FFullUnitName := aFullUnitName;

  Result := (uuResult as IClUsedUnit);
end;

function clCreatePreProcessorInfo(const aFilename: string): IClPreProcessorInfo;
var
  ppiResult: TClPreProcessorInfo;
begin
  ppiResult := TClPreProcessorInfo.Create;

  ppiResult.Filename := aFilename;

  Result := (ppiResult as IClPreProcessorInfo);
end;

// ------------------------------------------------------------------------------------------

end.

