(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwProject;

interface

uses
  SysUtils,
  Classes,
  Contnrs,
  ucwClasses,
  ucwPreprocessIntf,
  ucwAnalyzeParser,
  ucwPersistIntf,
  mwPasLexTypes;

type
  TcwProject = class(TInterfacedObject, IcwProject, IclPersist)
  private
    FParserList: TObjectList;
    FViewers: TInterfaceList;
    FProgressNotifiers: TInterfaceList;
    FProjectUnit: TcwUnit;
    FAllUnits: TcwUnitList;
    FCWOptions: IcwOptions;
    FReports: IcwReports;
    FEmptyClasses: TcwClassList;
    FEmptyObjects: TcwClassList;
    FEmptyInterfaces: TcwInterfaceList;
    FSystemSymbols: TcwSymbolList;
    FCanceled: Boolean;
    FPredefinedTypes: array[TcwTypeKind] of TcwSymbol;
    FFileNames: TcwFileNameList;
    FPreProcessorInfo: IClPreProcessorInfo;
    FMessages: IcwMessageList;
    FCodingConventions: IcwCodingConventions;
    FProjectOptions: IcwFileSettings;
    FRegularExpressions: IcwRegExpr;
    FFileSettings: IcwFileSettings;
    FSystemOptions: IcwSystemOptions;
    FSearchFiles: TStringList;
    procedure CreatePredefinedTypes;
    function CreateParser(const AFullFileName: string): TObject;
    procedure DoParse(const AUnit: TcwUnit; const aParseImpl: Boolean = True);
    procedure FullParse(const AUnit: TcwUnit);
  protected
    function GetProjectUnit: TcwUnit;
    function GetCanceled: Boolean;
    function GetUnparsedClasses: TcwClassList;
    function GetUnparsedObjects: TcwClassList;
    function GetUnparsedInterfaces: TcwInterfaceList;
    function GetUnparsedSymbols: TcwSymbolList;
    function GetAllUnits: TcwUnitList;
    function GetFileNames: TcwFileNameList;
    function GetCWOptions: IcwOptions;
    function GetReports: IcwReports;
    function GetMessages: IcwMessageList;
    function GetCodingConventions: IcwCodingConventions;
    function GetRegularExpressions: IcwRegExpr;
    function GetSystemOptions: IcwSystemOptions;
    function GetSystemFileSettings: IcwFileSettings;
    function GetProjectOptions: IcwFileSettings;
    function GetAsXML: widestring;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function AddUnit(const aUnitName: string; const aFullFileName: string = ''): TcwUnit;
    procedure Run(aParseAllImplementations: Boolean = True);
    procedure ParseInterface(const AUnit: TcwUnit);
    //cancel will return true if the project pasring has been cancelelled from a progress notifier
    property Canceled: Boolean read GetCanceled;
    //"library" classes - classes for which source code was not available for parsing
    property UnparsedClasses: TcwClassList read GetUnparsedClasses;
    //"library" objects - objects for which source code was not available for parsing
    property UnparsedObjects: TcwClassList read GetUnparsedObjects;
    //"library" interfaces - interfaces for which source code was not available for parsing
    property UnparsedInterfaces: TcwInterfaceList read GetUnparsedInterfaces;
    //returns an interface to a pre-defined type
    function GetPredefinedType(const aTypeKind: TcwTypeKind): TcwSymbol; overload;
    function GetPredefinedType(ATypeName: string): TcwSymbol; overload;
    //"system" symbols - symbols that have not been part of the parsing
    property UnparsedSymbols: TcwSymbolList read GetUnparsedSymbols;
    //all units used by the project
    property AllUnits: TcwUnitList read GetAllUnits;
    property FileNames: TcwFileNameList read GetFileNames;
    //options for the project
    property CWOptions: IcwOptions read GetCWOptions;
    //create a new CWSettings instance, to be used by reports and other plugins
    function CreateCWSettings(const aName: string): IcwSettings;
    //parsing proress notification
    procedure OnParseUnit(const aFullFileName: string; ParseState: TParserState);
    //from a unit name, will attempt to return the full file name
    //will search for the existance of the file as pecified in the search path
    //if it can not find a source file on the path, will return an empty string
    function ResolveUnitName(const aUnitName: string; const ARelativeToPath: string;
      out ADirectoryType: TDirectoryType): string;
    // shows the source code for a location
    procedure ShowSourceFile(const aFileName: string; aX, aY: Integer);
    // shows the source code for a location
    procedure ShowLocation(const aLocation: TcwCodeLocation);
    // registers a new code viewer
    procedure AddCodeViewer(const aViewer: IcwCodeViewer);
    // removes a code viewer
    procedure RemoveCodeViewer(const aViewer: IcwCodeViewer);
    // registers a progress notifier
    procedure AddProgressNotifier(const aNotifier: IcwProgressNotifier);
    // registers a progress notifier
    procedure RemoveProgressNotifier(const aNotifier: IcwProgressNotifier);
    property Reports: IcwReports read GetReports;
    property Messages: IcwMessageList read GetMessages;
    //coding conventions settings
    property CodingConventions: IcwCodingConventions read GetCodingConventions;
    property RegularExpressions: IcwRegExpr read GetRegularExpressions;
    property SystemOptions: IcwSystemOptions read GetSystemOptions;
    property SystemFileSettings: IcwFileSettings read GetSystemFileSettings;
    //interface to read and save options for the current project
    property ProjectOptions: IcwFileSettings read GetProjectOptions;
    // IclPersist
    function Load(const aLocation: widestring): Boolean;
    procedure Save(const aLocation: widestring);
    procedure IclPersist.Clear = ClearPersist;
    procedure ClearPersist;
    // IclXMLPersist
    procedure SetAsXML(const aValue: widestring);
    property ProjectUnit: TcwUnit read GetProjectUnit;
  end;

implementation

uses
  ucwPathUtils,
  ucwMessageList,
  ucwOptions,
  ucwReports,
  ucwStringList,
  ucwSettings,
  ucwCodingConventions,
  ucwRegExpr,
  ucwFileSettings,
  ucwSystemOptions;

type
  TcwSearchFile = class
  private
    FFullFileName: string;
    FDirectoryType: TDirectoryType;
  public
    constructor Create(AFullFileName: string; ADirectoryType: TDirectoryType);
    property FullFileName: string read FFullFileName;
    property DirectoryType: TDirectoryType read FDirectoryType;
  end;

{ TcwSearchFile }

constructor TcwSearchFile.Create(AFullFileName: string; ADirectoryType: TDirectoryType);
begin
  inherited Create;
  FFullFileName := AFullFileName;
  FDirectoryType := ADirectoryType;
end;

{ TcwProject }

constructor TcwProject.Create;
var
  ApplicationPath: string;
begin
  inherited;

  FProjectUnit := TcwUnit.Create(-1, 'TcwProject');

  FViewers := TInterfaceList.Create;
  FProgressNotifiers := TInterfaceList.Create;

  ApplicationPath := IncludeTrailingPathDelimiter(ExtractFilePath(paramstr(0)));

  FProjectOptions := TcwFileSettings.Create(ApplicationPath + 'projectoptions.ini');

  FAllUnits := TcwUnitList.Create;
  FCWOptions := TcwOptions.Create;
  FMessages := TcwMessageList.Create;
  FCodingConventions := TcwCodingConventions.Create(ApplicationPath + 'codeconventions.cw');
  FCodingConventions.Load;
  FEmptyClasses := TcwOwnedClassList.Create(nil);
  FEmptyObjects := TcwOwnedObjectList.Create(nil);
  FEmptyInterfaces := TcwOwnedInterfaceList.Create(nil);
  FSystemSymbols := TcwSymbolList.Create;
  FFileNames := TcwFileNameList.Create;
  FReports := TcwReports.Create;
  FRegularExpressions := TcwRegExpr.Create;

  FFileSettings := TcwFileSettings.Create(ApplicationPath + 'options.cw');

  FSystemOptions := TcwSystemOptions.Create;
  FSystemOptions.Load(FFileSettings);

  FSearchFiles := TStringList.Create;
  FSearchFiles.Sorted := True;
  FParserList := TObjectList.Create;
end;

destructor TcwProject.Destroy;
begin
  FProjectUnit.Free;
  FAllUnits.Free;
  FFileNames.Free;
  FSystemSymbols.Free;
  ClearAndFreeItems(FSearchFiles);
  FSearchFiles.Free;
  FreeAndNil(FViewers);
  FreeAndNil(FProgressNotifiers);
  FEmptyClasses.Free;
  FEmptyObjects.Free;
  FEmptyInterfaces.Free;
  FParserList.Free;
  inherited;
end;

procedure TcwProject.CreatePredefinedTypes;

  function AddType(const aTypeName: string; aTypeKind: TcwTypeKind): TcwType;
  begin
    Result := TcwType.Create(aTypeName, aTypeKind, vkNone, 0, 0, 0, False);
    FSystemSymbols.AddSymbol(Result);
  end;

  function AddArray(const aTypeName: string; aTypeKind: TcwTypeKind): TcwSymbol;
  begin
    Result := TcwArray.Create(aTypeName, vkNone, 0, 0, 0);
    FSystemSymbols.AddSymbol(Result);
  end;

begin
  //INTEGER
  FPredefinedTypes[tkBoolean] := AddType('Boolean', tkBoolean);
  FPredefinedTypes[tkByte] := AddType('Byte', tkByte);
  FPredefinedTypes[tkBytebool] := AddType('ByteBool', tkBytebool);
  FPredefinedTypes[tkCardinal] := AddType('Cardinal', tkCardinal);
  FPredefinedTypes[tkChar] := AddType('Char', tkChar);
  FPredefinedTypes[tkDWord] := AddType('DWord', tkDWord);
  FPredefinedTypes[tkInt64] := AddType('Int64', tkInt64);
  FPredefinedTypes[tkInteger] := AddType('Integer', tkInteger);
  FPredefinedTypes[tkLongBool] := AddType('LongBool', tkLongBool);
  FPredefinedTypes[tkLongInt] := AddType('LongInt', tkLongInt);
  FPredefinedTypes[tkLongWord] := AddType('LongWord', tkLongWord);
  FPredefinedTypes[tkPChar] := AddType('PChar', tkPChar);
  FPredefinedTypes[tkShortInt] := AddType('ShortInt', tkShortInt);
  FPredefinedTypes[tkSmallInt] := AddType('SmallInt', tkSmallInt);
  FPredefinedTypes[tkWideChar] := AddType('WideChar', tkWideChar);
  FPredefinedTypes[tkWord] := AddType('Word', tkWord);
  FPredefinedTypes[tkWordbool] := AddType('WordBool', tkWordbool);
  //FLOAT
  FPredefinedTypes[tkReal48] := AddType('Real48', tkReal48);
  FPredefinedTypes[tkReal] := AddType('Real', tkReal);
  FPredefinedTypes[tkSingle] := AddType('Single', tkSingle);
  FPredefinedTypes[tkDouble] := AddType('Double', tkDouble);
  FPredefinedTypes[tkExtended] := AddType('Extended', tkExtended);
  FPredefinedTypes[tkCurrency] := AddType('Currency', tkCurrency);
  FPredefinedTypes[tkComp] := AddType('Comp', tkComp);

  //STRING
  FPredefinedTypes[tkString] := AddType('String', tkString);
  FPredefinedTypes[tkAnsiString] := AddType('AnsiString', tkAnsiString);
  FPredefinedTypes[tkShortString] := AddType('ShortString', tkShortString);
  FPredefinedTypes[tkWideString] := AddType('WideString', tkWideString);

  //VARIANT
  FPredefinedTypes[tkOleVariant] := AddType('OleVariant', tkOleVariant);
  FPredefinedTypes[tkVariant] := AddType('Variant', tkVariant);

  FPredefinedTypes[tkArrayOfConst] := AddArray('Array of Const', tkArrayOfConst);
  FPredefinedTypes[tkArray] := AddArray('Array', tkArray);
  FPredefinedTypes[tkFile] := AddType('File', tkFile);
  FPredefinedTypes[tkPacked] := AddType('Packed', tkPacked);
  FPredefinedTypes[tkRecord] := AddType('Record', tkRecord);
  FPredefinedTypes[tkSet] := AddType('Set', tkSet);
  FPredefinedTypes[tkRoutine] := AddType('Routine', tkRoutine);
  FPredefinedTypes[tkPointer] := AddType('Pointer', tkPointer);
  FPredefinedTypes[tkClass] := AddType('Class', tkClass);
  FPredefinedTypes[tkInterface] := AddType('Interface', tkInterface);

  FPredefinedTypes[tkProcedure] := AddType('procedure', tkProcedure);
  FPredefinedTypes[tkFunction] := AddType('function', tkFunction);
  FPredefinedTypes[tkText] := AddType('Text', tkText);
  FPredefinedTypes[tkTextFile] := AddType('TextFile', tkTextFile);
end;

procedure TcwProject.Clear;
begin
  FParserList.Clear;
  FAllUnits.Clear;
  FProjectUnit.UsedUnits.Clear;
  Messages.Clear;
  Reports.Clear;
  FEmptyClasses.Clear;
  FEmptyObjects.Clear;
  FEmptyInterfaces.Clear;
  FMessages.Clear;
  FFileNames.Clear;
  Reports.Clear;
  ClearAndFreeItems(FSearchFiles);
end;

procedure TcwProject.Run(aParseAllImplementations: Boolean = True);
var
  AUsedUnit: TcwUsedUnit;
  i: integer;
  sCurrentDir: string;
  AUnit: TcwUnit;
begin
  if not FileExists(CWOptions.FileToAnalyze) then
    Exit;
  Clear;
  CreatePredefinedTypes;

  FCanceled := False;
  //move to the project dir, to resolve all relative paths
  sCurrentDir := GetCurrentDir;
  try
    SetCurrentDir(ExtractFilePath(CWOptions.FileToAnalyze));

    // preprocess main unit
    FPreProcessorInfo := clPreProcessorInfoFactory(CWOptions.FileToAnalyze);

    AUsedUnit := TcwUsedUnit.Create(AddUnit(ChangeFileExt(ExtractFileName(
      CWOptions.FileToAnalyze), ''), CWOptions.FileToAnalyze));
    ParseInterface(AUsedUnit.UnitIntf);

    ProjectUnit.UsedUnits.AddUsedUnit(AUsedUnit);

    if aParseAllImplementations then
    begin
      i := 0;
      //finish parsing all units that have had only their interface section parsed
      while i <= AllUnits.Count - 1 do
      begin
        AUnit := AllUnits[i];
        if AUnit.Parsed = psrInterfaceParsed then
          FullParse(AUnit);
        Inc(i);
      end;
    end else
      FullParse(AllUnits[0]);
  finally
    SetCurrentDir(sCurrentDir);                  
  end;
end;

function TcwProject.CreateParser(const AFullFileName: string): TObject;
var
  vDefines: TmwAssignAbleHash;
begin
  vDefines := TmwAssignAbleHash.Create(CWOptions.Defines.Text);
  //TODO add codewrench define
  try
    Result := TAnalyzeParser.Create(AFullFileName, vDefines, GlobalKeyHashTable,
      GlobalDirectiveHash);
  finally
    vDefines.Free;
  end;
end;

function TcwProject.AddUnit(const aUnitName: string; const aFullFileName: string = ''): TcwUnit;
var
  AFileExists: Boolean;
  AFileName: string;
  aUnitType: TDirectoryType;
  AFileIndex: Integer;
begin
  if aFullFileName = '' then
    AFileName := ResolveUnitName(aUnitName, '', aUnitType)
  else
    AFileName := ResolveUnitName(aFullFileName, '', aUnitType);

  AFileExists := AFileName <> '';
  if AFileExists then
  begin
    FParserList.Add(CreateParser(AFileName));
    AFileIndex := FileNames.Add(AFileName, AFileExists, AllUnits.Count)
  end else
  begin
    FParserList.Add(nil);
    AFileIndex := FileNames.Add(aUnitName, AFileExists, AllUnits.Count);
  end;

  Result := TcwUnit.Create(AFileIndex, aUnitName);
  AllUnits.AddUnit(Result);

  Result.UnitType := aUnitType;
end;

procedure TcwProject.DoParse(const AUnit: TcwUnit; const aParseImpl: Boolean = True);
begin
  if Length(AUnit.FullFileName) = 0 then
    Exit;
  if Canceled then
    Exit;

  if AUnit.Parsed = psrInterfaceParsing then
  begin
    OnParseUnit(AUnit.FullFileName, psrInterfaceParsing);
    (FParserList[AUnit.UnitIndex] as TAnalyzeParser).ParseInterface(AUnit);
  end;

  if aParseImpl then
  begin
    OnParseUnit(AUnit.FullFileName, psrFull);
    (FParserList[AUnit.UnitIndex] as TAnalyzeParser).ParseImplementation;
    FParserList[AUnit.UnitIndex] := nil;
    AUnit.Parsed := psrFull;
    if AUnit.DFMFileName <> '' then
      AUnit.DFMResource.ReadResource(AUnit, AUnit.DFMFileName);
  end;
end;

procedure TcwProject.FullParse(const AUnit: TcwUnit);
begin
  if AUnit.Parsed <> psrFull then
  begin
    AUnit.Parsed := psrFull;
    DoParse(AUnit, True);
  end;
end;

procedure TcwProject.ParseInterface(const AUnit: TcwUnit);
begin
  if AUnit.Parsed = psrNone then
  begin
    AUnit.Parsed := psrInterfaceParsing;
    DoParse(AUnit, False);
    AUnit.Parsed := psrInterfaceParsed;
  end;
end;

//cancel will return true if the project pasring has been cancelelled from a progress notifier
function TcwProject.GetCanceled: Boolean;
begin
  Result := FCanceled;
end;

function TcwProject.GetMessages: IcwMessageList;
begin
  Result := FMessages;
end;

//coding conventions settings
function TcwProject.GetCodingConventions: IcwCodingConventions;
begin
  Result := FCodingConventions;
end;

function TcwProject.GetRegularExpressions: IcwRegExpr;
begin
  Result := FRegularExpressions;
end;

//interface to read and save options for the current project
function TcwProject.GetProjectOptions: IcwFileSettings;
begin
  Result := FProjectOptions;
end;

function TcwProject.GetUnparsedClasses: TcwClassList;
begin
  Result := FEmptyClasses;
end;

function TcwProject.GetUnparsedObjects: TcwClassList;
begin
  Result := FEmptyObjects;
end;

//"library" interfaces - interfaces for which source code was not available for parsing
function TcwProject.GetUnparsedInterfaces: TcwInterfaceList;
begin
  Result := FEmptyInterfaces;
end;

//returns an interface to a pre-defined type
function TcwProject.GetPredefinedType(ATypeName: string): TcwSymbol;
begin
  Result := FSystemSymbols.Find(ATypeName) as TcwSymbol;
end;

function TcwProject.GetPredefinedType(const aTypeKind: TcwTypeKind): TcwSymbol;
begin
  Result := FPredefinedTypes[aTypeKind];
end;

//"system" symbols - symbols that have not been part of the parsing
function TcwProject.GetUnparsedSymbols: TcwSymbolList;
begin
  Result := FSystemSymbols;
end;

//all units used by the project
function TcwProject.GetAllUnits: TcwUnitList;
begin
  Result := FAllUnits;
end;

function TcwProject.GetFileNames: TcwFileNameList;
begin
  Result := FFileNames;
end;

//options for the project
function TcwProject.GetCWOptions: IcwOptions;
begin
  Result := FCWOptions;
end;

//create a new CWSettings instance, to be used by reports and other plugins
function TcwProject.CreateCWSettings(const aName: string): IcwSettings;
begin
  Result := TcwSettings.Create(aName);
end;

//parsing proress notification
procedure TcwProject.OnParseUnit(const aFullFileName: string; ParseState: TParserState);
var
  i: Integer;
begin
  for i := 0 to FProgressNotifiers.Count - 1 do
  begin
    IcwProgressNotifier(FProgressNotifiers[i]).ShowProgress(aFullFileName, ParseState, FCanceled);
    if FCanceled then
      Break;
  end;
end;

//from a unit name, will attempt to return the full file name
//will search for the existance of the file as pecified in the search path
//if it can not find a source file on the path, will return an empty string
function TcwProject.ResolveUnitName(const aUnitName: string; const ARelativeToPath: string;
  out ADirectoryType: TDirectoryType): string;

  procedure LoadSearchFiles(ADir: string; ADirectoryType: TDirectoryType);
  var
    AFullFileName: string;
    ASearchRec: TSearchRec;
    AFileName: string;
  begin
    ADir := ResolvePathName(ADir);
    if FindFirst(IncludeTrailingPathDelimiter(ADir) + '*.*', faAnyFile, ASearchRec) = 0 then
    begin
      try
        repeat
          if (((ASearchRec.Attr and faDirectory) = 0) and
            (ASearchRec.Name <> '.') and (ASearchRec.Name <> '..')) then
          begin
            AFullFileName := IncludeTrailingPathDelimiter(ADir) + ASearchRec.Name;
            //TODO: ignoring case sensitivity with file names...
            AFileName := LowerCase(ASearchRec.Name);
            if FSearchFiles.IndexOf(AFileName) = -1 then
              FSearchFiles.AddObject(AFileName, TcwSearchFile.Create(AFullFileName,
                ADirectoryType));
          end;
        until FindNext(ASearchRec) <> 0;
      finally
        SysUtils.FindClose(ASearchRec);
      end;
    end;
  end;

  function FileExistsOrExisted(AFullFileName: string): Boolean;
  begin
    if Assigned(FileNames.Find(AFullFileName)) then
      Result := True
    else
      Result := FileExists(AFullFileName);
  end;

var
  AIndex: Integer;
  AAliasStr: string;
  APos: Integer;
  AEqualPos: Integer;
  AUnitIndex: Integer;
  ASaveDir: string;
  i: integer;
  sPath, sExt, sProjectPath: string;
begin
  // init result
  Result := aUnitName;

  // account for aliases
  for AUnitIndex := 0 to CWOptions.Aliases.Count - 1 do
  begin
    AAliasStr := Trim(CWOptions.Aliases[AUnitIndex]);
    APos := Pos(UpperCase(Result), UpperCase(AAliasStr));
    if APos = 1 then
    begin
      AEqualPos := Pos('=', CWOptions.Aliases[AUnitIndex]);
      if AEqualPos > 0 then
        if SameText(Result, Trim(Copy(AAliasStr, 1, AEqualPos - 1))) then
        begin
          Result := Trim(Copy(AAliasStr, AEqualPos + 1, Length(AAliasStr)));
          Break;
        end;
    end;
  end;

  // get extension
  sExt := ExtractFileExt(Result);
  if sExt = '' then
    Result := ChangeFileExt(Result, '.pas');

  // get path
  sPath := ExtractFilePath(Result);
  ADirectoryType := dtProject;

  // find path if no path given
  if sPath = '' then
  begin
    if FSearchFiles.Count = 0 then
    begin
      sProjectPath := ExtractFileDir(CWOptions.FileToAnalyze);
      LoadSearchFiles(sProjectPath, dtProject);
      for i := 0 to CWOptions.ProjectSearchPath.Count - 1 do
        LoadSearchFiles(CWOptions.ProjectSearchPath[i], dtProject);
      for i := 0 to SystemOptions.LibrarySearchPath.Count - 1 do
        LoadSearchFiles(SystemOptions.LibrarySearchPath[i], dtLibrary);
    end;
    AIndex := FSearchFiles.IndexOf(Result);
    if AIndex >= 0 then
    begin
      Result := TcwSearchFile(FSearchFiles.Objects[AIndex]).FullFileName;
      ADirectoryType := TcwSearchFile(FSearchFiles.Objects[AIndex]).DirectoryType;
    end else
      Result := '';
  end else
  begin
    if ARelativeToPath <> '' then
    begin
      ASaveDir := GetCurrentDir;
      SetCurrentDir(ARelativeToPath);
    end;
    try
      if FileExistsOrExisted(Result) then
        Result := ExpandFileName(Result)
      else
        Result := '';
    finally
      if ARelativeToPath <> '' then
        SetCurrentDir(ASaveDir);
    end;
  end;
end;

procedure TcwProject.ShowSourceFile(const aFileName: string; aX,
  aY: Integer);
var
  i: Integer;
begin
  for i := 0 to FViewers.Count - 1 do
    IcwCodeViewer(FViewers[i]).ShowSourceFile(aFileName, aX, aY);
end;

// shows the source code for a location
procedure TcwProject.ShowLocation(const aLocation: TcwCodeLocation);
var
  sFileName: string;
var
  i: Integer;
begin
  if aLocation <> nil then
  begin
    sFileName := aLocation.FileName;
    for i := 0 to FViewers.Count - 1 do
      ShowSourceFile(sFileName, aLocation.X, aLocation.Y);
  end;
end;

procedure TcwProject.AddCodeViewer(const aViewer: IcwCodeViewer);
begin
  if aViewer <> nil then
    if FViewers.IndexOf(aViewer) < 0 then
      FViewers.Add(aViewer);
end;

procedure TcwProject.RemoveCodeViewer(const aViewer: IcwCodeViewer);
begin
  FViewers.Remove(aViewer);
end;

// registers a progress notifier
procedure TcwProject.AddProgressNotifier(const aNotifier: IcwProgressNotifier);
begin
  if aNotifier <> nil then
    if FProgressNotifiers.IndexOf(aNotifier) < 0 then
      FProgressNotifiers.Add(aNotifier);

end;

// registers a progress notifier
procedure TcwProject.RemoveProgressNotifier(const aNotifier: IcwProgressNotifier);
begin
  FProgressNotifiers.Remove(aNotifier);
end;

function TcwProject.GetReports: IcwReports;
begin
  Result := FReports;
end;

function TcwProject.Load(const aLocation: widestring): Boolean;
var
  i: integer;
  sFileName: string;
begin
  Result := FProjectOptions.Load(aLocation);
  if not Result then
    Exit;

  CWOptions.Load(FProjectOptions);

  for i := 0 to Reports.Count - 1 do
    Reports[i].Settings.LoadSettings(FProjectOptions);

  sFileName := CWOptions.FileToAnalyze;

  AddUnit(ChangeFileExt(ExtractFileName(sFileName), ''), sFileName);
end;

procedure TcwProject.Save(const aLocation: widestring);
var
  i: integer;
  VisibleReports: IcwStringList;
begin
  // save visible reports
  VisibleReports := TcwStringList.Create;
  for i := 0 to FReports.Count - 1 do
    if FReports[i].Visible then
      VisibleReports.Add(FReports[i].Name);
  ProjectOptions.SaveSection('VisibleReports', VisibleReports);

  // save options
  CWOptions.Save(FProjectOptions);
  FProjectOptions.Save(aLocation);

  FSystemOptions.Save(FFileSettings);
end;

procedure TcwProject.ClearPersist;
begin
  Clear;
  CWOptions.Clear;
end;

function TcwProject.GetAsXML: widestring;
begin

end;

procedure TcwProject.SetAsXML(const aValue: widestring);
begin

end;

function TcwProject.GetSystemOptions: IcwSystemOptions;
begin
  Result := FSystemOptions;
end;

function TcwProject.GetSystemFileSettings: IcwFileSettings;
begin
  Result := FFileSettings;
end;

function TcwProject.GetProjectUnit: TcwUnit;
begin
  Result := FProjectUnit;
end;

end.

