(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwReport;

interface
uses
  SysUtils, Classes, ucwClasses;

type
  TcwReport = class(TInterfacedObject, IcwReport)
  private
    FColumns: IcwColumns;
    FData: IcwReportData;
    FSettings: IcwSettings;
    FVisible: boolean;

  public
    constructor Create;
    //abstract routine to create columns - must be overriden
    procedure   CreateColumns;virtual;abstract;

    procedure CreateOptions;virtual;
      //report name
    function GetName : string;virtual;abstract;
    property Name : string read GetName;

    //options/settings for the report
    function GetSettings : IcwSettings;
    property Settings : IcwSettings read GetSettings;

     //if the report can be sorted by any column, the Sortable prop will return true
    //some reports make no sense to be sortable (like code metrics, which are line-by-line)
    //and for such non-srtable reports, this property will return false;
    //by default, the reports are sortable
    //to implement a non-sortable report, one needs to override GetSortable and return False
    function GetSortable : Boolean;virtual;
    property Sortable : Boolean read GetSortable;

    //if the report has "Location" details, HasDetails will return true
    //in this case, for each row of data, there is a corresponding UsedBy list
    function GetHasDetails : Boolean;virtual;
    property HasDetails : Boolean read GetHasDetails;

    //columns
    function GetColumns : IcwColumns;
    property Columns : IcwColumns read GetColumns;

    //data rows
    function GetData : IcwReportData;
    property Data : IcwReportData read GetData;

    //visible
    function GetVisible: boolean;
    procedure SetVisible(Value: boolean);

    //clear and run the report
    procedure DoExecute;virtual;

    //return the report in a CSV Format
    //if aIncludeHeaders = True, the CSV report will include the field names as first line
    //if aQuotedValues = True, the values will enclides in double quotes
    function  GetAsCSV(const aIncludeHeaders : Boolean; aQuotedValues : Boolean) : IcwStringList;

    //run the report, so it will be filled with data
    procedure Execute;


    //clear the report
    procedure Clear;
  end;

implementation

uses
  Variants,  ucwColumns, ucwReportData, ucwStringList ;

constructor TcwReport.Create;
begin
  inherited Create;
  FColumns := TcwColumns.Create;
  FData    := TcwReportData.Create;
  FSettings := gcwProject.CreateCWSettings(Name);
  CreateOptions;
  CreateColumns;
end;

procedure TcwReport.CreateOptions;
begin
end;

//options/settings for the report
function TcwReport.GetSettings : IcwSettings;
begin
  Result := FSettings;
end;

function TcwReport.GetSortable : Boolean;
begin
  Result := True;
end;

//if the report has "Location" details, HasDetails will return true
//in this case, for each row of data, there is a corresponding UsedBy list
function TcwReport.GetHasDetails : Boolean;
begin
  Result := False;
end;

//columns
function TcwReport.GetColumns : IcwColumns;
begin
  Result := FColumns;
end;


//data rows
function TcwReport.GetData : IcwReportData;
begin
  Result := FData;
end;

function TcwReport.GetVisible: boolean;
begin
  Result := FVisible;
end;

procedure TcwReport.SetVisible(Value: boolean);
begin
  FVisible := Value;
end;

procedure TcwReport.DoExecute;
begin

end;

//return the report in a CSV Format
//if aIncludeHeaders = True, the CSV report will include the field names as first line
//if aQuotedValues = True, the values will enclides in double quotes
function  TcwReport.GetAsCSV(const aIncludeHeaders : Boolean; aQuotedValues : Boolean) : IcwStringList;

  function GetVal (const aVal : string) : string;
  begin
    if aQuotedValues then
      Result := QuotedStr(aVal) else
        Result := aVal;
  end;

var i, nRow : Integer;
var s : string;
begin
  Result := TcwStringList.Create;
  if aIncludeHeaders then
  begin
    s := '';
    for i := 0 to Columns.Count - 1 do
    begin
      s := s + GetVal(Columns[i].Name);
      if i < Columns.Count - 1 then
        s := s + ',';
    end;
    Result.Add(s);
  end;

  for nRow := 0 to Data.Count - 1 do
  begin
    s := ''; 
    for i := 0 to Columns.Count - 1 do
    begin
      s := s + GetVal(VarToStr(Data[nRow, i]));
      if i < Columns.Count - 1 then
        s := s + ',';
    end;
    Result.Add(s);    
  end;
end;

//run the report, so it will be filled with data
//clear and run the report
procedure TcwReport.Execute;
begin
  Clear;
  DoExecute;
end;

//clear the report
procedure TcwReport.Clear;
begin
  FData.Clear;
end;

end.
