(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwRoutineParser;

interface

uses
  Types,
  Classes,
  ucwClasses,
  ucwStructuredTypeParser,
  mwSimplePasPar;

const
  // following is the "correct" typing of the result variable and is used in Lint report
  cResultVariableName = 'Result';

type
  TRoutineParser = class(TStructuredTypeParser)
  private
    function IsOverloadedInAncestors(ARoutineName: string; AClass: TcwClass = nil): Boolean;
  protected
    function ParseReturnType: TcwSymbol;
    function ParseProceduralDirective: TRoutineDirective;
    function ParseOldFormalParameterType: TcwSymbol;
    function ParseNewFormalParameterType(out APredefinedArrayType: TcwSymbol): TcwSymbol;
    procedure ParseParameterNameList(const ARoutine: TcwRoutine; const aParamKind: TParamKind;
      const AAddParameters: Boolean; const ASymbolList: TcwSymbolList);
    procedure ParseConstParameter(const ARoutine: TcwRoutine; const AAddParameters: Boolean;
      const ASymbolList: TcwSymbolList);
    procedure ParseVarParameter(const ARoutine: TcwRoutine; const AAddParameters: Boolean;
      const ASymbolList: TcwSymbolList);
    procedure ParseOutParameter(const ARoutine: TcwRoutine; const AAddParameters: Boolean;
      const ASymbolList: TcwSymbolList);
    procedure ParseParameterFormal(const ARoutine: TcwRoutine; const aParamKind: TParamKind;
      const AAddParameters: Boolean; const ASymbolList: TcwSymbolList);
    procedure ParseFormalParameterSection(var ARoutine: TcwRoutine; const AAddParameters: Boolean;
      const ASymbolList: TcwSymbolList);
    procedure ParseFormalParameterList(var ARoutine: TcwRoutine; const AAddParameters: Boolean;
      AFindOverloadedRoutine: Boolean);
    function ParseFunctionHeading(const AContainer: TcwVarContainer):
      TcwRoutine;
    function ParseProcedureHeading(const AContainer: TcwVarContainer):
      TcwRoutine;
    procedure ExportedHeading; override;
    procedure ParseFunctionProcedureBlock(const ARoutine: TcwRoutine);
    function ParseFunctionProcedureMethodDeclaration(const AContainer: TcwVarContainer): TcwRoutine;
    function ParseProcedureDeclarationSection(const AContainer: TcwVarContainer): TcwRoutine;
    procedure AddReturnType(const ARoutine: TcwRoutine);

    procedure StoreImplementationStart(const aRoutine: TcwRoutine; const aLocation: TPoint);
    procedure StoreImplementationEnd(const aRoutine: TcwRoutine; const aLocation: TPoint);
  public
  end;

implementation

uses
  SysUtils,
  mwPasLexTypes,
  mwSimplePasParTypes,
  mwPasLex,
  ucwUnitParser;

function TRoutineParser.ParseReturnType: TcwSymbol;
begin
  case TokenID of
    ptString: begin
      Result := gcwProject.GetPredefinedType(tkString);
      ParseStringType;
    end;
    else
      Result := ParseTypeId;
  end;
end;

procedure TRoutineParser.AddReturnType(const ARoutine: TcwRoutine);
var
  vResultVar: TcwVariable;
  vRoutineResultVar: TcwVariable;
  vReturnType: TcwSymbol;
begin
  vReturnType := ParseReturnType;
  if vReturnType <> nil then
  begin
    vResultVar := TcwVariable.Create(cResultVariableName, vkPublic, False,
      FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
    vResultVar.TypeIntf := vReturnType;
    ARoutine.ReturnType := vResultVar;
    ARoutine.AddSymbol(vResultVar);

    vRoutineResultVar := TcwVariable.Create(ARoutine.Name, vkPublic, False,
      FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
    vRoutineResultVar.TypeIntf := vReturnType;
    ARoutine.RoutineType := vRoutineResultVar;
    ARoutine.AddSymbol(vRoutineResultVar);
  end;
end;

function TRoutineParser.ParseProceduralDirective: TRoutineDirective;
begin
  case ExID of
    ptAbstract: Result := rdAbstract;
    ptCdecl: Result := rdCdecl;
    ptPascal: Result := rdPascal;
    ptRegister: Result := rdRegister;
    ptAssembler: Result := rdAssembler;
    ptSafeCall: Result := rdSafeCall;
    ptStdCall: Result := rdStdCall;
    ptExport: Result := rdExport;
    ptFar: Result := rdFar;
    ptNear: Result := rdNear;
    ptExternal: Result := rdExternal;
    ptDynamic: Result := rdDynamic;
    ptMessage: Result := rdMessage;
    ptOverload: Result := rdOverload;
    ptOverride: Result := rdOverride;
    ptReintroduce: Result := rdReintroduce;
    ptVirtual: Result := rdVirtual;
    ptLocal: Result := rdLocal;
    else
      Result := rdNone;
  end;

  inherited ProceduralDirective
end;

function TRoutineParser.ParseOldFormalParameterType: TcwSymbol;
begin
  case TokenID of
    ptString: begin
      Result := gcwProject.GetPredefinedType(tkString);
      NextToken;
    end;
    else
      Result := ParseTypeId;
  end;
end;

function TRoutineParser.ParseNewFormalParameterType(out APredefinedArrayType: TcwSymbol): TcwSymbol;
begin
  APredefinedArrayType := nil;
  case TokenID of
    ptArray: begin
      Result := gcwProject.GetPredefinedType(tkArray);
      NextToken;
      Expected(ptOf);
      case TokenID of
        (*new in ObjectPascal80*)
        ptConst: begin
          Result := gcwProject.GetPredefinedType(tkArrayOfConst);
          NextToken;
        end;
        else
          APredefinedArrayType := ParseOldFormalParameterType;
      end;
    end;
    else
      Result := ParseOldFormalParameterType;
  end;
end;

procedure TRoutineParser.ParseParameterNameList(const ARoutine: TcwRoutine;
  const aParamKind: TParamKind; const AAddParameters: Boolean; const ASymbolList: TcwSymbolList);

  procedure ParseParameterName;
  begin
    if AAddParameters then
      ARoutine.Parameters.AddParameter(Lexer.Token, aParamKind, FileIndex,
        Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
    ASymbolList.AddSymbol(Lexer.Token, vkNone, FileIndex, Lexer.PosXY.X,
      Lexer.PosXY.Y);
    ParameterName;
  end;

var
  ACount: Integer;
  Index: Integer;
begin
  ACount := ARoutine.Parameters.Count;
  ParseParameterName;
  while TokenID = ptComma do
  begin
    NextToken;
    ParseParameterName;
  end;
  if AAddParameters and (TokenID <> ptColon) then
    for Index := ACount to ARoutine.Parameters.Count - 1 do
      ARoutine.Parameters[Index].UntypedParameter := True;
end;

procedure TRoutineParser.ParseConstParameter(const ARoutine: TcwRoutine;
  const AAddParameters: Boolean; const ASymbolList: TcwSymbolList);
var
  APredefinedArrayType: TcwSymbol;
  ATypeCount: Integer;
  ACount: Integer;
  i: Integer;
  vType: TcwSymbol;
begin
  Expected(ptConst);
  ATypeCount := ASymbolList.Count;
  ACount := ARoutine.Parameters.Count;
  ParseParameterNameList(ARoutine, pkConst, AAddParameters, ASymbolList);
  case TokenID of
    ptColon: begin
      NextToken;
      vType := ParseNewFormalParameterType(APredefinedArrayType);
      for i := ATypeCount to ASymbolList.Count - 1 do
        ASymbolList[i].TypeIntf := vType;
      if TokenID = ptEqual then
      begin
        NextToken;
        TypedConstant;
      end;
      if AAddParameters then
        for i := ACount to ARoutine.Parameters.Count - 1 do
        begin
          ARoutine.Parameters[i].ParamType := vType;
          // in case parameter is a tkArray type...
          ARoutine.Parameters[i].PredefinedArrayType := APredefinedArrayType;
        end;
    end
  end;
end;

procedure TRoutineParser.ParseVarParameter(const ARoutine: TcwRoutine;
  const AAddParameters: Boolean; const ASymbolList: TcwSymbolList);
var
  APredefinedArrayType: TcwSymbol;
  ATypeCount: Integer;
  ACount: Integer;
  i: Integer;
  vType: TcwSymbol;
begin
  Expected(ptVar);
  ACount := ARoutine.Parameters.Count;
  ATypeCount := ASymbolList.Count;
  ParseParameterNameList(ARoutine, pkVar, AAddParameters, ASymbolList);
  case TokenID of
    ptColon: begin
      NextToken;
      vType := ParseNewFormalParameterType(APredefinedArrayType);
      for i := ATypeCount to ASymbolList.Count - 1 do
        ASymbolList[i].TypeIntf := vType;
      if AAddParameters then
        for i := ACount to ARoutine.Parameters.Count - 1 do
        begin
          ARoutine.Parameters[i].ParamType := vType;
          // in case parameter is a tkArray type...
          ARoutine.Parameters[i].PredefinedArrayType := APredefinedArrayType;
        end;
    end
  end;
end;

procedure TRoutineParser.StoreImplementationEnd(const aRoutine: TcwRoutine; const aLocation: TPoint);
begin
  if Assigned(aRoutine) then
  begin
    aRoutine.ImplEndX:=aLocation.X+1;
    aRoutine.ImplEndY:=aLocation.Y+1;
  end;
end;

procedure TRoutineParser.StoreImplementationStart(const aRoutine: TcwRoutine; const aLocation: TPoint);
begin
  if Assigned(aRoutine) then
  begin
    aRoutine.ImplStartX:=aLocation.X+1;
    aRoutine.ImplStartY:=aLocation.Y+1;
  end;
end;

procedure TRoutineParser.ParseOutParameter(const ARoutine: TcwRoutine;
  const AAddParameters: Boolean; const ASymbolList: TcwSymbolList);
var
  APredefinedArrayType: TcwSymbol;
  ATypeCount: Integer;
  ACount: Integer;
  i: Integer;
  vType: TcwSymbol;
begin
  ExpectedEx(ptOut);
  ACount := ARoutine.Parameters.Count;
  ATypeCount := ASymbolList.Count;
  ParseParameterNameList(ARoutine, pkOut, AAddParameters, ASymbolList);
  case TokenID of
    ptColon: begin
      NextToken;
      vType := ParseNewFormalParameterType(APredefinedArrayType);
      for i := ATypeCount to ASymbolList.Count - 1 do
        ASymbolList[i].TypeIntf := vType;
      if AAddParameters then
        for i := ACount to ARoutine.Parameters.Count - 1 do
        begin
          ARoutine.Parameters[i].ParamType := vType;
          // in case parameter is a tkArray type...
          ARoutine.Parameters[i].PredefinedArrayType := APredefinedArrayType;
        end;
    end
  end;
end;

procedure TRoutineParser.ParseParameterFormal(const ARoutine: TcwRoutine;
  const aParamKind: TParamKind; const AAddParameters: Boolean; const ASymbolList: TcwSymbolList);
var
  APredefinedArrayType: TcwSymbol;
  ATypeCount: Integer;
  ACount: Integer;
  i: Integer;
  vType: TcwSymbol;
begin
  case TokenID of
    ptIdentifier: begin
      ACount := ARoutine.Parameters.Count;
      ATypeCount := ASymbolList.Count;
      ParseParameterNameList(ARoutine, aParamKind, AAddParameters, ASymbolList);
      Expected(ptColon);
      vType := ParseNewFormalParameterType(APredefinedArrayType);
      for i := ATypeCount to ASymbolList.Count - 1 do
        ASymbolList[i].TypeIntf := vType;
      if TokenID = ptEqual then
      begin
        NextToken;
        TypedConstant;
      end;
      if AAddParameters then
        for i := ACount to ARoutine.Parameters.Count - 1 do
        begin
          ARoutine.Parameters[i].ParamType := vType;
          // in case parameter is a tkArray type...
          ARoutine.Parameters[i].PredefinedArrayType := APredefinedArrayType;
        end;
    end;
    else
      SynError(InvalidParameter);
  end;
end;

procedure TRoutineParser.ParseFormalParameterSection(var ARoutine: TcwRoutine;
  const AAddParameters: Boolean; const ASymbolList: TcwSymbolList);
begin
  case TokenID of
    ptConst: ParseConstParameter(ARoutine, AAddParameters, ASymbolList);
    ptIdentifier:
      case ExID of
        ptOut: ParseOutParameter(ARoutine, AAddParameters, ASymbolList);
        else
          ParseParameterFormal(ARoutine, pkNone, AAddParameters, ASymbolList);
      end;
    ptIn: begin
      NextToken;
      ParseParameterFormal(ARoutine, pkIn, AAddParameters, ASymbolList);
    end;
    ptVar: ParseVarParameter(ARoutine, AAddParameters, ASymbolList);
  end;
end;

procedure TRoutineParser.ParseFormalParameterList(var ARoutine: TcwRoutine;
  const AAddParameters: Boolean; AFindOverloadedRoutine: Boolean);

  function RoutineMatchesTypeList(const ARoutine: TcwRoutine; const ASymbolList: TcwSymbolList):
    Boolean;
  var
    Index: Integer;
    AType: TcwSymbol;
  begin
    Result := ASymbolList.Count = ARoutine.Parameters.Count;
    if Result then
      for Index := 0 to ASymbolList.Count - 1 do
      begin
        AType := ASymbolList[Index].TypeIntf;
        Result := Result and
          ((ARoutine.Parameters[Index].UntypedParameter and not ASymbolList[Index].TypeAssigned) or
           (Assigned(ARoutine.Parameters[Index].ParamType) and
            SameText(ARoutine.Parameters[Index].ParamType.Name, AType.Name)));
        if not Result then
          Break;
      end;
  end;

var
  AVarContainer: TcwVarContainer;
  AParameter: TcwParameter;
  Index: Integer;
  ASymbolList: TcwSymbolList;
  ATempRoutine: TcwRoutine;
begin
  ASymbolList := TcwSymbolList.Create;
  try
    Expected(ptRoundOpen);
    ParseFormalParameterSection(ARoutine, AAddParameters, ASymbolList);
    while TokenID = ptSemiColon do
    begin
      SEMICOLON;
      ParseFormalParameterSection(ARoutine, AAddParameters, ASymbolList);
    end;
    Expected(ptRoundClose);

    if AFindOverloadedRoutine then
    begin
      AVarContainer := ARoutine.RoutineList.Container;
      // see if type list matches current routine, and if not find and set the correct routine...
      if not RoutineMatchesTypeList(ARoutine, ASymbolList) then
      begin
        for Index := 0 to AVarContainer.Routines.Count - 1 do
        begin
          ATempRoutine := AVarContainer.Routines[Index];
          if (ATempRoutine <> ARoutine) and SameText(ATempRoutine.Name, ARoutine.Name) and
            RoutineMatchesTypeList(ATempRoutine, ASymbolList) then
          begin
            ARoutine := ATempRoutine;
            Break;
          end;
        end;
      end;
    end;
  
    // now that we know the correct routine, update parameter usedby...
    for Index := 0 to ASymbolList.Count - 1 do
    begin
      AParameter := ARoutine.Parameters.Find(ASymbolList[Index].Name);
      if Assigned(AParameter) then
        AParameter.UsedBy.AddUsedBy(UsedByInfo(FileIndex,
          ASymbolList[Index].Name <> AParameter.Name, ASymbolList[Index].X, ASymbolList[Index].Y))
      else
        if Assigned(OnMessage) then
          OnMessage(Self, Format('Parameter, %s, not found.', [ASymbolList[Index].Name]),
            ASymbolList[Index].X, ASymbolList[Index].Y);
    end;
  finally
    ASymbolList.Free;
  end;
end;

function TRoutineParser.ParseFunctionHeading(const AContainer: TcwVarContainer): TcwRoutine;
begin
  Expected(ptFunction);
  Result := AContainer.Routines.AddRoutine(Lexer.Token, UnitVisibility,
    rkFunction, FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
  FunctionProcedureName;
  if TokenID = ptRoundOpen then
    ParseFormalParameterList(Result, True, False);
  Expected(ptColon);
  AddReturnType(Result);
end;

function TRoutineParser.ParseProcedureHeading(const AContainer: TcwVarContainer): TcwRoutine;
begin
  Expected(ptProcedure);
  Result := AContainer.Routines.AddRoutine(Lexer.Token, UnitVisibility,
    rkProcedure, FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
  FunctionProcedureName;
  if TokenID = ptRoundOpen then
    ParseFormalParameterList(Result, True, False);
end;

procedure TRoutineParser.ExportedHeading;
var
  vRoutine: TcwRoutine;
  Directive: TRoutineDirective;
begin
  vRoutine := nil;
  case TokenID of
    ptFunction: vRoutine := ParseFunctionHeading(CurrentContainer);
    ptProcedure: vRoutine := ParseProcedureHeading(CurrentContainer);
    else
      SynError(InvalidExportedHeading);
  end;
  StoreDeclarationEnd(vRoutine);

  if TokenID = ptSemiColon then
    SEMICOLON;

  case ExID of
    ptForward: begin
      ForwardDeclaration; //jdj added 02/07/2001
      //        NextToken;
      //        SEMICOLON;
    end;
    ptAssembler: begin
      NextToken;
      SEMICOLON;
      if Exid = ptForward then
        ForwardDeclaration; //jdj added 02/07/2001
    end;

    else
    begin
      while (ExID in ([ptAbstract, ptCdecl, ptDynamic, ptExport, ptExternal, ptFar,
        ptMessage, ptNear, ptOverload, ptOverride, ptPascal, ptAssembler, ptRegister,
        ptReintroduce, ptSafeCall, ptStdCall, ptVirtual] + [ptDeprecated, ptPlatform])) or
        (TokenID = ptLibrary) do
      begin
        if (ExID in [ptDeprecated, ptPlatform]) or (TokenID = ptLibrary) then
          HintDirectives
        else
        begin
          Directive := ParseProceduralDirective;
          if (vRoutine <> nil) and (Directive <> rdNone) then
            vRoutine.AddDirective(Directive);
        end;
        if TokenID = ptSemiColon then
          SEMICOLON;
      end;
      if TokenID = ptSemiColon then
        SEMICOLON;
      if ExId = ptForward then
        ForwardDeclaration; //jdj added 02/07/2001
    end;
  end;
end;

procedure TRoutineParser.ParseFunctionProcedureBlock(const ARoutine: TcwRoutine);
var
  NoExternal: Boolean;
  Directive: TRoutineDirective;
  ASaveVisibility: TVisibilityKind;
begin
  ASaveVisibility := UnitVisibility;
  try
    UnitVisibility := vkNone;
    NoExternal := True;
    if TokenID = ptSemiColon then
      SEMICOLON;
    if ExID = ptForward then
    begin
      NextToken;
      SEMICOLON;
    end else
    begin
      while ExID in [ptAbstract, ptCdecl, ptDynamic, ptExport, ptExternal, ptFar,
        ptMessage, ptNear, ptOverload, ptOverride, ptPascal, ptAssembler, ptRegister,
        ptReintroduce, ptSafeCall, ptStdCall, ptVirtual, ptLocal] do
        case ExId of
          ptExternal:
          begin
            Directive := ParseProceduralDirective;
            if (Directive <> rdNone) then
              ARoutine.AddDirective(Directive);
            StoreImplementationEnd(ARoutine, Lexer.PosXY);
            if TokenID = ptSemiColon then
              SEMICOLON;
            NoExternal := False;
          end;
        else
          begin
            Directive := ParseProceduralDirective;
            if (Directive <> rdNone) then
              ARoutine.AddDirective(Directive);
            if TokenID = ptSemiColon then
              SEMICOLON;
          end;
        end;

      if NoExternal then
      begin
        if ExId = ptAssembler then
        begin
          NextToken;
          SEMICOLON;
        end;
        case TokenID of
          ptAsm: AsmStatement;
          else
            ParseBlock(ARoutine);
        end;
        StoreImplementationEnd(ARoutine, Lexer.PosXY);
        SEMICOLON;
      end;
    end;
  finally
    UnitVisibility := ASaveVisibility;
  end;
end;

function TRoutineParser.IsOverloadedInAncestors(ARoutineName: string; AClass: TcwClass = nil):
  Boolean;
var
  ASymbol: TcwSymbol;
begin
  Result := False;
  if Assigned(AClass) then
  begin
    ASymbol := AClass.FindSymbol(UnitIntf, AClass, ARoutineName);
    if Assigned(ASymbol) and (ASymbol is TcwRoutine) then
      Result := rdOverload in (ASymbol as TcwRoutine).Directives;
  end;
end;

function TRoutineParser.ParseFunctionProcedureMethodDeclaration(const AContainer: TcwVarContainer):
  TcwRoutine;
var
  ARoutineKind: TRoutineKind;
  ASymbol: TcwSymbol;
  AWasOverloaded: boolean;
  LocationXY: TPoint;
  vClass: TcwClass;
  ARoutineAdded: Boolean;
  AIsFunction: Boolean;
begin
  LocationXY := Lexer.PosXY;
  AIsFunction := TokenID = ptFunction;
  if TokenID in [ptConstructor, ptDestructor, ptProcedure, ptFunction] then
    NextToken
  else
    SynError(InvalidProcedureMethodDeclaration);
  Lexer.InitAhead;
  vClass := nil;
  if Lexer.AheadTokenID = ptPoint then
  begin
    vClass := UnitIntf.Classes.Find(Lexer.Token);
    if vClass = nil then
      vClass := UnitIntf.Objects.Find(Lexer.Token);
    if Assigned(vClass) then
      vClass.UsedBy.AddUsedBy(UsedByInfo(FileIndex, Lexer.Token <> vClass.Name,
        Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1));
    ObjectNameOfMethod;
    Expected(ptPoint);
    if Assigned(vClass) then
      Result := vClass.Routines.Find(Lexer.Token)
    else
      Result := nil;
    //todo: .sm if routine couldn't be found, this is actually an error
    //currently the routine is just added to the unit which is totally wrong!
  end else
  begin
    //todo: .sm instead of searching in UnitIntf, search in AContainer? check with
    //local methods which have a forward declaration
    ASymbol := UnitIntf.FindSymbol(UnitIntf, nil, Lexer.Token, True, GetUsedByInfo);
    if ASymbol is TcwRoutine then
      Result := ASymbol as TcwRoutine
    else
      Result := nil;
  end;

  AWasOverloaded := Assigned(Result) and ((rdOverload in Result.Directives) or
    ((Result.RoutineList.Container is TcwClass) and (rdOverride in Result.Directives) and
    IsOverloadedInAncestors(Result.Name,
      (Result.RoutineList.Container as TcwClass).InheritsFromClass)));
  ARoutineAdded := not Assigned(Result) or
    (AWasOverloaded and (Result.Visibilty = vkImplementation));
  if ARoutineAdded then
  begin
    if AIsFunction then
      ARoutineKind := rkFunction
    else
      ARoutineKind := rkProcedure;
    Result := AContainer.Routines.AddRoutine(Lexer.Token, UnitVisibility,
      ARoutineKind, FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
  end;

  FunctionProcedureName;
  if TokenID = ptRoundOpen then
    ParseFormalParameterList(Result, ARoutineAdded, AWasOverloaded);

  if ARoutineAdded then
    StoreDeclarationEnd(Result);

  PushRoutine(Result, vClass);
  try
    StoreImplementationStart(Result, LocationXY);
    if AIsFunction then
    begin
      case TokenID of
        ptSemiColon: ParseFunctionProcedureBlock(Result);
        else
        begin
          Expected(ptColon);
          if ARoutineAdded then
            AddReturnType(Result)
          else
            ParseReturnType;
          ParseFunctionProcedureBlock(Result);
        end;
      end;
    end else
      ParseFunctionProcedureBlock(Result);
  finally
    PopRoutine;
  end;
end;

function TRoutineParser.ParseProcedureDeclarationSection(const AContainer: TcwVarContainer):
  TcwRoutine;
begin
  Result := nil;
  if TokenID = ptClass then
    NextToken;
  case TokenID of
    ptConstructor, ptDestructor, ptProcedure:
      Result := ParseFunctionProcedureMethodDeclaration(AContainer);
    ptFunction:
      ParseFunctionProcedureMethodDeclaration(AContainer);
    else
      SynError(InvalidProcedureDeclarationSection);
  end;
end;

end.
