(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwStructuredTypeParser;

interface

uses
  Types,
  Classes,
  ucwClasses,
  ucwVarParser,
  mwSimplePasPar;

type
  TStructuredTypeParser = class(TVarParser)
  private
    procedure ParseRecordVariant(const aRecord: TcwRecord);
    procedure ParseRecordVariantSection(const aRecord: TcwRecord);
  protected
    procedure ParseFieldDeclaration(const aRecord: TcwRecord);
    procedure ParseFieldList(const aRecord: TcwRecord);
    function ParseArrayType(const aContainer: TcwSymbolContainer; const aTypeName: string;
      const aCodeLocation: TPoint): TcwType;
    function ParseRecordType(const aContainer: TcwVarContainer; const aTypeName: string;
      const aCodeLocation: TPoint): TcwSymbol;
    function ParseSetType(const aContainer: TcwVarContainer; const aTypeName: string;
      const aCodeLocation: TPoint): TcwSymbol;
    function ParseStructuredType(const aContainer: TcwSymbolContainer; const aTypeName: string;
      const aCodeLocation: TPoint; out ATypeAdded: Boolean): TcwSymbol; override;
  end;

implementation

uses
  SysUtils,
  mwPasLexTypes,
  mwSimplePasParTypes,
  mwPasLex, ucwUnitParser;

procedure TStructuredTypeParser.ParseFieldDeclaration(const aRecord: TcwRecord);
var
  vVarList: TStringList;
  i: Integer;
  Loc: TSmallPoint;
  vNewType: TcwSymbol;
  vField: TcwVariable;
begin
  vVarList := TStringList.Create;
  try
    GetIdentifierList(vVarList);
    Expected(ptColon);
    vNewType := ParseTypeKind(aRecord, Lexer.Token, Lexer.PosXY, False);
    for i := 0 to vVarList.Count - 1 do
    begin
      Loc := TSmallPoint(vVarList.Objects[i]);
      vField := aRecord.Fields.AddVariable(vVarList[i], vkPublic, False,
        FileIndex, Loc.x,  Loc.y);
      vField.TypeIntf := vNewType;
    end;
  finally
    vVarList.Free;
  end;
end;

procedure TStructuredTypeParser.ParseFieldList(const aRecord: TcwRecord);
begin
  while TokenID = ptIdentifier do
  begin
    ParseFieldDeclaration(aRecord);
    SEMICOLON;
  end;
  if TokenID = ptCase then
    ParseRecordVariantSection(aRecord);
  SEMICOLON;
end;

function TStructuredTypeParser.ParseArrayType(const aContainer: TcwSymbolContainer;
  const aTypeName: string; const aCodeLocation: TPoint): TcwType;

  procedure ParseNextDimension(const aContainer: TcwSymbolContainer; const AArray: TcwArray);
  var
    AResultExpected: Boolean;
  begin
    NextToken;
    case TokenID of
      ptIdentifier: begin
        AArray.AddDimension(ParseVariableReference(nil));
        case TokenID of
          ptComma, ptDotDot, ptSquareClose: ;
          else
            Expression;
        end;
        if TokenID = ptDotDot then
          // recurse!!
          ParseNextDimension(AContainer, AArray);
      end;
      else
        AArray.AddDimension(ParseOrdinalType(aContainer, Lexer.PosXY, AResultExpected,
          Lexer.Token));
    end;
  end;

var
  AArray: TcwArray;
begin
  AArray := (aContainer as TcwVarContainer).Arrays.AddArray(aTypeName, UnitVisibility, FileIndex,
    aCodeLocation.X + 1, aCodeLocation.Y + 1);
  Result := AArray;
  Expected(ptArray);
  if TokenID = ptSquareOpen then
  begin
    ParseNextDimension(aContainer, AArray);
    while TokenID = ptComma do
      ParseNextDimension(aContainer, AArray);
    Expected(ptSquareClose);
  end;
  Expected(ptOf);
  AArray.ArrayType := ParseTypeKind(aContainer, Lexer.Token, Lexer.PosXY, False);
  StoreDeclarationEnd(AArray);
end;

function TStructuredTypeParser.ParseRecordType(const aContainer: TcwVarContainer;
  const aTypeName: string; const aCodeLocation: TPoint): TcwSymbol;
var
  vRecord: TcwRecord;
begin
  if (Length(aTypeName) > 0) then
    vRecord := aContainer.Records.AddRecord(aTypeName, UnitVisibility,
      FileIndex, aCodeLocation.X + 1, aCodeLocation.Y + 1)
  else
    vRecord := aContainer.Records.AddRecord('record', UnitVisibility, FileIndex,
      aCodeLocation.X + 1, aCodeLocation.Y + 1);

  Result := vRecord;
  Expected(ptRecord);
  ParseFieldList(vRecord);
  Expected(ptEnd);
  StoreDeclarationEnd(vRecord);

  HintDirectives;
end;

function TStructuredTypeParser.ParseSetType(const aContainer: TcwVarContainer; const aTypeName: string;
  const aCodeLocation: TPoint): TcwSymbol;
var
  AReferencedEnumeration: TcwSymbol;
  ASet: TcwSet;
begin
  ASet := aContainer.Sets.AddItem(aTypeName, UnitVisibility,
    FileIndex, aCodeLocation.X+1, aCodeLocation.Y+1);
  Expected(ptSet);
  Expected(ptOf);
  case TokenID of
    ptIdentifier: begin
      Lexer.InitAhead;
      case Lexer.AheadTokenID of
        ptPoint: Expression;
        ptRoundOpen: ConstantExpression;
        else
        begin
          // explicit enumeration (TSet = set of TEnum);
          AReferencedEnumeration := FindType(Lexer.Token);

          if AReferencedEnumeration is TcwEnumeration then
            ASet.Enumeration := AReferencedEnumeration as TcwEnumeration;

          Expected(ptIdentifier);
        end;
      end;
    end;
    // implicit enumeration (TSet = set of (one, two);
    ptRoundOpen: ParseEnumerationMembers(aContainer, ASet);
    ptSquareOpen: begin
      NextToken;
      SubrangeType;
      Expected(ptSquareClose);
    end;
    else
      Expression;
  end;

  if TokenID = ptDotDot then
  begin
    NextToken;
    ConstantExpression;
  end;

  StoreDeclarationEnd(ASet);

  Result := ASet as TcwSymbol;
end;

function TStructuredTypeParser.ParseStructuredType(const aContainer: TcwSymbolContainer;
  const aTypeName: string; const aCodeLocation: TPoint; out ATypeAdded: Boolean): TcwSymbol;
begin
  Result := nil;
  ATypeAdded := False;
  if TokenID = ptPacked then
    NextToken;
  case TokenID of
    ptArray: begin
      Result := ParseArrayType(aContainer, aTypeName, aCodeLocation);
      ATypeAdded := True;
    end;
    ptFile: begin
      Result := FindType(Lexer.Token);
      FileType;
    end;
    ptRecord: begin
      Result := ParseRecordType(aContainer as TcwVarContainer, aTypeName, aCodeLocation);
      ATypeAdded := True;
    end;
    ptSet: begin
      Result := ParseSetType(aContainer as TcwVarContainer, aTypeName, aCodeLocation);
      ATypeAdded := True;
    end;
    else
      SynError(InvalidStructuredType);
  end;
end;

procedure TStructuredTypeParser.ParseRecordVariant(const aRecord: TcwRecord);
begin
  ConstantExpression;
  while (TokenID = ptComma) do
  begin
    NextToken;
    ConstantExpression;
  end;
  Expected(ptColon);
  Expected(ptRoundOpen);
  if TokenID in [ptIdentifier, ptCase] then
    ParseFieldList(aRecord);
  Expected(ptRoundClose);
end;

procedure TStructuredTypeParser.ParseRecordVariantSection(const aRecord: TcwRecord);

  procedure ParseTagField;
  var
    AField: TcwVariable;
    ATagName: string;
    AType: TcwSymbol;
    APoint: TPoint;
  begin
    APoint := Lexer.PosXY;
    ATagName := Lexer.Token;
    Expected(ptIdentifier);

    case Lexer.TokenID of
      ptColon: begin
        NextToken;
        AType := ParseTypeKind(ARecord, Lexer.Token, Lexer.PosXY, False);
        AField := ARecord.Fields.AddVariable(ATagName, vkPublic, False, 
          FileIndex, APoint.X + 1, APoint.Y + 1);
        AField.TypeIntf := AType;
      end;
    end;
  end;

begin
  Expected(ptCase);
  ParseTagField;
  Expected(ptOf);
  ParseRecordVariant(aRecord);
  while TokenID = ptSemiColon do
  begin
    SEMICOLON;
    case TokenID of
      ptRoundClose, ptElse, ptEnd: ;
      else
        ParseRecordVariant(aRecord);
    end;
  end;
end;

end.
