(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwTreeNode;

interface

uses SysUtils, ComCtrls, ucwClasses;

type
  TcwSymbolType = (
       stUnit,
       stUsedUnits,
       stUsedBy,
       stClasses,
       stClass,
       stObjects,
       stObject,
       stInterfaces,
       stInterface,
       stRecords,
       stRecord,
       stRoutines,
       stRoutine,
       stVars,
       stVar,
       stConst,
       stConsts,
       stProperty,
       stTypes,
       stType,
       stLocation,
       stLocationWrongCase,
       stParameters,
       stParameter,
       stEnumerations,
       stEnumeration,
       stSets,
       stSet,
       stArrays,
       stArray


       );


const
  VirtualSymbolTreeNodes = [stUsedUnits, stUsedBy, stClasses, stInterfaces, stUsedUnits,
    stObjects, stRecords, stVars, stConsts, stTypes, stRoutines, stEnumerations, stSets, stArrays];

  ImageIndexes: array[TcwSymbolType] of Integer = (
       0, // stUnit
       1, // stUsedUnits
       1, // stUsedBy
       11, // stClasses
       10, // stClass
       13, // stObjects
       9, // stObject
       12, // stInterfaces
       8, // stInterface
       15, // stRecords
       14, // stRecord
       2, // stRoutines
       2, // stRoutine
       3, // stVars
       3, // stVar
       3, // stConst
       3, // stConsts
       4, // stProperty
       5, // stTypes
       5, // stType
       6, // stLocation
       7, // stLocationWrongCase
       3, // stParameters
       3, // stParameter
       3, //stEnumerations,
       3, //stEnumeration
       3, //stSets
       3, //stSet
       3, // stArrays
       3  // stArray
       );



type
  TcwTreeNode = class(TTreeNode)
  private
    FLocation: TcwCodeLocation;
    FUnit: TcwUnit;
    FSymbol: TcwSymbol;
    FNodeType: TcwSymbolType;
    function  VarContainerHasChildren(const aContainer : TcwVarContainer) : Boolean;

    procedure AddTypes(const aContainer : TcwVarContainer);
    function  AddSymbol(const aSymbol: TcwSymbol; aType: TcwSymbolType) : TcwTreeNode;
    procedure AddRoutineReturnValue(const aRoutine : TcwRoutine);
    procedure AddRoutineParameters(const aRoutine : TcwRoutine);
    procedure AddVars(const aContainer : TcwVarContainer;const aValue: TcwSymbolType);
    procedure AddEnumerations(const aValue : TcwVarContainer);
    procedure AddSets(const aValue : TcwVarContainer);
    procedure AddArrays(const aValue : TcwVarContainer);
    procedure AddRoutines(const aContainer : TcwVarContainer;const aValue: TcwSymbolType);
    procedure AddLocation(const aLocation: TcwUsedByLocation);
    procedure AddClassMembers(const aName: string; const aMembersList : TcwSymbolList);
    procedure AddClassesNode(const aContainer : TcwVarContainer);
    procedure AddObjectsNode(const aContainer : TcwVarContainer);
    procedure AddUsedBy(const aParentNode : TcwTreeNode; const aUsedByList: TcwUsedByList);
    procedure AddUsedUnits(const aUsedUnitList: TcwUsedUnitList);
    procedure AddInterfacesNode(const aContainer : TcwVarContainer);
    procedure AddRecordsNode(const aContainer : TcwVarContainer);
    procedure AddClassNode(const aClass: TcwClass; const aUnit: TcwUnit; const aType: TcwSymbolType);
    procedure ExpandUsedBy;
    procedure ExpandClass;
    procedure ExpandVar;
    procedure ExpandVars(const aContainer : TcwVarContainer);
    procedure ExpandRoutines(const aContainer : TcwVarContainer);
    procedure ExpandRoutine;
    procedure ExpandParameters;
    procedure ExpandParameter;
    procedure ExpandType;
    procedure ExpandTypes(const aContainer : TcwVarContainer);
    procedure ExpandEnumeration;
    procedure ExpandEnumerations;
    procedure ExpandSet;
    procedure ExpandSets;
    procedure ExpandArray;
    procedure ExpandArrays;

    procedure ExpandConsts(const aContainer : TcwVarContainer);
    procedure ExpandClasses(const aContainer : TcwVarContainer);
    procedure ExpandObjects(const aContainer : TcwVarContainer);
    procedure ExpandInterfaces(const aContainer : TcwVarContainer);
    procedure ExpandVarContainer(const aContainer : TcwVarContainer);
    procedure ExpandUnit;
    procedure ExpandRecords(const aContainer : TcwVarContainer);
    procedure ExpandRecord;

    procedure ExpandClassDescendants;
    procedure ExpandInterfaceDescendants;
    procedure ExpandClassesDescendants;
    procedure ExpandObjectsDescendants;
    procedure ExpandInterfacesDescendants;
    procedure SetNodeType(const aValue: TcwSymbolType);
  public
    procedure Execute;
    procedure ExpandNode(ShowUsageTree : Boolean);
    function AddUnit(aParent: TTreeNode; const aUnit: TcwUnit) : TcwTreeNode;
    property UnitIntf: TcwUnit read FUnit write FUnit;
    property NodeType: TcwSymbolType read FNodeType write SetNodeType;
    property SymbolIntf: TcwSymbol read FSymbol write FSymbol;
    property Location: TcwCodeLocation read FLocation write FLocation;
  end;
  
implementation

procedure TcwTreeNode.AddClassesNode(const aContainer : TcwVarContainer);
var
  NewNode: TcwTreeNode;
begin
  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self,
        Format('Classes (%d)', [aContainer.Classes.Count])));

  NewNode.SymbolIntf := aContainer;
  NewNode.UnitIntf := UnitIntf;
  NewNode.NodeType := stClasses;

  NewNode.HasChildren := True;
end;

procedure TcwTreeNode.AddObjectsNode(const aContainer : TcwVarContainer);
var
  NewNode: TcwTreeNode;
begin
  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self,
        Format('Objects (%d)', [aContainer.Objects.Count])));

  NewNode.SymbolIntf := aContainer;
  NewNode.UnitIntf := UnitIntf;
  NewNode.NodeType := stObjects;

  NewNode.HasChildren := True;
end;


procedure TcwTreeNode.AddInterfacesNode(const aContainer : TcwVarContainer);
var
  NewNode: TcwTreeNode;
begin
  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self,
        Format('Interfaces (%d)', [aContainer.Interfaces.Count])));

  NewNode.SymbolIntf := aContainer;
  NewNode.UnitIntf := UnitIntf;
  NewNode.NodeType := stInterfaces;

  NewNode.HasChildren := True;
end;


procedure TcwTreeNode.AddRecordsNode(const aContainer : TcwVarContainer);
var
  NewNode: TcwTreeNode;
begin
  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self,
        Format('Records (%d)', [aContainer.Records.Count])));

  NewNode.SymbolIntf := aContainer;
  NewNode.UnitIntf := UnitIntf;
  NewNode.NodeType := stRecords;

  NewNode.HasChildren := True;
end;


procedure TcwTreeNode.AddTypes(const aContainer : TcwVarContainer);
var
  s: string;
  NewNode: TcwTreeNode; 
begin
  s := Format('Types (%d)', [aContainer.Types.Count]);
  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, s));
  NewNode.UnitIntf := FUnit;
  NewNode.SymbolIntf := aContainer;
  NewNode.NodeType := stTypes;
  NewNode.HasChildren := (aContainer.Types.Count > 0);
end;

procedure TcwTreeNode.AddVars(const aContainer : TcwVarContainer;const aValue: TcwSymbolType);
var
  s: string;
  aList: TcwVariablesList;
  NewNode: TcwTreeNode;
begin
  if aValue = stVars then
  begin
    s := 'Variables';
    aList := aContainer.Variables;
  end
  else
  begin
    s := 'Constants';
    aList := aContainer.Constants;
  end;

  s := Format(s + ' (%d)', [aList.Count]);

  // add a class node
  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, s));
  NewNode.UnitIntf := FUnit;
  NewNode.SymbolIntf := aContainer;
  NewNode.NodeType := aValue;
  NewNode.HasChildren := (aList.Count > 0);
end;

procedure TcwTreeNode.AddEnumerations(const aValue : TcwVarContainer);
var
  NewNode: TcwTreeNode;
begin
  // add a class node
  if aValue.Enumerations.Count <= 0 then EXIT;

  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, Format('Enumerations (%d)', [aValue.Enumerations.Count])));
  NewNode.UnitIntf := FUnit;
  NewNode.SymbolIntf := aValue;
  NewNode.NodeType := stEnumerations;
  NewNode.HasChildren := (aValue.Enumerations.Count > 0);
end;

procedure TcwTreeNode.AddSets(const aValue : TcwVarContainer);
var
  NewNode: TcwTreeNode;
begin
  // add a class node
  if aValue.Sets.Count <= 0 then EXIT;

  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, Format('Sets (%d)', [aValue.Sets.Count])));
  NewNode.UnitIntf := FUnit;
  NewNode.SymbolIntf := aValue;
  NewNode.NodeType := stSets;
  NewNode.HasChildren := (aValue.Sets.Count > 0);
end;

procedure TcwTreeNode.AddArrays(const aValue : TcwVarContainer);
var
  NewNode: TcwTreeNode;
begin
  // add a class node
  if aValue.Arrays.Count <= 0 then EXIT;

  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, Format('Arrays (%d)', [aValue.Arrays.Count])));
  NewNode.UnitIntf := FUnit;
  NewNode.SymbolIntf := aValue;
  NewNode.NodeType := stArrays;
  NewNode.HasChildren := (aValue.Arrays.Count > 0);
end;

procedure TcwTreeNode.AddRoutines(const aContainer : TcwVarContainer;const aValue: TcwSymbolType);
var
  NewNode: TcwTreeNode;
begin
  // add a class node
  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, Format('Routines (%d)', [aContainer.Routines.Count])));
  NewNode.UnitIntf := FUnit;
  NewNode.SymbolIntf := aContainer;
  NewNode.NodeType := aValue;
  NewNode.HasChildren := (aContainer.Routines.Count > 0);
end;

procedure TcwTreeNode.ExpandVarContainer(const aContainer : TcwVarContainer);
begin
  if aContainer.Classes.Count > 0 then
    AddClassesNode(aContainer);
  if aContainer.Objects.Count > 0 then
    AddObjectsNode(aContainer);
  if aContainer.Interfaces.Count > 0 then
    AddInterfacesNode(aContainer);
  if aContainer.Records.Count > 0 then
    AddRecordsNode(aContainer);
  if aContainer.Types.Count > 0 then
    AddTypes(aContainer);
  if aContainer.Variables.Count > 0 then
    AddVars(aContainer,stVars);
  if aContainer.Constants.Count > 0 then
    AddVars(aContainer,stConsts);
  if aContainer.Enumerations.Count > 0 then
    AddEnumerations(aContainer);
  if aContainer.Sets.Count > 0 then
    AddSets(aContainer);
  if aContainer.Arrays.Count > 0 then
    AddArrays(aContainer);
  if aContainer.Routines.Count > 0 then
    AddRoutines(aContainer,stRoutines);
end;

procedure TcwTreeNode.ExpandUnit;
begin
  if FUnit = nil then
    Exit;
  ExpandVarContainer(FUnit);

  if FUnit.UsedUnits.Count > 0 then
     AddUsedUnits(FUnit.UsedUnits);
  if FUnit.UsedBy.Count > 0 then
     AddUsedBy(Self, FUnit.UsedBy);
end;

procedure TcwTreeNode.AddClassNode(const aClass: TcwClass; const aUnit: TcwUnit; const aType: TcwSymbolType);
var
  s: string;
  i: Integer;
  NewNode : TcwTreeNode;
begin
  // get the class name
  s := '';
  if aClass.InheritsFromClass <> nil then
    s := s + aClass.InheritsFromClass.Name;

  for i := 0 to aClass.Interfaces.Count - 1 do
  begin
    if Length(s) > 0 then
      s := s + ', ';
    s := s + aClass.Interfaces.SortedInterfaces[i].Name;
  end;
  if length(s) > 0 then
    s := Format('%s (%s)',[aClass.Name, s]) else
      s := aClass.Name;

  // add a class node
  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, s));
  NewNode.UnitIntf := aUnit;
  NewNode.SymbolIntf := aClass;
  NewNode.NodeType := aType;

  NewNode.HasChildren := (aClass.Fields.Count > 0) or
                         (aClass.Properties.Count > 0) or
                         (aClass.Routines.Count > 0) or
                         (aClass.UsedBy.Count > 0);
end;

procedure TcwTreeNode.ExpandClasses(const aContainer : TcwVarContainer);
var
  i: Integer;
  fClass : TcwClass;
begin
  if aContainer = nil then
    Exit;
  TTreeView(TreeView).Items.BeginUpdate;
  try
    for i := 0 to aContainer.Classes.Count - 1 do
    begin
      fClass := aContainer.Classes.SortedClasses[i];
      AddClassNode(fClass, FUnit,stClass);
    end;
  finally
    TTreeView(TreeView).Items.EndUpdate;
  end;
end;

procedure TcwTreeNode.ExpandObjects(const aContainer : TcwVarContainer);
var
  i: Integer;
  fClass : TcwClass;
begin
  if aContainer = nil then EXIT;
  TTreeView(TreeView).Items.BeginUpdate;
  try
    for i := 0 to aContainer.Objects.Count - 1 do
    begin
      fClass := aContainer.Objects.SortedClasses[i];
      AddClassNode(fClass, FUnit,stObject);
    end;
  finally
    TTreeView(TreeView).Items.EndUpdate;
  end;
end;

procedure TcwTreeNode.ExpandInterfaces(const aContainer : TcwVarContainer);
var
  i: Integer;
  fClass : TcwClass;
begin
  if aContainer = nil then EXIT;
  TTreeView(TreeView).Items.BeginUpdate;
  try
    for i := 0 to aContainer.Interfaces.Count - 1 do
    begin
      fClass := aContainer.Interfaces.SortedInterfaces[i];
      AddClassNode(fClass, FUnit, stInterface);
    end;
  finally
    TTreeView(TreeView).Items.EndUpdate;
  end;

end;


procedure TcwTreeNode.ExpandRecord;
var fRecord : TcwRecord;
var aVariable : TcwVariable;
var i         : Integer;
begin
  if (FUnit = nil) or (FSymbol = nil) then
     Exit;
  fRecord := FSymbol as TcwRecord;
  TTreeView(TreeView).Items.BeginUpdate;
  try
    for i := 0 to fRecord.Fields.Count - 1 do
    begin
       aVariable := fRecord.Fields.SortedVariables[i];
       AddSymbol(aVariable, stVar);
    end;
    // add a used by node
    AddUsedBy(Self, FSymbol.UsedBy);
  finally
    TTreeView(TreeView).Items.EndUpdate;
  end;
end;

procedure TcwTreeNode.ExpandRecords(const aContainer : TcwVarContainer);
var
  i: Integer;
  fRecord : TcwRecord;
  NewNode : TcwTreeNode;
begin
  if aContainer = nil then EXIT;
  TTreeView(TreeView).Items.BeginUpdate;
  try
    for i := 0 to aContainer.Records.Count - 1 do
    begin
      fRecord := aContainer.Records.SortedRecords[i];
      // add a class node
      NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, fRecord.Name));
      NewNode.UnitIntf := FUnit;
      NewNode.SymbolIntf := fRecord;
      NewNode.NodeType := stRecord;

      NewNode.HasChildren := (fRecord.Fields.Count > 0) or
                             (fRecord.UsedBy.Count > 0);

    end;

  finally
    TTreeView(TreeView).Items.EndUpdate;
  end;

end;

procedure TcwTreeNode.AddClassMembers(const aName: string; const aMembersList : TcwSymbolList);
var
  i: Integer;
  aSymbol: TcwSymbol;
  aCaptionNode,
  aSubNode:  TcwTreeNode;
  aProperty: TcwProperty;
  aRoutine: TcwRoutine;
  aVariable: TcwVariable;
begin
  if aMembersList.Count <= 0 then
    Exit;
  aCaptionNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, Format('%s (%d)',
       [aName, aMembersList.Count])));
  aCaptionNode.HasChildren := true;

  for i := 0 to aMembersList.Count - 1 do
  begin
    aSymbol := aMembersList.SortedSymbols[i];
    aSubNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(aCaptionNode, aSymbol.Name));
    aSubNode.HasChildren := False;

    if aSymbol is TcwProperty then
    begin
      aProperty := aSymbol as TcwProperty;
      aSubNode.UnitIntf := FUnit;
      aSubNode.SymbolIntf := aProperty;
      aSubNode.NodeType := stProperty;
      aSubNode.Text := 'Property ' + aSubNode.text;
      if aProperty.TypeIntf <> nil then
        aSubNode.Text := aSubNode.Text + ' : ' + aProperty.TypeIntf.Name;
    end else
    if aSymbol is TcwRoutine then
    begin
      aRoutine := aSymbol as TcwRoutine;
      aSubNode.UnitIntf := FUnit;
      aSubNode.SymbolIntf := aRoutine;
      aSubNode.NodeType := stRoutine;
      aSubNode.Text := RoutineKindName[aRoutine.RoutineKind] + ' ' + aRoutine.Name;
      if aRoutine.ReturnType <> nil then
        aSubNode.Text := aSubNode.Text + ' : ' + aRoutine.ReturnType.Name;
    end else
    if aSymbol is TcwVariable then
    begin
      aVariable := aSymbol as TcwVariable;
      aSubNode.UnitIntf := FUnit;
      aSubNode.SymbolIntf := aVariable;
      aSubNode.NodeType := stVar;
      aSubNode.Text := aVariable.Name;
      if aVariable.TypeIntf <> nil then
        aSubNode.Text := aSubNode.Text + ' : ' + aVariable.TypeIntf.Name;

    end;
    AddUsedBy(aSubNode, aSymbol.UsedBy);
  end;

  aMembersList.Free;
end;

function TcwTreeNode.AddUnit(aParent: TTreeNode; const aUnit: TcwUnit) : TcwTreeNode;
begin
  Result := TcwTreeNode(TTreeView(TreeView).Items.AddChild(aParent, aUnit.Name));
  Result.UnitIntf := aUnit;
  Result.SymbolIntf := aUnit;
  Result.NodeType := stUnit;

  Result.HasChildren := (aUnit.UsedUnits.Count > 0) or (aUnit.UsedBy.Count > 0) or
                         VarContainerHasChildren(aUnit);

                           
end;

procedure TcwTreeNode.AddUsedUnits(const aUsedUnitList: TcwUsedUnitList);
var
  i: Integer;
  FolderNode: TcwTreeNode;
  aUnit: TcwUnit;
begin
  if AUsedUnitList.Count <= 0 then
    Exit;    
  FolderNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, Format('Used Units (%d)',
      [AUsedUnitList.Count])));
  FolderNode.FUnit := FUnit;
  FolderNode.FSymbol := nil;
  FolderNode.NodeType := stUsedUnits;
  for i := 0 to AUsedUnitList.Count - 1 do
  begin
    aUnit := AUsedUnitList.SortedUsedUnits[i].UnitIntf;
    AddUnit(FolderNode, aUnit);
  end;
end;

procedure TcwTreeNode.AddUsedBy(const aParentNode : TcwTreeNode; const aUsedByList: TcwUsedByList);
var
  FolderNode : TcwTreeNode;
begin
  if aUsedByList.Count <= 0 then
    Exit;
  FolderNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(aParentNode, Format('Used By (%d)',
      [aUsedByList.Count])));
  FolderNode.HasChildren := (aUsedByList.Count > 0);
  FolderNode.UnitIntf := aParentNode.FUnit;
  FolderNode.SymbolIntf := aParentNode.FSymbol;
  FolderNode.NodeType := stUsedBy;
end;

procedure TcwTreeNode.AddLocation(const aLocation: TcwUsedByLocation);
var
  NewNode: TcwTreeNode;
  fName : string;
begin
  fName := ExtractFileName (aLocation.FileName);
  NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, Format('%s (%d,%d)',[fName,
                                                                            aLocation.X,
                                                                            aLocation.Y])));
  NewNode.UnitIntf := nil;
  NewNode.Location := aLocation;
  NewNode.SymbolIntf := nil;
  if aLocation.IsWrongCase then
    NewNode.NodeType := stLocationWrongCase else
      NewNode.NodeType := stLocation;
  NewNode.HasChildren := False;
end;

procedure TcwTreeNode.ExpandUsedBy;
var
  i: Integer;
  aLocation: TcwUsedByLocation;
  aUnit    : TcwUnit;
begin
  if (FUnit <> nil) then
  begin
    if (FSymbol <> nil) then
    begin
      for i := 0 to FSymbol.UsedBy.Count - 1 do
      begin
        aLocation := FSymbol.UsedBy[i];
        AddLocation(aLocation);
      end;
    end else
    begin
      for i := 0 to FUnit.UsedBy.Count - 1 do
      begin
        aLocation := FUnit.UsedBy[i];
        aUnit := aLocation.UnitIntf;
        AddUnit(Self,aUnit).FLocation :=  aLocation
      end;
    end;
  end;
end;


function TcwTreeNode.AddSymbol(const aSymbol: TcwSymbol; aType: TcwSymbolType) : TcwTreeNode;
var
  sText : string;
  sType : string;
begin
  sText := aSymbol.Name;
  if (aSymbol.TypeIntf <> nil)  then
    sType := aSymbol.TypeIntf.Name;
  begin
    if aSymbol is TcwType then
      sType :=  aSymbol.Name;
  end;
  if (sText <> sType) then
    sText := sText + ' : ' + sType;

  Result := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, sText));
  Result.UnitIntf := FUnit;
  Result.SymbolIntf := aSymbol;
  Result.NodeType := aType;
  Result.HasChildren := aSymbol.UsedBy.Count > 0;
end;


procedure TcwTreeNode.AddRoutineReturnValue(const aRoutine : TcwRoutine);
begin
  if aRoutine.ReturnType <> nil then
    AddSymbol(aRoutine.ReturnType,stParameter);
end;

procedure TcwTreeNode.AddRoutineParameters(const aRoutine : TcwRoutine);
var
  Node: TcwTreeNode;
begin
  if aRoutine.Parameters.Count > 0 then
  begin
    Node := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, Format('Parameters (%d)',[aRoutine.Parameters.Count])));
    Node.UnitIntf := FUnit;
    Node.SymbolIntf := aRoutine;
    Node.NodeType := stParameters;
    Node.HasChildren := True;
  end;
end;
procedure TcwTreeNode.ExpandVar;
begin
  if FUnit = nil then
    Exit;
  AddUsedBy(Self, FSymbol.UsedBy);
end;

procedure TcwTreeNode.ExpandVars(const aContainer : TcwVarContainer);
var
  i: Integer;
  aVariable: TcwVariable;
begin
  if aContainer = nil then EXIT;
  for i := 0 to aContainer.Variables.Count - 1 do
  begin
     aVariable := aContainer.Variables.SortedVariables[i];
     AddSymbol(aVariable, stVar);
  end;
end;

function TcwTreeNode.VarContainerHasChildren(const aContainer : TcwVarContainer) : Boolean;
begin
  Result := (aContainer.Classes.Count > 0) or (aContainer.Routines.Count > 0) or
            (aContainer.Variables.Count > 0) or (aContainer.Interfaces.Count > 0) or
            (aContainer.Constants.Count > 0) or (aContainer.Enumerations.Count > 0) or
            (aContainer.Sets.Count > 0) or (aContainer.Arrays.Count > 0);
end;

procedure TcwTreeNode.ExpandRoutines(const aContainer : TcwVarContainer);
var
  i: Integer;
  aRoutine: TcwRoutine;
  vnewNode : TcwTreeNode;
begin
  if aContainer = nil then Exit;
  for i := 0 to aContainer.Routines.Count - 1 do
  begin
     aRoutine := aContainer.Routines.SortedRoutines[i];
     vnewNode := AddSymbol(aRoutine, stRoutine);
     vnewNode.HasChildren := VarContainerHasChildren(aRoutine) or (aRoutine.UsedBy.Count > 0);
  end;
end;

procedure TcwTreeNode.ExpandRoutine;
begin
  if FUnit = nil then
    Exit;
  //AddRoutineReturnType(FSymbol as IcwRoutine);

  AddRoutineReturnValue(FSymbol as TcwRoutine);
  AddRoutineParameters(FSymbol as TcwRoutine);

  ExpandVarContainer(FSymbol as TcwRoutine);
  
  AddUsedBy(Self, FSymbol.UsedBy);
end;


procedure TcwTreeNode.ExpandParameters;
var
  i: Integer;
  aRoutine: TcwRoutine;
  aParameter : TcwParameter;
begin
  if FSymbol = nil then
    Exit;
  aRoutine := FSymbol as TcwRoutine;
  for i := 0 to aRoutine.Parameters.Count - 1 do
  begin
     aParameter := aRoutine.Parameters[i];
     AddSymbol(aParameter, stParameter);
  end;
end;

procedure TcwTreeNode.ExpandParameter;
begin
  if FUnit = nil then
    Exit;
  AddUsedBy(Self, FSymbol.UsedBy);
end;


procedure TcwTreeNode.ExpandType;
begin
  if FUnit = nil then
    Exit;
  AddUsedBy(Self, FSymbol.UsedBy);
end;

procedure TcwTreeNode.ExpandTypes(const aContainer : TcwVarContainer);
var
  i: Integer;
  aType: TcwSymbol;
begin
  if aContainer = nil then Exit;
  for i := 0 to aContainer.Types.Count - 1 do
  begin
     aType := aContainer.Types.SortedTypes[i];
     AddSymbol(aType, stType);
  end;
end;


procedure TcwTreeNode.ExpandEnumeration;
var vEnum : TcwEnumeration;
var i     : Integer;
begin
  vEnum := FSymbol as TcwEnumeration;
  for i := 0 to vEnum.Items.Count - 1 do
  begin
    AddSymbol(vEnum.Items[i], stType);
  end;
   // add a used by node
  AddUsedBy(Self, vEnum.UsedBy);
end;

procedure TcwTreeNode.ExpandSet;
var vSet : TcwSet;
var i     : Integer;
begin
  vSet := FSymbol as TcwSet;
  for i := 0 to vSet.Items.Count - 1 do
  begin
    AddSymbol(vSet.Items[i], stType);
  end;
   // add a used by node
  AddUsedBy(Self, vSet.UsedBy);
end;

procedure TcwTreeNode.ExpandEnumerations;
var
  i: Integer;
  aEnum: TcwEnumeration;
  aContainer: TcwVarContainer;
  NewNode   : TcwTreeNode;
begin
  if FUnit = nil then  Exit;
  if FSymbol = nil then  Exit;
  aContainer := FSymbol as TcwVarContainer;
  for i := 0 to aContainer.Enumerations.Count - 1 do
  begin
     aEnum := aContainer.Enumerations[i];
     NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, aEnum.Name));
     NewNode.UnitIntf := FUnit;
     NewNode.SymbolIntf := aEnum;
     NewNode.NodeType := stEnumeration;
     NewNode.HasChildren := (aEnum.Items.Count > 0);
  end;
end;

procedure TcwTreeNode.ExpandSets;
var
  i: Integer;
  aSet: TcwSet;
  aContainer: TcwVarContainer;
  NewNode   : TcwTreeNode;
begin
  if FUnit = nil then  Exit;
  if FSymbol = nil then  Exit;
  aContainer := FSymbol as TcwVarContainer;
  for i := 0 to aContainer.Sets.Count - 1 do
  begin
     aSet := aContainer.Sets[i];
     NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, aSet.Name));
     NewNode.UnitIntf := FUnit;
     NewNode.SymbolIntf := aSet;
     NewNode.NodeType := stSet;
     NewNode.HasChildren := (aSet.Items.Count > 0);
  end;
end;

procedure TcwTreeNode.ExpandArray;
var vArray : TcwArray;
//var i     : Integer;
begin
  vArray := FSymbol as TcwArray;

(*
  //dimensions not implemented yet;
  for i := 0 to vArray.DimensionCount - 1 do
  begin
    AddSymbol(vArray.Dimensions[i], stType);
  end;
*)

  // add a used by node
  AddUsedBy(Self, vArray.UsedBy);
end;

procedure TcwTreeNode.ExpandArrays;
var
  i: Integer;
  aArray: TcwArray;
  aContainer: TcwVarContainer;
  NewNode   : TcwTreeNode;
begin
  if FUnit = nil then  Exit;
  if FSymbol = nil then  Exit;
  aContainer := FSymbol as TcwVarContainer;
  for i := 0 to aContainer.Arrays.Count - 1 do
  begin
     aArray := aContainer.Arrays[i];
     NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, aArray.Name));
     NewNode.UnitIntf := FUnit;
     NewNode.SymbolIntf := aArray;
     NewNode.NodeType := stArray;
     NewNode.HasChildren := (aArray.DimensionCount > 0);
  end;
end;


procedure TcwTreeNode.ExpandConsts(const aContainer : TcwVarContainer);
var
  i: Integer;
  aVariable: TcwVariable;
begin
  if aContainer = nil then Exit;
  for i := 0 to aContainer.Constants.Count - 1 do
  begin
     aVariable := aContainer.Constants.SortedVariables[i];
     AddSymbol(aVariable, stConst);
  end;
end;

procedure TcwTreeNode.ExpandClass;
var aMembersList : TcwSymbolList;
begin
  if (FUnit = nil) or (FSymbol = nil) then
     Exit;
  TTreeView(TreeView).Items.BeginUpdate;
  try
    // add private data
    TcwClass(FSymbol).GetMembersWithVisibility(aMembersList, [vkPrivate]);
    AddClassMembers('Private', aMembersList);

    // add protected
    TcwClass(FSymbol).GetMembersWithVisibility(aMembersList,[vkProtected]);
    AddClassMembers('Protected', aMembersList);

    // add published
    TcwClass(FSymbol).GetMembersWithVisibility(aMembersList,[vkPublished]);
    AddClassMembers('Published', aMembersList);

    // add automated
    TcwClass(FSymbol).GetMembersWithVisibility(aMembersList,[vkAutomated]);
    AddClassMembers('Automated', aMembersList);

    // add public
    TcwClass(FSymbol).GetMembersWithVisibility(aMembersList,[vkPublic]);
    AddClassMembers('Public', aMembersList);

    // add a used by node
    AddUsedBy(Self, FSymbol.UsedBy);
  finally
    TTreeView(TreeView).Items.EndUpdate;
  end;
end;


procedure TcwTreeNode.ExpandClassDescendants;
var i : Integer;
var vClass       : TcwClass;
var NewNode      : TcwTreeNode;
begin
  if (FUnit = nil) or (FSymbol = nil) then
     Exit;
  TTreeView(TreeView).Items.BeginUpdate;
  try
    // add descendants
    for i := 0 to  TcwClass(FSymbol).Descendants.Count - 1 do
    begin
      vClass := TcwClass(FSymbol).Descendants[i] as TcwClass;
      // add a class node
      NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, vClass.Name));
      NewNode.UnitIntf := gcwProject.AllUnits[vClass.UnitIndex];
      NewNode.SymbolIntf := vClass;
      NewNode.NodeType := stClass;

      NewNode.HasChildren := (vClass.Descendants.Count > 0);

    end;
    // add a used by node
    AddUsedBy(Self, FSymbol.UsedBy);
  finally
    TTreeView(TreeView).Items.EndUpdate;
  end;
end;


procedure TcwTreeNode.ExpandInterfaceDescendants;
var i            : Integer;
var vInterface   : TcwClass;
var NewNode      : TcwTreeNode;
begin
  if (FUnit = nil) or (FSymbol = nil) then
     Exit;
  TTreeView(TreeView).Items.BeginUpdate;
  try
    // add descendants
    for i := 0 to  TcwClass(FSymbol).Descendants.Count - 1 do
    begin
      vInterface := TcwInterface(FSymbol).Descendants[i] as TcwClass;
      // add a class node
      NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, vInterface.Name));
      NewNode.UnitIntf := vInterface.UnitIntf;
      NewNode.SymbolIntf := vInterface;
      NewNode.NodeType := stInterface;

      NewNode.HasChildren := (vInterface.Descendants.Count > 0);

    end;
    // add a used by node
    AddUsedBy(Self, FSymbol.UsedBy);

  finally
    TTreeView(TreeView).Items.EndUpdate;
  end;
end;


procedure TcwTreeNode.ExpandClassesDescendants;
var i, j : Integer;
var vUnit  : TcwUnit;
var vClass : TcwClass;
var NewNode : TcwTreeNode;
begin
    for i := 0 to gcwProject.AllUnits.Count - 1 do
    begin
      vUnit := gcwProject.AllUnits[i];
      for j := 0 to vUnit.Classes.Count - 1 do
      begin
        vClass := vUnit.Classes[j];
        if vClass.InheritsFromClass = nil then
        begin
           NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self,  vClass.Name));
           NewNode.UnitIntf := vUnit;
           NewNode.SymbolIntf := vClass;
           NewNode.NodeType := stClass;
           NewNode.HasChildren := (vClass.Descendants.Count > 0) ;
        end;
      end;
    end;
    for j := 0 to gcwProject.UnparsedClasses.Count - 1 do
    begin
      vClass := gcwProject.UnparsedClasses[j];
      if vClass.InheritsFromClass = nil then
      begin
        NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self,  vClass.Name));
         NewNode.UnitIntf := gcwProject.ProjectUnit;
        NewNode.SymbolIntf := vClass;
        NewNode.NodeType := stClass;
        NewNode.HasChildren := (vClass.Descendants.Count > 0) ;
      end;
    end;
end;


procedure TcwTreeNode.ExpandObjectsDescendants;
var i, j : Integer;
var vUnit  : TcwUnit;
var vClass : TcwClass;
var NewNode : TcwTreeNode;
begin
    for i := 0 to gcwProject.AllUnits.Count - 1 do
    begin
      vUnit := gcwProject.AllUnits[i];
      for j := 0 to vUnit.Objects.Count - 1 do
      begin
        vClass := vUnit.Objects[j];
        if vClass.InheritsFromClass = nil then
        begin
           NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self,  vClass.Name));
           NewNode.UnitIntf := vUnit;
           NewNode.SymbolIntf := vClass;
           NewNode.NodeType := stObject;
           NewNode.HasChildren := (vClass.Descendants.Count > 0) ;
        end;
      end;
    end;
    for j := 0 to gcwProject.UnparsedObjects.Count - 1 do
    begin
      vClass := gcwProject.UnparsedObjects[j];
      if vClass.InheritsFromClass = nil then
      begin
         NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self,  vClass.Name));
         NewNode.UnitIntf := gcwProject.ProjectUnit;
         NewNode.SymbolIntf := vClass;
         NewNode.NodeType := stObject;
         NewNode.HasChildren := (vClass.Descendants.Count > 0) ;
      end;
    end;
end;

procedure TcwTreeNode.ExpandInterfacesDescendants;
var i, j       : Integer;
var vUnit      : TcwUnit;
var vInterface : TcwClass;
var NewNode    : TcwTreeNode;
begin
    for i := 0 to gcwProject.AllUnits.Count - 1 do
    begin
      vUnit := gcwProject.AllUnits[i];
      for j := 0 to vUnit.Interfaces.Count - 1 do
      begin
        vInterface := vUnit.Interfaces[j];
        if vInterface.Interfaces.Count = 0 then
        begin
           NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self,  vInterface.Name));
           NewNode.UnitIntf := vUnit;
           NewNode.SymbolIntf := vInterface;
           NewNode.NodeType := stInterface;
           NewNode.HasChildren := (vInterface.Descendants.Count > 0) ;
        end;
      end;
    end;
    for j := 0 to gcwProject.UnparsedInterfaces.Count - 1 do
    begin
      vInterface := gcwProject.UnparsedInterfaces[j];
      if vInterface.Interfaces.Count = 0 then
      begin
         NewNode := TcwTreeNode(TTreeView(TreeView).Items.AddChild(Self, vInterface.Name));
         NewNode.UnitIntf := gcwProject.ProjectUnit;
         NewNode.SymbolIntf := vInterface;
         NewNode.NodeType := stInterface;
         NewNode.HasChildren := (vInterface.Descendants.Count > 0) ;
      end;
    end;
end;


procedure TcwTreeNode.ExpandNode(ShowUsageTree : Boolean);
begin
  if ShowUsageTree then
  begin
    case NodeType of
      stUnit: ExpandUnit;
      stInterfaces: ExpandInterfaces(FSymbol as TcwVarContainer);

      stRecords: ExpandRecords(FSymbol as TcwVarContainer);
      stRecord : ExpandRecord;
      stClasses: ExpandClasses(FSymbol as TcwVarContainer);
      stObjects: ExpandObjects(FSymbol as TcwVarContainer);
      stInterface: ExpandClass;
      stClass: ExpandClass;
      stObject: ExpandClass;
      stUsedBy: ExpandUsedBy;
      stVars: ExpandVars(FSymbol as TcwVarContainer);
      stVar: ExpandVar;
      stRoutines: ExpandRoutines(FSymbol as TcwVarContainer);
      stRoutine: ExpandRoutine;
      stConsts: ExpandConsts(FSymbol as TcwVarContainer);
      stConst: ExpandVar;
      stTypes: ExpandTypes(FSymbol as TcwVarContainer);
      stType: ExpandType;
      stEnumerations : ExpandEnumerations;
      stEnumeration  : ExpandEnumeration;
      stSets : ExpandSets;
      stSet  : ExpandSet;
      stArrays       : ExpandArrays;
      stArray        : ExpandArray;
      stParameters : ExpandParameters;
      stParameter  : ExpandParameter;
    end;
  end else
  begin
    case NodeType of
      stClass: ExpandClassDescendants;
      stClasses: ExpandClassesDescendants;
      stObject: ExpandClassDescendants;
      stObjects: ExpandObjectsDescendants;

      stInterface: ExpandInterfaceDescendants;
      stInterfaces: ExpandInterfacesDescendants;      
      stUsedBy: ExpandUsedBy;
    end;
  end;
end;

procedure TcwTreeNode.SetNodeType(const aValue: TcwSymbolType);
begin
  FNodeType := aValue;
  ImageIndex := ImageIndexes[aValue];
  SelectedIndex := ImageIndexes[aValue];
end;

procedure TcwTreeNode.Execute;
var
  sFileName: string;
  vX, vY: Integer;
  aLoc: TcwCodeLocation;
begin
  sFileName := '';
  vX := 0;
  vY := 0;
  
  aLoc := Location;
  if aLoc = nil then
    aLoc := SymbolIntf;
  
  if aLoc <> nil then
  begin
    sFileName := aLoc.FileName;
    vX := aLoc.X;
    vY := aLoc.Y;
  end else
  if UnitIntf <> nil then
  begin
    sFileName := UnitIntf.FullFileName;
    vX := 0;
    vY := 0;
  end;
  
  gcwProject.ShowSourceFile(sFileName, vX, vY);
end;

end.
