(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwTypeParser;

interface

uses
  SysUtils,
  Classes,
  Types,
  mwPasLexTypes,
  mwSimplePasParTypes,
  mwSimplePasPar,
  ucwClasses,
  ucwInterfaceParser;

type
  TTypeParser = class(TInterfaceParser)
  protected
    procedure ParseExportsClause;
    procedure ParseLabelId(const aVarContainer: TcwVarContainer);
    procedure ParseLabelDeclarationSection(const aVarContainer: TcwVarContainer);
    procedure ParseTypeSection(const aVarContainer: TcwVarContainer);
    function ParseDeclarationSection(const aVarContainer: TcwVarContainer): TcwRoutine; override;
    procedure InterfaceDeclaration; override;
  end;

implementation

uses
  ucwUnitParser;

procedure TTypeParser.ParseExportsClause;
begin
  Expected(ptExports);
  ExportsElement;
  while TokenID = ptComma do
  begin
    NextToken;
    ExportsElement;
  end;
  SEMICOLON;
end;

procedure TTypeParser.ParseLabelId(const aVarContainer: TcwVarContainer);
begin
  case TokenID of
    ptIntegerConst: begin
      aVarContainer.Labels.AddLabel(Lexer.Token, True, FileIndex,
        Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
      NextToken;
    end;
    ptIdentifier: begin
      aVarContainer.Labels.AddLabel(Lexer.Token, False, FileIndex,
        Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
      NextToken;
    end;
    else
      SynError(InvalidLabelId);
  end;
end;

procedure TTypeParser.ParseLabelDeclarationSection(const aVarContainer: TcwVarContainer);
begin
  Expected(ptLabel);
  ParseLabelId(aVarContainer);
  while (TokenID = ptComma) do
  begin
    NextToken;
    ParseLabelId(aVarContainer);
  end;
  SEMICOLON;
end;

procedure TTypeParser.ParseTypeSection(const aVarContainer: TcwVarContainer);

  procedure ParseTypeDeclaration(const aVarContainer: TcwVarContainer);
  var
    vClassName: string;
    vTypeName: string;
    vTypeLocation: TPoint;
    AClass: TcwSymbol;
  begin
    vTypeName := Lexer.Token;
    vTypeLocation := Lexer.PosXY;
    TypeName;
    Expected(ptEqual);
    if TokenID = ptType then
      ExpliciteType;
    Lexer.InitAhead;
    case TokenID of
      ptClass: begin
        case Lexer.AheadTokenID of
          ptOf: begin
            Expected(ptClass);
            Expected(ptOf);
            vClassName := Lexer.Token;

            AClass := FindSymbol(vClassName, True, GetUsedByInfo, False);
            if not Assigned(AClass) or not (AClass is TcwClass) then
              ParseClassForward(vClassName, vTypeLocation);

            ParseTypeDefinition(aVarContainer, vTypeName, vTypeLocation);
          end;
          ptSemiColon: begin
            Expected(ptClass);
            AClass := FindSymbol(vTypeName, True, GetUsedByInfo, False);
            if not Assigned(AClass) or not (AClass is TcwClass) then
              ParseClassForward(vTypeName, vTypeLocation);
          end;
          else
            ParseClassType(vTypeName, vTypeLocation);
        end;
      end;
      ptInterface: begin
        case Lexer.AheadTokenID of
          ptSemiColon: begin
            Expected(ptInterface);
            ParseInterfaceForward(vTypeName, vTypeLocation);
          end;
          else
            ParseInterfaceType(vTypeName, vTypeLocation);
        end;
      end;
      ptDispInterface: begin
        case Lexer.AheadTokenID of
          ptSemiColon: begin
            Expected(ptDispInterface);
            ParseInterfaceForward(vTypeName, vTypeLocation);
          end;
          else
            ParseInterfaceType(vTypeName, vTypeLocation);
            //InterfaceType;
        end;
      end;
      ptObject: begin
        case Lexer.AheadTokenID of
          ptSemiColon: ParseObjectForward(vTypeName, vTypeLocation);
          else
            ParseObjectType(vTypeName, vTypeLocation);
        end;
      end;
      else
        ParseTypeDefinition(aVarContainer, vTypeName, vTypeLocation);
    end;
  end;

begin
  IncIndentLevel;
  Expected(ptType);
  while TokenID = ptIdentifier do
  begin
    ParseTypeDeclaration(aVarContainer);
    if TokenId = ptEqual then //jdj 8/2/00
      TypedConstant;
    SEMICOLON;
  end;
  DecIndentLevel;
end;

function TTypeParser.ParseDeclarationSection(const aVarContainer: TcwVarContainer): TcwRoutine;
begin
  Result := nil;
  case TokenID of
    ptClass: ParseProcedureDeclarationSection(aVarContainer);
    ptConst: ParseConstSection(aVarContainer);
    ptConstructor: ParseProcedureDeclarationSection(aVarContainer);
    ptDestructor: ParseProcedureDeclarationSection(aVarContainer);
    ptExports: ParseExportsClause;
    ptFunction: ParseProcedureDeclarationSection(aVarContainer);
    ptLabel: ParseLabelDeclarationSection(aVarContainer);
    ptProcedure: ParseProcedureDeclarationSection(aVarContainer);
    ptResourceString: ParseConstSection(aVarContainer);
    ptType: ParseTypeSection(aVarContainer);
    ptThreadVar: ParseVarSection(aVarContainer);
    ptVar: ParseVarSection(aVarContainer);
    else
      SynError(InvalidDeclarationSection);
  end;
end;

procedure TTypeParser.InterfaceDeclaration;
begin
  case TokenID of
    ptConst: ParseConstSection(UnitIntf);
    ptFunction: ExportedHeading;
    ptProcedure: ExportedHeading;
    ptResourceString: ParseConstSection(UnitIntf);
    ptType: ParseTypeSection(UnitIntf);
    ptThreadVar: ParseVarSection(UnitIntf);
    ptVar: ParseVarSection(UnitIntf);
    else
      SynError(InvalidInterfaceDeclaration);
  end;
end;

end.
