(*  Copyright (c) 2002, the CodeLens project
  All rights reserved.

  Redistribution and use in source and binary forms,
  with or without modification, are permitted provided
  that the following conditions are met:

  Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.

  Redistributions in binary form must reproduce the
  above copyright notice, this list of conditions
  and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
  Neither the name of the CodeLens project nor the names
  of its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
  AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. *)

unit ucwUnitParser;

interface

uses
  SysUtils,
  Classes,
  Types,
  Contnrs,
  mwDirectiveLex,
  mwPasLexTypes,
  mwSimplePasPar,
  mwPasLex,
  ucwClasses;

const
  SSysInitPas = 'SysInit';
  SSystemPas = 'System';
  SCodeWrenchSystemPas = 'CodeWrenchSystem';

type
  DPointer = record
    FRoutine: TcwRoutine;
    FClass: TcwClass;
  end;

  TUnitParser = class(TmwSimplePasPar)
  private
    FInsideExternalDirective: Boolean;
    FFullFileName: string;
    FFileStream: TMemoryStream;
    FIncludeList: TList;
    FFileIndex: Integer;
    FFileIndexStack: TStack;
    FUnitIntf: TcwUnit;
    //FUnitStats: TcwCodeStatistics;
    FScopeVariables: TcwHashedObjectList;
    FInterfaceParsed: Boolean;
    FParseInterfaceOnly: Boolean;
    //stack of routines currently being parsed
    FRoutinesStack: array of DPointer;
    FStackCount: Integer;
    FStackLength: Integer;
    //current unit section (interface/implementation)
    FUnitVisibility: TVisibilityKind;
    FFirstInLine: Boolean;
    FIndentLevel: Integer;
    FIndentSpaces: Integer;
    function RoutineStackCount: Integer;
    procedure ProcessUsedUnit(const aUnitName: string; const aFullFileName: string; aX, aY: Integer);
    procedure OnParseMessage(Sender: TObject; const aMsg: string; X, Y: Integer);
    procedure OnLineEnd(Sender: TmwBasePasLex);
    procedure HandleLineIndent;
    function DoIncludeDirective(Sender: TmwBasePasLex; Value: String): PChar;
    procedure DoFinishedIncludeDirective(Sender: TmwBasePasLex);
    function GetUnitIntf: TcwUnit;
  protected
    FCurrentClass: TcwClass;
    FParsingClassInterface: TcwClass;

    property UnitIntf: TcwUnit read GetUnitIntf;
    property FileIndex: Integer read FFileIndex;
    property UnitVisibility: TVisibilityKind read FUnitVisibility write FUnitVisibility;

    procedure PushRoutine(const aRoutine: TcwRoutine; aClass: TcwClass);
    procedure PopRoutine;

    //if we are in a routine, will return the current routine, else the unit
    function CurrentContainer: TcwVarContainer;

    function FindSymbol(const aSymbolName: string; const AAddUsedBy: Boolean;
      const AUsedByInfo: TcwUsedByInfo; AResultExpected: Boolean = True): TcwSymbol;
    function FindType(const aSymbolName: string; AForPointerType: Boolean = False): TcwSymbol;
    function FindVisibleClass(const aClassName: string): TcwClass;
    function FindVisibleObject(const aClassName: string): TcwClass;
    function FindVisibleInterface(const aInterfaceName: string): TcwInterface;

    procedure GetQualifiedIdentifierList(const aStrings: TStrings);
    procedure GetIdentifierList(const aStrings: TStrings);
    function ParseOrdinalIdentifier: TcwSymbol;
    function ParseRealIdentifier: TcwSymbol;
    function ParseStringIdentifier: TcwSymbol;
    function ParseVariantIdentifier: TcwSymbol;
    function ParseTypeId(AForPointerType: Boolean = False): TcwSymbol;
    function ParseOrdinalType(const aContainer: TcwSymbolContainer; const aCodeLocation: TPoint;
      out AResultExpected: Boolean; const AEnumName: string = ''): TcwSymbol;
    function ParseRealType: TcwType;
    procedure ParseStringType;
    procedure ParseCharString;
    function ParseSimpleType(const aContainer: TcwSymbolContainer; const aTypeName: string;
      const aCodeLocation: TPoint; out ATypeAdded: Boolean; out AResultExpected: Boolean):
        TcwSymbol;
    function ParseStructuredType(const aContainer: TcwSymbolContainer; const aTypeName: string;
      const aCodeLocation: TPoint; out ATypeAdded: Boolean): TcwSymbol; virtual;
    function ParsePointerType(const aContainer: TcwVarContainer; const aTypeName: string;
      const aCodeLocation: TPoint): TcwSymbol;
    function ParseEnumeratedType(const aContainer: TcwVarContainer; const aCodeLocation: TPoint;
      const aTypeName: string = ''): TcwSymbol;
    procedure ParseEnumerationMembers(const aContainer: TcwVarContainer;
      const aMemberContainer: TcwEnumerationMemberContainer = nil);
    function ParseTypeKind(const aContainer: TcwSymbolContainer; const aTypeName: string;
      const aCodeLocation: TPoint; AParsingType: Boolean): TcwSymbol;
    function ParseTypeDefinition(const aContainer: TcwVarContainer; const aTypeName: string;
      const aCodeLocation: TPoint): TcwSymbol;
    function ParseDeclarationSection(const aVarContainer: TcwVarContainer): TcwRoutine; virtual;
    procedure ParseCompoundStatement(const aVarContainer: TcwVarContainer);
    procedure ParseBlock(const aVarContainer: TcwVarContainer);
    function GetUsedByInfo: TcwUsedByInfo;
    procedure ExternalDirective; override;
  protected
    property ScopeVariables: TcwHashedObjectList read FScopeVariables;
    procedure IncIndentlevel;
    procedure DecIndentlevel;

    procedure ExceptionHandler; override;
    procedure Designator; override;
    procedure ExportsElement; override;
    function ParseQualifiedIdentifier(const aSymbol: TcwSymbol): TcwSymbol; virtual; abstract;
    function ParseVariableReference(const aSymbol: TcwSymbol): TcwSymbol; virtual; abstract;

    procedure InterfaceSection; override;
    procedure ImplementationSection; override;
    procedure MainUsesClause; override;
    procedure MainUsedUnitStatement; override;

    procedure ContainsClause; override;
    procedure ContainsStatement; override;

    procedure NextToken; override;
    procedure ProgramBlock; override;
    procedure ResourceIncluded(const aResName: string); override;

    procedure UnitFile; override;
    procedure UsesClause; override;
    procedure UsedUnitName; override;
    procedure ParseInitializationSection;

    //.sm helper method to save codelocation of declaration end
    procedure StoreDeclarationEnd(aCodeLocation: TcwCodeLocation);
  public
    constructor Create(const AFullFileName: string; const aDefines: TmwAssignAbleHash;
      AKeyHashTable: TmwPasHash; ADirectiveHash: TmwDirectiveHash);
    destructor Destroy; override;

    //parse the implementation section of the unit
    procedure ParseImplementation;

    //parse the interface section of the unit
    procedure ParseInterface(AUnitIntf: TcwUnit);
  end;

implementation

uses
  mwSimplePasParTypes;

constructor TUnitParser.Create(const AFullFileName: string; const aDefines: TmwAssignAbleHash;
  AKeyHashTable: TmwPasHash; ADirectiveHash: TmwDirectiveHash);
begin
  inherited Create(AKeyHashTable, ADirectiveHash);
  FFullFileName := AFullFileName;
  FFileStream := TMemoryStream.Create;
  FFileStream.LoadFromFile(AFullFileName);
  FIncludeList := TList.Create;
  FFileIndex := -1;
  FFileIndexStack := TStack.Create;
  FScopeVariables := TcwHashedObjectList.Create;
  FIndentSpaces := gcwProject.CodingConventions.IndentationConventions.SpacesPerIndent;
  FFirstInLine := True;
  OnMessage := OnParseMessage;
  Lexer.OnLineEnd := OnLineEnd;
  Lexer.DefinedList.Assign(aDefines);
  Lexer.OnIncludeDirect := DoIncludeDirective;
  Lexer.OnFinishedIncludeDirect := DoFinishedIncludeDirective;
end;

destructor TUnitParser.Destroy;
var
  Index: Integer;
begin
  FScopeVariables.Free;
  for Index := 0 to FIncludeList.Count - 1 do
    StrDispose(PChar(FIncludeList[Index]));
  FIncludeList.Free;
  FFileIndexStack.Free;
  FFileStream.Free;
  inherited;
end;

function TUnitParser.DoIncludeDirective(Sender: TmwBasePasLex; Value: String): PChar;
var
  ASourceFile: TcwSourceFile;
  AFileName: string;
  AUnitType: TDirectoryType;
  AStringList: TStringList;
begin
  AFileName := gcwProject.ResolveUnitName(Value, ExtractFileDir(UnitIntf.FileName),
    AUnitType);
  if AFileName = '' then
  begin
    Result := nil;
    if Assigned(OnMessage) then
      OnMessage(Self, Format('Include file, %s, does not exist.', [Value]), Lexer.PosXY.X + 1,
        Lexer.PosXY.Y + 1);
  end else
  begin
    AStringList := TStringList.Create;
    try
      FFileIndexStack.Push(Pointer(FFileIndex));
      ASourceFile := gcwProject.FileNames.Find(AFileName);
      if Assigned(ASourceFile) then
        FFileIndex := ASourceFile.FileIndex
      else
        FFileIndex := gcwProject.FileNames.Add(AFileName, True, UnitIntf.UnitIndex);
      AStringList.LoadFromFile(AFileName);
      Result := StrAlloc(Length(AStringList.Text) + 1);
      StrPCopy(Result, AStringList.Text);
      FIncludeList.Add(Result);
    finally
      AStringList.Free;
    end;
  end;
end;

procedure TUnitParser.DoFinishedIncludeDirective(Sender: TmwBasePasLex);
begin
  if FFileIndexStack.Count > 0 then
    FFileIndex := Integer(FFileIndexStack.Pop)
  else
    FFileIndex := -1;
end;

procedure TUnitParser.PushRoutine(const aRoutine: TcwRoutine; aClass: TcwClass);
begin
  if FStackLength <= FStackCount then
  begin
    Inc(FStackLength, 50);
    SetLength(FRoutinesStack, FStackLength);
  end;
  FRoutinesStack[FStackCount].FRoutine := aRoutine;
  FRoutinesStack[FStackCount].FClass := aClass;
  if aClass <> nil then
    FCurrentClass := aClass;
  //save routine impl start pos ; .sm 21042006;
  if Assigned(aRoutine) then
  begin
    ARoutine.ImplStartX := Lexer.PosXY.X + 1;
    ARoutine.ImplStartY := Lexer.PosXY.Y + 1;
  end;
  inc(FStackCount);
end;

procedure TUnitParser.PopRoutine;
begin
  if FStackCount > 0 then
  begin
    if (FRoutinesStack[FStackCount - 1].FClass = FCurrentClass) then
      FCurrentClass := nil;
    Dec(FStackCount);
    FRoutinesStack[FStackCount].FRoutine := nil;
    FRoutinesStack[FStackCount].FClass := nil;
  end else
    FCurrentClass := nil;
end;

//if we are in a routine, will return the current routine, else the unit
function TUnitParser.CurrentContainer: TcwVarContainer;
begin
  if RoutineStackCount > 0 then
    Result := FRoutinesStack[RoutineStackCount - 1].FRoutine
  else
    Result := UnitIntf;
end;

function TUnitParser.RoutineStackCount: Integer;
begin
  Result := FStackCount;
end;

procedure TUnitParser.StoreDeclarationEnd(aCodeLocation: TcwCodeLocation);
begin
  if Assigned(aCodeLocation) then
  begin
    aCodeLocation.EndX:=Lexer.PosXY.X+1;
    aCodeLocation.EndY:=Lexer.PosXY.Y+1;
  end;
end;

procedure TUnitParser.IncIndentlevel;
begin
  Inc(FIndentLevel, FIndentSpaces);
end;

procedure TUnitParser.DecIndentlevel;
begin
  if FIndentLevel > 0 then
    Dec(FIndentLevel, FIndentSpaces);
end;

procedure TUnitParser.OnParseMessage(Sender: TObject; const aMsg: string; X, Y: Integer);
begin
  gcwProject.Messages.AddMessage(aMsg, FFileIndex, X + 1, Y + 1, not FInterfaceParsed);
  if FInterfaceParsed then
    FUnitIntf.UnitWarnings := FUnitIntf.UnitWarnings + [uwImplementation]
  else
    FUnitIntf.UnitWarnings := FUnitIntf.UnitWarnings + [uwInterface];
end;

procedure TUnitParser.ProcessUsedUnit(const aUnitName: string; const aFullFileName: string;
  aX, aY: Integer);
var
  NewUnit: TcwUnit;
begin
  NewUnit := gcwProject.AllUnits.FindUnit(aUnitName);
  if not Assigned(NewUnit) then
    NewUnit := gcwProject.AddUnit(aUnitName, aFullFileName);

  UnitIntf.UsedUnits.AddUsedUnit(TcwUsedUnit.Create(NewUnit));
  if gcwProject.FileNames[NewUnit.FileIndex].FileExists then
    gcwProject.ParseInterface(NewUnit);

  {if (NewUnit.Parsed = psrInterface) and
     (FUnitVisibility = vkImplementation) then
       NewUnit.FullParse;}
  NewUnit.UsedBy.AddUsedBy(UsedByInfo(FFileIndex, NewUnit.Name <> aUnitName, aX, aY));
end;

function TUnitParser.FindSymbol(const aSymbolName: string; const AAddUsedBy: Boolean;
  const AUsedByInfo: TcwUsedByInfo; AResultExpected: Boolean = True): TcwSymbol;
var
  AUsedUnit: TcwUsedUnit;
  i: Integer;
  vRoutine: TcwRoutine;
  vType: TcwSymbol;
  vClass: TcwClass;
  vUnit: TcwUnit;
  aFromClass: TcwClass;
begin
  for i := FScopeVariables.Count - 1 downto 0 do
  begin
    vType := (FScopeVariables[i] as TcwSymbol).TypeIntf;
    if (vType <> nil) and (vType.TypeIntf <> nil) then
    begin
      if vType is TcwClass then
        aFromClass := vType as TcwClass
      else
        aFromClass := FCurrentClass;
      Result := vType.TypeIntf.FindSymbol(UnitIntf, aFromClass, aSymbolName, AAddUsedBy,
        AUsedByInfo);
      if Assigned(Result) then
        Exit;
    end;
  end;

  //first locate in current stack of routines
  for i := RoutineStackCount - 1 downto 0 do
  begin
    vRoutine := FRoutinesStack[i].FRoutine;
    Result := vRoutine.FindSymbol(UnitIntf, FCurrentClass, aSymbolName, AAddUsedBy,
      AUsedByInfo);
    if Assigned(Result) then
      Exit;
    vClass := FRoutinesStack[i].FClass;
    if vClass <> nil then
    begin
      Result := vClass.FindSymbol(UnitIntf, FCurrentClass, aSymbolName, AAddUsedBy,
        AUsedByInfo);
      if Assigned(Result) then
        Exit;
    end;
  end;

  // if we're parsing the class interface section, and the class refers to one of
  // its own members, e.g. property Test stored IsTestStored;
  // IsTestStored should be found as belonging to the current class
  if Assigned(FParsingClassInterface) then
  begin
    Result := FParsingClassInterface.FindSymbol(UnitIntf, FParsingClassInterface, aSymbolName,
      AAddUsedBy, AUsedByInfo);
    if Assigned(Result) then
      Exit;
  end;

  //find in current unit
  Result := UnitIntf.FindSymbol(UnitIntf, FCurrentClass, aSymbolName, AAddUsedBy,
    AUsedByInfo);
  if Assigned(Result) then
    Exit;

  AUsedUnit := UnitIntf.UsedUnits.FindUsedUnit(aSymbolName);
  if Assigned(AUsedUnit) then
  begin
    Result := AUsedUnit.UnitIntf;
    Exit;
  end;

  //find in all used units
  for i := UnitIntf.UsedUnits.Count - 1 downto 0 do
  begin
    vUnit := UnitIntf.UsedUnits[i].UnitIntf;
    Result := vUnit.FindSymbol(UnitIntf, FCurrentClass, aSymbolName, AAddUsedBy,
      AUsedByInfo);
    if Assigned(Result) then
      Exit;
  end;

  Result := gcwProject.GetPredefinedType(ASymbolName);
  if Assigned(Result) then
    Exit;

  //TODO: suppress warning messages option
  if AResultExpected and Assigned(OnMessage) then
    OnMessage(Self, Format('Symbol, %s, not found in project', [ASymbolName]), Lexer.PosXY.X,
      Lexer.PosXY.Y);
end;

function TUnitParser.FindType(const aSymbolName: string; AForPointerType: Boolean = False):
  TcwSymbol;
var
  i: Integer;
  vRoutine: TcwRoutine;
  vUnit: TcwUnit;
begin
  //first locate in current stack of routines
  for i := RoutineStackCount - 1 downto 0 do
  begin
    vRoutine := FRoutinesStack[i].FRoutine;
    Result := vRoutine.Types.Find(aSymbolName);
    if Assigned(Result) then
      Exit;
  end;

  //find in current unit
  Result := UnitIntf.Types.Find(aSymbolName);
  if Assigned(Result) then
    Exit;

  //find in all used units
  for i := UnitIntf.UsedUnits.Count - 1 downto 0 do
  begin
    vUnit := UnitIntf.UsedUnits[i].UnitIntf;
    Result := vUnit.Types.Find(aSymbolName);
    if Assigned(Result) then
      Exit;
  end;

  Result := gcwProject.GetPredefinedType(ASymbolName);
  if Assigned(Result) then
    Exit;

  //TODO: the type might be forward declaration, like a pointer to a type thats not defined yet
  // in this case we're going to call CreateSymbolByName, but we need to check that the symbol
  // that gets created will have its type set later on
  //TODO: suppress warning messages option
  if not AForPointerType and Assigned(OnMessage) then
    OnMessage(Self, Format('Type, %s, not found in project', [ASymbolName]), Lexer.PosXY.X,
      Lexer.PosXY.Y);
end;

function TUnitParser.FindVisibleClass(const aClassName: string): TcwClass;
var
  i: Integer;
  vUnit: TcwUnit;
begin
  //find in current unit
  Result := UnitIntf.Classes.Find(aClassName);
  if Assigned(Result) then
    Exit;

  //find in all used units
  for i := UnitIntf.UsedUnits.Count - 1 downto 0 do
  begin
    vUnit := UnitIntf.UsedUnits[i].UnitIntf;
    Result := vUnit.Classes.Find(aClassName);
    if Assigned(Result) then
      Exit;
  end;
end;

function TUnitParser.FindVisibleObject(const aClassName: string): TcwClass;
var
  i: Integer;
  vUnit: TcwUnit;
begin
  //find in current unit
  Result := UnitIntf.Objects.Find(aClassName);
  if Assigned(Result) then
    Exit;

  //find in all used units
  for i := UnitIntf.UsedUnits.Count - 1 downto 0 do
  begin
    vUnit := UnitIntf.UsedUnits[i].UnitIntf;
    Result := vUnit.Objects.Find(aClassName);
    if Assigned(Result) then
      Exit;
  end;
end;

function TUnitParser.FindVisibleInterface(const aInterfaceName: string): TcwInterface;
var
  i: Integer;
  vUnit: TcwUnit;
begin
  //find in current unit
  Result := UnitIntf.Interfaces.Find(aInterfaceName);
  if Assigned(Result) then
    Exit;

  //find in all used units
  for i := UnitIntf.UsedUnits.Count - 1 downto 0 do
  begin
    vUnit := UnitIntf.UsedUnits[i].UnitIntf;
    Result := vUnit.Interfaces.Find(aInterfaceName);
    if Assigned(Result) then
      Exit;
  end;
end;

procedure TUnitParser.GetIdentifierList(const aStrings: TStrings);
var
  Pt: TSmallPoint;
begin
  Pt.x := Lexer.PosXY.X + 1;
  Pt.y := Lexer.PosXY.y + 1;
  if Lexer.Token <> '' then
    aStrings.AddObject(Lexer.Token, Pointer(Pt));
  Expected(ptIdentifier);
  while (TokenID = ptComma) do
  begin
    NextToken;
    Pt.x := Lexer.PosXY.X + 1;
    Pt.y := Lexer.PosXY.y + 1;
    if Lexer.Token <> '' then
      aStrings.AddObject(Lexer.Token, Pointer(Pt));
    Expected(ptIdentifier);
  end;
end;

procedure TUnitParser.GetQualifiedIdentifierList(const aStrings: TStrings);
var
  Pt: TSmallPoint;
begin
  Pt.x := Lexer.PosXY.X + 1;
  Pt.y := Lexer.PosXY.y + 1;
  aStrings.AddObject(Lexer.Token, Pointer(Pt));
  ParseQualifiedIdentifier(nil);
  while (TokenID = ptComma) do
  begin
    NextToken;
    Pt.x := Lexer.PosXY.X + 1;
    Pt.y := Lexer.PosXY.y + 1;
    aStrings.AddObject(Lexer.Token, Pointer(Pt));
    ParseQualifiedIdentifier(nil);
  end;
end;

function TUnitParser.GetUnitIntf: TcwUnit;
begin
  Result := FUnitIntf as TcwUnit;
end;

function TUnitParser.ParseOrdinalIdentifier: TcwSymbol;
begin
  Result := nil;
  case ExID of
    ptBoolean:
      begin
        Result := gcwProject.GetPredefinedType(tkBoolean);
        NextToken;
      end;
    ptByte:
      begin
        Result := gcwProject.GetPredefinedType(tkByte);
        NextToken;
      end;
    ptBytebool:
      begin
        Result := gcwProject.GetPredefinedType(tkBytebool);
        NextToken;
      end;
    ptCardinal:
      begin
        Result := gcwProject.GetPredefinedType(tkCardinal);
        NextToken;
      end;
    ptChar:
      begin
        Result := gcwProject.GetPredefinedType(tkChar);
        NextToken;
      end;
    ptDWord:
      begin
        Result := gcwProject.GetPredefinedType(tkDWord);
        NextToken;
      end;
    ptInt64:
      begin
        Result := gcwProject.GetPredefinedType(tkInt64);
        NextToken;
      end;
    ptInteger:
      begin
        Result := gcwProject.GetPredefinedType(tkInteger);
        NextToken;
      end;
    ptLongBool:
      begin
        Result := gcwProject.GetPredefinedType(tkLongBool);
        NextToken;
      end;
    ptLongInt:
      begin
        Result := gcwProject.GetPredefinedType(tkLongInt);
        NextToken;
      end;
    ptLongWord:
      begin
        Result := gcwProject.GetPredefinedType(tkLongWord);
        NextToken;
      end;
    ptPChar:
      begin
        Result := gcwProject.GetPredefinedType(tkPChar);
        NextToken;
      end;
    ptShortInt:
      begin
        Result := gcwProject.GetPredefinedType(tkShortInt);
        NextToken;
      end;
    ptSmallInt:
      begin
        Result := gcwProject.GetPredefinedType(tkSmallInt);
        NextToken;
      end;
    ptWideChar:
      begin
        Result := gcwProject.GetPredefinedType(tkWideChar);
        NextToken;
      end;
    ptWord:
      begin
        Result := gcwProject.GetPredefinedType(tkWord);
        NextToken;
      end;
    ptWordbool:
      begin
        Result := gcwProject.GetPredefinedType(tkWordbool);
        NextToken;
      end;
    else
      begin
        SynError(InvalidOrdinalIdentifier);
      end;
  end;
end;

function TUnitParser.ParseRealIdentifier: TcwSymbol;
begin
  Result := nil;
  case ExID of
    ptReal48:
      begin
        Result := gcwProject.GetPredefinedType(tkReal48);
        NextToken;
      end;
    ptReal:
      begin
        Result := gcwProject.GetPredefinedType(tkReal);
        NextToken;
      end;
    ptSingle:
      begin
        Result := gcwProject.GetPredefinedType(tkSingle);
        NextToken;
      end;
    ptDouble:
      begin
        Result := gcwProject.GetPredefinedType(tkDouble);
        NextToken;
      end;
    ptExtended:
      begin
        Result := gcwProject.GetPredefinedType(tkExtended);
        NextToken;
      end;
    ptCurrency:
      begin
        Result := gcwProject.GetPredefinedType(tkCurrency);
        NextToken;
      end;
    ptComp:
      begin
        Result := gcwProject.GetPredefinedType(tkComp);
        NextToken;
      end;
    else
      begin
        SynError(InvalidRealIdentifier);
      end;
  end;
end;

function TUnitParser.ParseStringIdentifier: TcwSymbol;
begin
  Result := nil;
  case ExID of
    ptAnsiString:
      begin
        Result := gcwProject.GetPredefinedType(tkAnsiString);
        NextToken;
      end;
    ptShortString:
      begin
        Result := gcwProject.GetPredefinedType(tkShortString);
        NextToken;
      end;
    ptWideString:
      begin
        Result := gcwProject.GetPredefinedType(tkWideString);
        NextToken;
      end;
    else
      begin
        SynError(InvalidStringIdentifier);
      end;
  end;
end;

function TUnitParser.ParseVariantIdentifier: TcwSymbol;
begin
  Result := nil;
  case ExID of
    ptOleVariant:
      begin
        Result := gcwProject.GetPredefinedType(tkOleVariant);
        NextToken;
      end;
    ptVariant:
      begin
        Result := gcwProject.GetPredefinedType(tkVariant);
        NextToken;
      end;
    else
      begin
        SynError(InvalidVariantIdentifier);
      end;
  end;
end;

function TUnitParser.ParseTypeId(AForPointerType: Boolean = False): TcwSymbol;
var
  aX, aY: Integer;
  vIsWrongCase: Boolean;
  aUsedUnit: TcwUsedUnit;
begin
  aUsedUnit := nil;
  vIsWrongCase := False;
  Lexer.InitAhead;
  if Lexer.AheadTokenID = ptPoint then
  begin
    aUsedUnit := UnitIntf.UsedUnits.FindUsedUnit(Lexer.Token);
    if not Assigned(aUsedUnit) and Assigned(OnMessage) then
      //TODO: suppress warning messages option
      OnMessage(Self, Format('Unit, %s, not found among used units', [Lexer.Token]),
        Lexer.PosXY.X, Lexer.PosXY.Y);
    UnitId;
    Expected(ptPoint);
  end;
  aX := Lexer.PosXY.X + 1;
  aY := Lexer.PosXY.Y + 1;
  case GenID of
    ptBoolean, ptByte, ptChar, ptDWord, ptInt64, ptInteger, ptLongInt,
    ptLongWord, ptPChar, ptShortInt, ptSmallInt, ptWideChar, ptWord, ptCardinal:
      Result := ParseOrdinalIdentifier;
    ptComp, ptCurrency, ptDouble, ptExtended, ptReal, ptReal48, ptSingle:
      Result := ParseRealIdentifier;
    ptAnsiString, ptShortString, ptWideString: Result := ParseStringIdentifier;
    ptOleVariant, ptVariant: Result := ParseVariantIdentifier;
    ptString: begin
      Result := gcwProject.GetPredefinedType(tkString);
      ParseStringType;
    end;
    else
    begin
      if Assigned(aUsedUnit) then
        Result := aUsedUnit.UnitIntf.Types.Find(Lexer.Token)
      else
        Result := FindType(Lexer.Token, AForPointerType);
      vIsWrongCase := (Assigned(Result)) and (Lexer.Token <> Result.Name);
      case GenID of
        ptFile: Expected(ptFile);
        else
          Expected(ptIdentifier);
      end;
    end;
  end;
  if Assigned(Result) then
    Result.UsedBy.AddUsedBy(UsedByInfo(FFileIndex, vIsWrongCase, aX, aY));
end;

function TUnitParser.ParseOrdinalType(const aContainer: TcwSymbolContainer;
  const aCodeLocation: TPoint; out AResultExpected: Boolean; const AEnumName: string = ''):
    TcwSymbol;
begin
  AResultExpected := False;
  Result := nil;
  case TokenID of
    ptIdentifier: begin
      Lexer.InitAhead;
      case Lexer.AheadTokenID of
        ptPoint: Expression;
        ptRoundOpen: ConstantExpression; //jdj
        else
        begin
          AResultExpected := True;
          Result := ParseTypeId;
        end;
      end;
    end;
    ptRoundOpen: begin
      AResultExpected := True;
      Result := ParseEnumeratedType(aContainer as TcwVarContainer, aCodeLocation, AEnumName);
    end;
    ptSquareOpen: begin
      NextToken;
      SubrangeType;
      Expected(ptSquareClose);
    end;
    else
      Expression;
  end;
  if TokenID = ptDotDot then
  begin
    NextToken;
    ConstantExpression;
  end;
end;

function TUnitParser.ParseRealType: TcwType;
begin
  Result := nil;
  case TokenID of
    ptMinus:
      begin
        NextToken;
      end;
    ptPlus:
      begin
        NextToken;
      end;
  end;
  case TokenId of
    ptFloat:
      begin
        NextToken;
      end;
    else
      begin
        ParseVariableReference(nil);
      end;
  end;
end;

procedure TUnitParser.ParseStringType;
begin
  case TokenID of
    ptString: begin
      NextToken;
      if TokenID = ptSquareOpen then
      begin
        NextToken;
        ConstantExpression;
        Expected(ptSquareClose);
      end;
    end;
    else
      VariableReference;
  end;
end;

procedure TUnitParser.ParseCharString;
begin
  UnitIntf.StringLiterals.AddSymbol(Lexer.Token, vkPublic, FFileIndex, Lexer.PosXY.X + 1,
    Lexer.PosXY.Y + 1);
  case TokenID of
    ptAsciiChar, ptIdentifier, ptRoundOpen, ptStringConst: begin
      while (TokenID in [ptAsciiChar, ptIdentifier, ptPlus, ptRoundOpen, ptStringConst]) and
        // not a hint directive
        {(TokenID <> ptLibrary) and } not (ExID in [ptDeprecated, ptPlatform]) and
        // not a external directive where ptName is used...
        not (FInsideExternalDirective and (ExID = ptName)) do
        case TokenID of
          ptIdentifier, ptRoundOpen: ParseVariableReference(nil);
          else
            NextToken;
        end;
    end;
    else
      SynError(InvalidCharString);
  end;
end;

procedure TUnitParser.ExportsElement;
begin
  Expected(ptIdentifier);
  if Lexer.ExID = ptIndex then
  begin
    NextToken;
    Expected(ptIntegerConst);
  end;
  if Lexer.ExID = ptName then
  begin
    NextToken;
    ParseCharString;
  end;
  //  if TokenID = ptResident then
  if Lexer.ExID = ptResident then //jdj 20001207
  begin
    NextToken;
  end;
end;

procedure TUnitParser.Designator;
begin
  ParseVariableReference(nil);
end;

function TUnitParser.ParseSimpleType(const aContainer: TcwSymbolContainer; const aTypeName: string;
  const aCodeLocation: TPoint; out ATypeAdded: Boolean; out AResultExpected: Boolean): TcwSymbol;
begin
  AResultExpected := False;
  Result := nil;
  ATypeAdded := False;
  case TokenID of
    ptMinus: NextToken;
    ptPlus: NextToken;
  end;
  case Lexer.TokenID of
    ptAsciiChar, ptIntegerConst: Result := ParseOrdinalType(aContainer, aCodeLocation,
      AResultExpected, aTypeName);
    ptFloat: begin
      AResultExpected := True;
      Result := ParseRealType;
    end;
    ptIdentifier: begin
      Lexer.InitAhead;
      case Lexer.AheadTokenID of
        ptPoint, ptSemiColon: begin
          AResultExpected := True;
          Result := ParseTypeId;
        end;
        else
        begin
          SimpleExpression;
          if Lexer.TokenID = ptDotDot then
          begin
            NextToken;
            SimpleExpression;
          end;
        end;
      end;
    end;
    ptRoundOpen: begin
      ParseEnumeratedType(aContainer as TcwVarContainer, aCodeLocation, aTypeName);
      ATypeAdded := True;
    end;
    ptSquareOpen: SubrangeType;
    else
      ParseVariableReference(nil);
  end;
  StoreDeclarationEnd(Result);
end;

function TUnitParser.ParseStructuredType(const aContainer: TcwSymbolContainer; const aTypeName: string; const
  aCodeLocation: TPoint; out ATypeAdded: Boolean): TcwSymbol;
begin
  Result := nil;
  Assert(False, 'Abstract method call');
end;

function TUnitParser.ParsePointerType(const aContainer: TcwVarContainer; const aTypeName: string;
  const aCodeLocation: TPoint): TcwSymbol;
var
  PtrTo: TcwSymbol;
  APointerTypeName: string;
begin
  Expected(ptPointerSymbol);
  // (cbm) save the forward pointer name (e.g. linked list)
  if GenID = ptIdentifier then
    APointerTypeName := Lexer.Token
  else
    APointerTypeName := '';
  PtrTo := ParseTypeId(True);

  Assert(aTypeName <> '');
  Result := aContainer.Types.AddType(aTypeName, UnitVisibility, FFileIndex,
    aCodeLocation.X + 1, aCodeLocation.Y + 1, True);
  Result.TypeIntf := PtrTo;
  Result.PointerType := True;
  Result.PointerTypeName := APointerTypeName;
  aContainer.AddForwardPointerType(Result as TcwType);
end;

function TUnitParser.ParseEnumeratedType(const aContainer: TcwVarContainer;
  const aCodeLocation: TPoint; const aTypeName: string = ''): TcwSymbol;
var
  Enum: TcwEnumeration;
begin
  if aTypeName <> '' then
    Enum := aContainer.Enumerations.AddItem(aTypeName, UnitVisibility,
      FFileIndex, aCodeLocation.X+1, aCodeLocation.Y+1)
  else
    Enum := nil;
  ParseEnumerationMembers(aContainer, Enum);
  StoreDeclarationEnd(Enum);

  Result := Enum as TcwSymbol;
end;

procedure TUnitParser.ParseEnumerationMembers(const aContainer: TcwVarContainer;
  const aMemberContainer: TcwEnumerationMemberContainer = nil);

  procedure ParseEnumerationMember;
  var
    ASymbol: TcwSymbol;
  begin
    if Assigned(aMemberContainer) then
    begin
      ASymbol := aMemberContainer.Items.AddSymbol(Lexer.Token, UnitVisibility, FFileIndex, Lexer.PosXY.X + 1,
        Lexer.PosXY.y + 1);
      aContainer.AddSymbol(ASymbol);
    end else
      UnitIntf.Constants.AddVariable(Lexer.Token, UnitVisibility, True, FFileIndex,
        Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
    Expected(ptIdentifier);
    if TokenID = ptEqual then
    begin
      Expected(ptEqual);
      Expression;
    end;
  end;

begin
  Expected(ptRoundOpen);
  ParseEnumerationMember;
  while TokenID = ptComma do
  begin
    NextToken;
    ParseEnumerationMember;
  end;
  Expected(ptRoundClose);
end;

function TUnitParser.ParseTypeKind(const aContainer: TcwSymbolContainer; const aTypeName: string;
  const aCodeLocation: TPoint; AParsingType: Boolean): TcwSymbol;

  procedure AddParsedType(const ATypeName: string; const ANewTypeType: TcwSymbol;
    ACodeLocation: TPoint; const AResultExpected: Boolean);
  var
    ANewType: TcwSymbol;
  begin
    if Assigned(ANewTypeType) or not AResultExpected then
    begin
      if AParsingType then
      begin
        Assert(aContainer is TcwVarContainer);
        ANewType := (aContainer as TcwVarContainer).Types.AddType(aTypeName, vkInterface,
          FFileIndex, ACodeLocation.X + 1, ACodeLocation.Y + 1, AResultExpected);
        ANewType.TypeIntf := ANewTypeType;
      end else
        if Assigned(ANewTypeType) then
          ANewTypeType.UsedBy.AddUsedBy(GetUsedByInfo);
    end else
      if Assigned(OnMessage) then
        OnMessage(Self, Format('Type not found for type named, %s', [ATypeName]),
          ACodeLocation.X, ACodeLocation.Y);
  end;

var
  AResultExpected: Boolean;
  ATypeAdded: Boolean;
begin
  Result := nil;
  case TokenID of
    ptAsciiChar, ptFloat, ptIntegerConst, ptMinus, ptNil, ptPlus, ptRoundOpen,
    ptSquareOpen, ptStringConst: begin
      Result := ParseSimpleType(aContainer, aTypeName, aCodeLocation, ATypeAdded, AResultExpected);
      if not ATypeAdded then
        AddParsedType(ATypeName, Result, ACodeLocation, AResultExpected);
    end;
    ptArray, ptFile, ptPacked, ptRecord, ptSet: begin
      Result := ParseStructuredType(aContainer, aTypeName, aCodeLocation, ATypeAdded);
      if not ATypeAdded then
        AddParsedType(ATypeName, Result, ACodeLocation, True);
    end;
    ptFunction, ptProcedure: begin
      // get the predefined type...
      Result := FindType(Lexer.Token);
      ProceduralType;
      AddParsedType(ATypeName, Result, ACodeLocation, True);
    end;
    ptIdentifier: begin
      Lexer.InitAhead;
      case Lexer.AheadTokenID of
        // (cbm) added ptEquals so we get the type for initialized vars
        // (cbm) added ptRoundClose in case we're ending a variant section in a variant record
        // (cbm) added ptOf for record variants, parsing the type for tag fields...
        ptOf, ptRoundClose, ptEqual, ptPoint, ptSemiColon: begin
          Result := ParseTypeId;
          AddParsedType(ATypeName, Result, ACodeLocation, True);
        end;
        else
        begin
          if (Lexer.AheadExID in [ptDeprecated, ptPlatform]) or
            (Lexer.AheadTokenID = ptLibrary) then
          begin
            Result := ParseTypeId;
            AddParsedType(ATypeName, Result, ACodeLocation, False);
            HintDirectives;
          end else
          begin
            AddParsedType(ATypeName, nil, ACodeLocation, False);
            SimpleExpression;
            if Lexer.TokenID = ptDotDot then
            begin
              NextToken;
              SimpleExpression;
            end;
          end;
        end;
      end;
    end;
    ptPointerSymbol: begin
      if AParsingType then
        Result := ParsePointerType(aContainer as TcwVarContainer, ATypeName, aCodeLocation)
        // in this case Result is already added to UnitIntf.Types
        //AddParsedType(ATypeName, Result, ACodeLocation);
      else
      begin
        Expected(ptPointerSymbol);
        Result := ParseTypeId(True);
      end;
    end;
    ptString: begin
      Result := gcwProject.GetPredefinedType(tkString);
      AddParsedType(ATypeName, Result, ACodeLocation, True);
      ParseStringType;
    end;
    else
      SynError(InvalidTypeKind);
  end;
end;

function TUnitParser.ParseTypeDefinition(const aContainer: TcwVarContainer; const aTypeName: string;
  const aCodeLocation: TPoint): TcwSymbol;
begin
  Result := ParseTypeKind(aContainer, aTypeName, aCodeLocation, True);
end;

function TUnitParser.ParseDeclarationSection(const aVarContainer: TcwVarContainer): TcwRoutine;
begin
  Result := nil;
  Assert(False, 'Abstract method call');
end;

procedure TUnitParser.ParseCompoundStatement(const aVarContainer: TcwVarContainer);
begin
  IncIndentLevel;
  Expected(ptBegin);
  StatementList;
  DecIndentLevel;
  Expected(ptEnd);
end;

procedure TUnitParser.ParseBlock(const aVarContainer: TcwVarContainer);
begin
  while TokenID in [ptClass, ptConst, ptConstructor, ptDestructor, ptExports,
    ptFunction, ptLabel, ptProcedure, ptResourceString, ptThreadVar, ptType, ptVar] do
    ParseDeclarationSection(aVarContainer);
  case TokenID of
    ptAsm: AsmStatement;
    else
      ParseCompoundStatement(aVarContainer);
  end;
end;

procedure TUnitParser.InterfaceSection;
begin
  FUnitVisibility := vkInterface;
  UnitIntf.InterfaceLine:=Lexer.PosXY.Y+1; 
  inherited;
end;

procedure TUnitParser.ImplementationSection;
begin
  FUnitVisibility := vkImplementation;
  UnitIntf.ImplementationLine:=Lexer.PosXY.Y+1; 
  Expected(ptImplementation);
  if TokenID = ptUses then
    UsesClause;
  while TokenID in [ptClass, ptConst, ptConstructor, ptDestructor, ptFunction,
    ptLabel, ptProcedure, ptResourceString, ptThreadVar, ptType, ptVar] do //ptResourceString added jdj
    ParseDeclarationSection(UnitIntf);
end;

procedure TUnitParser.MainUsesClause;
begin
  IncIndentlevel;
  Expected(ptUses);
  MainUsedUnitStatement;
  while TokenID = ptComma do
  begin
    NextToken;
    MainUsedUnitStatement;
  end;
  DecIndentlevel;
  SEMICOLON;
end;

procedure TUnitParser.MainUsedUnitStatement;
var
  UnitToken: string;
  UnitPos: TPoint;
begin
  UnitToken := Lexer.Token;
  UnitPos := Lexer.PosXY;

  MainUsedUnitName;
  if Lexer.TokenID = ptIn then
  begin
    NextToken;
    UnitToken := AnsiDequotedStr(Lexer.Token, '''');
    MainUsedUnitExpression;
    ProcessUsedUnit(ChangeFileExt(ExtractFileName(UnitToken), ''), UnitToken, UnitPos.X + 1, UnitPos.Y + 1);
  end else
    ProcessUsedUnit(UnitToken, '', UnitPos.X + 1, UnitPos.Y + 1);
end;

procedure TUnitParser.ContainsClause;
begin
  //inherited ContainsClause;

  IncIndentlevel;
  ExpectedEx(ptContains);
  ContainsStatement;
  while TokenID = ptComma do
  begin
    NextToken;
    ContainsStatement;
  end;
  DecIndentlevel;
  SEMICOLON;
end;

procedure TUnitParser.ContainsStatement;
var
  UnitToken: string;
  UnitPos: TPoint;
begin
  //inherited ContainsStatement;

  UnitToken := Lexer.Token;
  UnitPos := Lexer.PosXY;

  ContainsIdentifier;
  if Lexer.TokenID = ptIn then
  begin
    NextToken;
    UnitToken := AnsiDequotedStr(Lexer.Token, '''');
    ContainsExpression;
    ProcessUsedUnit(ChangeFileExt(ExtractFileName(UnitToken), ''), UnitToken, UnitPos.X + 1, UnitPos.Y + 1);
  end else
    ProcessUsedUnit(UnitToken, '', UnitPos.X + 1, UnitPos.Y + 1);
end;

procedure TUnitParser.HandleLineIndent;
begin
  if FFirstInLine then
  begin
    if Lexer.PosXY.X <> FIndentLevel then
    begin
      UnitIntf.CodingErrorsList.AddError(ekBadIndentation, FFileIndex,
        Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
    end;
    FFirstInLine := False;
  end;
end;

procedure TUnitParser.NextToken;
begin
  Lexer.Next;
  SkipJunk;
  if TokenID = ptEnd then
    DecIndentLevel;
  HandleLineIndent;
end;

procedure TUnitParser.ProgramBlock;
begin
  if TokenID = ptUses then
    MainUsesClause;
  ParseBlock(UnitIntf);
end;

procedure TUnitParser.ResourceIncluded(const aResName: string);
begin
  if SameText(aResName, '*.dfm') then
    UnitIntf.LoadFormResource(ChangeFileExt(UnitIntf.FullFileName, '.dfm'));
end;

procedure TUnitParser.OnLineEnd(Sender: TmwBasePasLex);
begin
  if Sender = Lexer then
  begin
//TODO    if Sender.TokenID = ptSlashesComment then
//      FUnitStats.IncSlashCommentLines
//    else if Sender.TokenID = ptAnsiComment then
//      FUnitStats.IncAnsiCommentLines
//    else if Sender.TokenID = ptBorComment then
//      FUnitStats.IncBorCommentLines
//    else if Sender.TokenID = ptCRLF then
//      FUnitStats.IncEmptyLines
//    else
//      FUnitStats.IncCodeLines;
//    FFirstInLine := True;
  end;
end;

procedure TUnitParser.UnitFile;
begin
  HandleLineIndent;
  Expected(ptUnit);

  UnitIntf.X := Lexer.PosXY.X + 1;
  UnitIntf.Y := Lexer.PosXY.Y + 1;

  Expected(ptIdentifier);
  HintDirectives;
  SEMICOLON;
  InterfaceSection;
  FInterfaceParsed := True;
  if FParseInterfaceOnly = True then
    Exit;

  ImplementationSection;
  ParseInitializationSection;
  Expected(ptPoint);
end;

procedure TUnitParser.UsesClause;
begin
  IncIndentlevel;
  Expected(ptUses);
  UsedUnitsList;
  DecIndentlevel;
  SEMICOLON;
end;

procedure TUnitParser.UsedUnitName;
begin
  ProcessUsedUnit(Lexer.Token, '', Lexer.posXY.X + 1, Lexer.posXY.Y + 1);
  inherited;
end;

procedure TUnitParser.ParseImplementation;
begin
  if FInterfaceParsed then
  begin
    ImplementationSection;
    ParseInitializationSection;
    Expected(ptPoint);
  end;
end;

procedure TUnitParser.ParseInitializationSection;
begin
  case TokenID of
    ptInitialization: begin
      NextToken;
      StatementList();
      if TokenID = ptFinalization then
      begin
        NextToken;
        StatementList;
      end;
      Expected(ptEnd);
    end;
    ptFinalization: begin
      //finalization without an initiaization section
      NextToken;
      StatementList;
      Expected(ptEnd);
    end;

    ptBegin: ParseCompoundStatement(UnitIntf);
    ptEnd: NextToken;
    else
      SynError(InvalidInitializationSection);
  end;
end;

procedure TUnitParser.ParseInterface(AUnitIntf: TcwUnit);
begin
  FUnitVisibility := vkInterface;
  FUnitIntf := AUnitIntf;
  FFileIndex := AUnitIntf.FileIndex;
  //TODO: FUnitStats := AUnitIntf.CodeStats;
  FParseInterfaceOnly := True;

//  Run(FFullFileName, FFileStream);
//  Exit;

  // CodeWrenchSystem.pas should not use any other unit...
  if AUnitIntf.Name <> SCodeWrenchSystemPas then
  begin
    ProcessUsedUnit(SCodeWrenchSystemPas, '', -1, -1);
    if not SameText(AUnitIntf.Name, SSystemPas) then
    begin
      if not SameText(AUnitIntf.Name, SSysInitPas) then
        ProcessUsedUnit(SSysInitPas, '', -1, -1);
      // Note: System.pas has SysInit in its implementation uses clause...
      ProcessUsedUnit(SSystemPas, '', -1, -1);
    end;
  end;
  Run(FFullFileName, FFileStream);
end;

procedure TUnitParser.ExceptionHandler;

  procedure ExceptionIdentifier;
  begin
    Lexer.InitAhead;
    case Lexer.AheadTokenID of
      ptPoint: ExceptionClassTypeIdentifier;
      else
      begin
        CurrentContainer.Variables.AddVariable(Lexer.Token, UnitIntf.Visibilty, False,
          FileIndex, Lexer.PosXY.X + 1, Lexer.PosXY.Y + 1);
        NextToken;
        case Lexer.TokenID of
          ptColon: begin
            NextToken;
            ParseTypeKind(CurrentContainer, Lexer.Token, Lexer.PosXY, False);
          end;
        end;
      end;
    end;
  end;

begin
  ExpectedEx(ptOn);
  ExceptionIdentifier;
  Expected(ptDo);
  Statement;
end;

function TUnitParser.GetUsedByInfo: TcwUsedByInfo;
begin
  Result.FileIndex := FFileIndex;
  Result.X := Lexer.PosXY.X + 1;
  Result.Y := Lexer.PosXY.Y + 1;
  Result.IsWrongCase := False;
end;

procedure TUnitParser.ExternalDirective;
begin
  FInsideExternalDirective := True;
  try
    inherited;
  finally
    FInsideExternalDirective := False;
  end;
end;

end.
